theory Constructing_Markov_Chain
  imports Discrete_Markov_Kernel
begin

section {* Markov Chain as Stochastic Process *} text_raw {* \label{sec:stochasticprocess} *}

text {*
  We can construct for each time-homogeneous discrete-time Markov process a corresponding
  probability space using @{const Discrete_Markov_Kernel}. The constructed probability space
  has the same probabilities.
*}

locale Discrete_Time_Markov_Chain = M: prob_space +
  fixes S :: "'s set" and X :: "nat \<Rightarrow> 'a \<Rightarrow> 's"
  assumes S: "countable S"
  assumes X[measurable]: "\<And>t. X t \<in> measurable M (count_space S)"
  assumes MC: "\<And>n s s'.
    \<P>(\<omega> in M. \<forall>t\<le>n. X t \<omega> = s t ) \<noteq> 0 \<Longrightarrow>
    \<P>(\<omega> in M. X (Suc n) \<omega> = s' \<bar> \<forall>t\<le>n. X t \<omega> = s t ) =
    \<P>(\<omega> in M. X (Suc n) \<omega> = s' \<bar> X n \<omega> = s n )"
  assumes TH: "\<And>n m s t.
    \<P>(\<omega> in M. X n \<omega> = t) \<noteq> 0 \<Longrightarrow> \<P>(\<omega> in M. X m \<omega> = t) \<noteq> 0 \<Longrightarrow>
    \<P>(\<omega> in M. X (Suc n) \<omega> = s \<bar> X n \<omega> = t) = \<P>(\<omega> in M. X (Suc m) \<omega> = s \<bar> X m \<omega> = t)"
begin

definition "S' = Some ` S \<union> {None}"

lemma None_in_S'[simp]: "None \<in> S'"
  by (simp add: S'_def)

lemma S_not_empty: "S \<noteq> {}"
  using X[THEN measurable_space] M.not_empty by auto

lemma S'_not_empty: "S' \<noteq> {}"
  by (auto simp: S'_def)

lemma measurable_Some[measurable]: "Some \<in> measurable (count_space S) (count_space S')"
  by (auto simp: S'_def)

lemma countable_S': "countable S'"
  by (auto simp: S S'_def)

definition "with P f d = (if \<exists>x. P x then f (SOME x. P x) else d)"

lemma withI[case_names default exists]:
  "((\<And>x. \<not> P x) \<Longrightarrow> Q d) \<Longrightarrow> (\<And>x. P x \<Longrightarrow> Q (f x)) \<Longrightarrow> Q (with P f d)"
  unfolding with_def by (auto intro: someI2)

subsection {* Construct @{text K'} *} text_raw {* \label{sec:mcstoch} *}

primrec K' :: "'s option \<Rightarrow> 's option measure" where
  "K' None = distr M (count_space S') (\<lambda>\<omega>. Some (X 0 \<omega>))"
| "K' (Some s) = 
    with (\<lambda>n. 0 < \<P>(\<omega> in M. X n \<omega> = s))
      (\<lambda>n. distr (uniform_measure M {\<omega>\<in>space M. X n \<omega> = s}) (count_space S') (\<lambda>\<omega>. Some (X (Suc n) \<omega>)))
      (point_measure S' (\<lambda>x. if x = Some (SOME s. s \<in> S) then 1 else 0))"

lemma sets_K[simp]: "sets (K' s) = Pow S'"
proof (cases s)
  case (Some s') then show ?thesis
    apply simp
    apply (rule withI)
    apply (simp_all add: sets_point_measure)
    done
qed simp

lemma space_K[simp]: "space (K' s) = S'"
  using sets_eq_imp_space_eq[of "K' s" "count_space S'"] by simp

subsection {* Equations for @{const K'} *}

lemma emeasure_K_S[simp]:
  assumes "0 < \<P>(\<omega> in M. X n \<omega> = s)"
  shows "emeasure (K' (Some s)) {Some t} = \<P>(\<omega> in M. X (Suc n) \<omega> = t \<bar> X n \<omega> = s)"
  unfolding K'.simps
proof (induct rule: withI)
  case default with assms show ?case by simp
next
  case (exists m)
  let ?M = "distr (uniform_measure M {\<omega> \<in> space M. X m \<omega> = s}) (count_space S') (\<lambda>\<omega>. Some (X (Suc m) \<omega>))"
  show ?case
  proof cases
    assume "t \<in> S"
    then have t: "Some t \<in> S'"
      by (simp add: S'_def)
    have "emeasure ?M {Some t} = \<P>(\<omega> in M. X (Suc m) \<omega> = t \<bar> X m \<omega> = s)"
      using exists
      by (auto simp add: emeasure_distr t cond_prob_def emeasure_eq_measure
               intro!: arg_cong[where f=prob])
    also have "\<dots> = \<P>(\<omega> in M. X (Suc n) \<omega> = t \<bar> X n \<omega> = s)"
      using assms exists by (auto intro!: TH)
    finally show ?thesis .
  next
    assume "t \<notin> S"
    then have "emeasure ?M {Some t} = 0"
      by (intro emeasure_notin_sets) (auto simp: S'_def)
    moreover
    have *: "{\<omega> \<in> space M. X (Suc n) \<omega> = t \<and> X n \<omega> = s} = {}"
      using `t \<notin> S` measurable_space[OF X] by auto
    have "0 = \<P>(\<omega> in M. X (Suc n) \<omega> = t \<bar> X n \<omega> = s)"
      by (simp add: cond_prob_def *)
    ultimately show ?thesis
      by simp
  qed
qed

lemma emeasure_K_None:
  "emeasure (K' (Some s)) {None} = 0"
  unfolding K'.simps by (rule withI) (auto simp: emeasure_point_measure emeasure_distr vimage_def)

lemma emeasure_K_init[simp]: "emeasure (K' None) {t} = \<P>(\<omega> in M. Some (X 0 \<omega>) = t)"
proof cases
  assume "t \<notin> S'"
  then have *: "{\<omega>\<in>space M. Some (X 0 \<omega>) = t} = {}"
    using measurable_space[OF X] by (auto simp: S'_def)
  from `t \<notin> S'` have "emeasure (distr M (count_space S') (\<lambda>\<omega>. Some (X 0 \<omega>))) {t} = 0"
    by (intro emeasure_notin_sets) (auto simp: S'_def)
  then show ?thesis
    by (simp add: *)
qed (auto simp add: emeasure_distr emeasure_eq_measure intro!: arg_cong[where f=prob])

sublocale K!: prob_space "K' s" for s
proof (cases s)
  case (Some s')
  show "prob_space (K' s)"
    unfolding Some K'.simps
  proof (induct rule: withI)
    case (exists n) then show ?case
      by (intro prob_space.prob_space_distr prob_space_uniform_measure)
         (auto simp: emeasure_eq_measure)
  qed (default, simp add: space_point_measure emeasure_point_measure someI_in S'_def S_not_empty)
qed (auto intro!: prob_space_distr)

sublocale MC: Discrete_Markov_Kernel S' K'
  by default (auto simp add: S'_def S)

declare K'.simps[simp del]

subsection {* Markov Kernel's Path Space equals Stochastic Process *} text_raw {* \label{sec:KhasX} *}

lemma paths_None_eq:
  "paths None = distr M S_seq (\<lambda>\<omega> n. Some (X n \<omega>))" (is "?L = ?R")
proof (rule nat_eq_cyl[OF countable_S'])
  interpret R!: prob_space ?R
      by (intro M.prob_space_distr) simp
    
  show "sets ?L = sets MC.S_seq" "sets ?R = sets MC.S_seq" "emeasure ?R (space MC.S_seq) \<noteq> \<infinity>"
    by simp_all

  fix \<omega>' :: "nat \<Rightarrow> 's option" and n :: nat assume \<omega>': "\<And>i. \<omega>' i \<in> S'"
  then have "emeasure ?L {\<omega> \<in> space ?L. \<forall>i<Suc n. \<omega> i = \<omega>' i} = 
    \<P>(\<omega> in M. Some (X 0 \<omega>) = \<omega>' 0) * (\<Prod>i<n. emeasure (K' (\<omega>' i)) {\<omega>' (Suc i)})"  
    by (subst emeasure_path) (auto simp: lessThan_Suc_eq_insert_0 zero_notin_Suc_image setprod_reindex)
  also have "\<P>(\<omega> in M. Some (X 0 \<omega>) = \<omega>' 0) * (\<Prod>i<n. emeasure (K' (\<omega>' i)) {\<omega>' (Suc i)})
    = emeasure M {\<omega> \<in> space M. \<forall>i\<le>n. Some (X i \<omega>) = \<omega>' i}"
  proof cases
    assume *: "\<exists>i\<le>n. \<omega>' i = None"
    with smallest[of "\<lambda>i. \<omega>' i = None"]
    obtain i where i: "\<omega>' i = None" "\<And>j. j < i \<Longrightarrow> \<omega>' j \<noteq> None" "i \<le> n"
      by (metis order_trans)

    then have "{\<omega> \<in> space M. \<forall>i\<le>n. Some (X i \<omega>) = \<omega>' i} = {}"
      by auto
    then have "emeasure M {\<omega> \<in> space M. \<forall>i\<le>n. Some (X i \<omega>) = \<omega>' i} = 0"
      by (simp del: Collect_empty_eq)
    also have "\<P>(\<omega> in M. Some (X 0 \<omega>) = \<omega>' 0) * (\<Prod>i<n. emeasure (K' (\<omega>' i)) {\<omega>' (Suc i)}) = 0"
    proof (cases i)
      case 0 with i show ?thesis
        by simp
    next
      case (Suc j)
      with i have "\<omega>' j \<noteq> None" by auto
      with Suc i(1) have "emeasure (K' (\<omega>' j)) {\<omega>' (Suc j)} = 0"
        by (auto simp add: emeasure_K_None)
      with Suc i(3) show ?thesis
        by (auto simp: setprod_ereal_0)
    qed
    finally show ?thesis .
  next
    assume "\<not> (\<exists>i\<le>n. \<omega>' i = None)"
    then obtain \<omega> where \<omega>: "\<And>i. i \<le> n \<Longrightarrow> \<omega>' i = Some (\<omega> i)"
      by atomize_elim (rule choice, auto)

    have "\<P>(\<omega>' in M. X 0 \<omega>' = \<omega> 0) * (\<Prod>i<n. emeasure (K' (Some (\<omega> i))) {Some (\<omega> (Suc i))}) =
      emeasure M {\<omega>' \<in> space M. \<forall>i\<le>n. X i \<omega>' = \<omega> i}"
    proof (induct n)
      case 0 then show ?case
        by (simp add: M.emeasure_eq_measure)
    next
      case (Suc n)
      have le: "\<P>(\<omega>' in M. X (Suc n) \<omega>' = \<omega> (Suc n) \<and> (\<forall>i\<le>n. X i \<omega>' = \<omega> i)) \<le> \<P>(\<omega>' in M. \<forall>i\<le>n. X i \<omega>' = \<omega> i)"
        by (auto intro!: M.finite_measure_mono)

      have "\<P>(\<omega>' in M. X 0 \<omega>' = \<omega> 0) * (\<Prod>i<Suc n. emeasure (K' (Some (\<omega> i))) {Some (\<omega> (Suc i))}) =
          emeasure M {\<omega>' \<in> space M. \<forall>i\<le>n. X i \<omega>' = \<omega> i} * emeasure (K' (Some (\<omega> n))) {Some (\<omega> (Suc n))}"
        unfolding Suc[symmetric] by (simp add: lessThan_Suc zero_notin_Suc_image setprod_reindex ac_simps)
      also have "\<dots> = emeasure M {\<omega>' \<in> space M. \<forall>i\<le>n. X i \<omega>' = \<omega> i} *
        \<P>(\<omega>' in M. X (Suc n) \<omega>' = \<omega> (Suc n) \<bar> \<forall>i\<le>n. X i \<omega>' = \<omega> i )"
      proof (intro ereal_left_mult_cong)
        assume "emeasure M {\<omega>' \<in> space M. \<forall>i\<le>n. X i \<omega>' = \<omega> i} \<noteq> 0"
        then have **: "0 < \<P>(\<omega>' in M. \<forall>i\<le>n. X i \<omega>' = \<omega> i)"
          by (simp add: Ball_def M.emeasure_eq_measure measure_not_0)
        also have "\<dots> \<le> \<P>(\<omega>' in M. X n \<omega>' = \<omega> n)"
          by (auto intro!: M.finite_measure_mono)
        finally have *: "0 < \<P>(\<omega>' in M. X n \<omega>' = \<omega> n)"
          by simp
  
        have "emeasure (K' (Some (\<omega> n))) {Some (\<omega> (Suc n))} = \<P>(\<omega>' in M. X (Suc n) \<omega>' = \<omega> (Suc n) \<bar> X n \<omega>' = \<omega> n )"
          using * by (rule emeasure_K_S)
        also have "\<dots> = \<P>(\<omega>' in M. X (Suc n) \<omega>' = \<omega> (Suc n) \<bar> \<forall>i\<le>n. X i \<omega>' = \<omega> i )"
          using ** by (auto intro!: MC[symmetric])
        finally show "emeasure (K' (Some (\<omega> n))) {Some (\<omega> (Suc n))} =
            \<P>(\<omega>' in M. X (Suc n) \<omega>' = \<omega> (Suc n) \<bar> \<forall>i\<le>n. X i \<omega>' = \<omega> i )"
          using \<omega> by (simp add: Ball_def)
      qed
      also have "\<dots> = emeasure M {\<omega>' \<in> space M. \<forall>i\<le>Suc n. X i \<omega>' = \<omega> i}"
        using le measure_nonneg[of M "{\<omega>' \<in> space M. X (Suc n) \<omega>' = \<omega> (Suc n) \<and> (\<forall>i\<le>n. X i \<omega>' = \<omega> i)}"]
        by (simp add: M.emeasure_eq_measure emeasure_distr_Collect cond_prob_def Ball_def le_Suc_eq all_conj_distrib conj_commute)
      finally show ?case .
    qed
    with \<omega> show ?thesis
      by simp
  qed
  finally show "emeasure ?L {\<omega> \<in> space ?L. \<forall>i\<le>n. \<omega> i = \<omega>' i} = emeasure ?R {\<omega> \<in> space ?R. \<forall>i\<le>n. \<omega> i = \<omega>' i}"
    by (subst emeasure_distr_Collect') (auto simp: less_Suc_eq_le)
qed

end

section {* Discrete Markov Kernel Constructs a DTMC *} text_raw {* \label{sec:kerneldtmc} *}

context Discrete_Markov_Kernel
begin

lemma split_at:
  assumes [simp]: "s \<in> S" and 2: "\<P>(\<omega>' in paths s. \<omega>' n = \<omega> n) \<noteq> 0"
  shows "\<P>(\<omega>' in paths s. \<omega>' (Suc n) = s' \<bar> \<omega>' n = \<omega> n ) = \<P>(\<omega>' in paths (\<omega> n). \<omega>' 0 = s')"
proof -
  have "\<P>(\<omega>' in paths s. \<omega>' (Suc n) = s' \<and> \<omega>' n = \<omega> n) =
      (\<integral>\<omega>'. \<P>(\<omega>' in paths (\<omega> n). \<omega>' 0 = s') * indicator {\<omega>'\<in>space (paths s). \<omega>' n = \<omega> n} \<omega>' \<partial>paths s)"
    by (subst prob_split_Collect[where i="Suc n"])
       (auto intro!: integral_cong split: split_indicator
             simp: comb_seq_less less_Suc_eq_le comb_seq_id
             simp del: integral_cmult)
  also have "\<dots> = \<P>(\<omega>' in paths (\<omega> n). \<omega>' 0 = s') * \<P>(\<omega>' in paths s. \<omega>' n = \<omega> n)"
    by (simp add: emeasure_eq_measure)
  finally show ?thesis
    using 2 unfolding cond_prob_def by simp
qed

lemma is_DTMC:
  assumes [simp]: "s \<in> S"
  shows "Discrete_Time_Markov_Chain (paths s) S (\<lambda>n \<omega>. \<omega> n)"
proof
  show "countable S"
    by (rule countable_space)
  show "\<And>t. (\<lambda>\<omega>. \<omega> t) \<in> measurable (paths s) (count_space S)"
    by simp
next
  fix n :: nat and \<omega> s'
  assume 1: "\<P>(\<omega>' in paths s. \<forall>t\<le>n. \<omega>' t = \<omega> t ) \<noteq> 0"
  then have 2: "\<P>(\<omega>' in paths s. \<omega>' n = \<omega> n ) \<noteq> 0"
    by (rule finite_measure_not_0_imp_not_0) auto
  have "\<P>(\<omega>' in paths s. \<omega>' (Suc n) = s' \<and> (\<forall>t\<le>n. \<omega>' t = \<omega> t)) =
      (\<integral>\<omega>'. \<P>(\<omega>' in paths (\<omega> n). \<omega>' 0 = s') * indicator {\<omega>'\<in>space (paths s). \<forall>t\<le>n. \<omega>' t = \<omega> t} \<omega>' \<partial>paths s)"
    by (subst prob_split_Collect[where i="Suc n"])
       (auto intro!: integral_cong split: split_indicator
             simp: comb_seq_less less_Suc_eq_le comb_seq_id prob_eq_0
             simp del: integral_cmult)
  also have "\<dots> = \<P>(\<omega>' in paths (\<omega> n). \<omega>' 0 = s') * \<P>(\<omega>' in paths s. \<forall>t\<le>n. \<omega>' t = \<omega> t)"
    by (simp add: emeasure_eq_measure)
  finally have "\<P>(\<omega>' in paths s. \<omega>' (Suc n) = s' \<bar> (\<forall>t\<le>n. \<omega>' t = \<omega> t) ) = \<P>(\<omega>' in paths (\<omega> n). \<omega>' 0 = s')"
    using 1 unfolding cond_prob_def by simp
  then show "\<P>(\<omega>' in paths s. \<omega>' (Suc n) = s' \<bar> (\<forall>t\<le>n. \<omega>' t = \<omega> t) ) = \<P>(\<omega>' in paths s. \<omega>' (Suc n) = s' \<bar> \<omega>' n = \<omega> n )"
    using 2 split_at[OF `s\<in>S`, of n \<omega> s'] by simp
next
  fix n m a b
  assume 1: "\<P>(\<omega>' in paths s. \<omega>' m = a ) \<noteq> 0"
  assume 2: "\<P>(\<omega>' in paths s. \<omega>' n = a ) \<noteq> 0"
  show "\<P>(\<omega>' in paths s. \<omega>' (Suc n) = b \<bar> \<omega>' n = a ) = \<P>(\<omega>' in paths s. \<omega>' (Suc m) = b \<bar> \<omega>' m = a )"
    using split_at[OF `s\<in>S` 1] split_at[OF `s\<in>S` 2] by simp
qed

end

end

