theory Example_A
  imports Classifying_Markov_Chains
begin

section {* Example A *} text_raw {* \label{ex:A} *}

text {*

We formalize the following Markov chain:

\begin{center}
\begin{tikzpicture}[thick]

  \path [fill, color = gray!30] (0, 0) circle(0.6) ;

  \path [fill, color = gray!30] (1, 1) circle(0.6) ;

  \path [fill, color = gray!30] (4.5, 0.66) ellipse(2 and 1.9) ;

  \node (bot)  at (-1, 0) {} ;

  \node[draw,circle] (A)  at (0, 0) {$A$} ;

  \node[draw,circle] (B)  at (1, 1) {$B$} ;

  \node[draw,circle] (C1) at (3, 0) {$C_1$} ;

  \node[draw,circle] (C2) at (6, 0) {$C_2$} ;

  \node[draw,circle] (C3) at (4.5, 2) {$C_3$} ;

  \path[->, >=latex]
    (bot) edge (A)
    (A)   edge                node [above] {$\frac{1}{2}$} (B)
          edge                node [below] {$\frac{1}{2}$} (C1)
    (B)   edge [loop above]   node [left]  {$\frac{1}{2}$} (B)
          edge [out = 0]      node [above] {$\frac{1}{2}$} (C1)
    (C1)  edge [loop above]   node [above] {$\frac{1}{3}$} (C1)
          edge [bend left=15] node [above] {$\frac{1}{3}$} (C2)
          edge [bend left=15] node [above] {$\frac{1}{3}$} (C3)
    (C2)  edge [loop right]   node [above] {$\frac{1}{3}$} (C2)
          edge [bend left=15] node [above] {$\frac{1}{3}$} (C1)
          edge [bend left=15] node [above] {$\frac{1}{3}$} (C3)
    (C3)  edge [loop right]   node [above] {$\frac{1}{2}$} (C3)
          edge [bend left=15] node [above] {$\frac{1}{4}$} (C1)
          edge [bend left=15] node [above] {$\frac{1}{4}$} (C2) ;

\end{tikzpicture}
\end{center}

First we define the state space as its own type:

*}

datatype state = A | B | C1 | C2 | C3

text {* Now the state space is @{text "UNIV :: state set"} *}

lemma UNIV_state: "UNIV = {A, B, C1, C2, C3}"
  using state.nchotomy by auto

instance state :: finite
  by default (simp add: UNIV_state)

text {* The transition function @{text tau} is easily defined using the case statement, this allows
us to give a sparse specification as all @{text 0} cases are collected at the end. *}

definition tau :: "state \<Rightarrow> state \<Rightarrow> real" where
  "tau s t = (case (s, t) of
      (A,  B)  \<Rightarrow> 1 / 2 | (A,  C1) \<Rightarrow> 1 / 2
    | (B,  B)  \<Rightarrow> 1 / 2 | (B,  C1) \<Rightarrow> 1 / 2
    | (C1, C1) \<Rightarrow> 1 / 3 | (C1, C2) \<Rightarrow> 1 / 3 | (C1, C3) \<Rightarrow> 1 / 3
    | (C2, C1) \<Rightarrow> 1 / 3 | (C2, C2) \<Rightarrow> 1 / 3 | (C2, C3) \<Rightarrow> 1 / 3
    | (C3, C1) \<Rightarrow> 1 / 4 | (C3, C2) \<Rightarrow> 1 / 4 | (C3, C3) \<Rightarrow> 1 / 2
    | _ \<Rightarrow> 0)"

text {* We use the @{text finite_pmf}-locale which introduces the point measure @{text tau.M}, and
  provides us with the necessary simplifier setup. *}

interpretation tau: finite_pmf UNIV "tau s" for s
  by unfold_locales (simp_all add: tau_def UNIV_state split: state.split)

interpretation A: Discrete_Markov_Kernel UNIV tau.M
  by default simp_all

subsection {* The essential classs @{term "{C1, C2, C3}"} *}

lemma A_E_eq:
  "A.E x = (case x of A \<Rightarrow> {B, C1} | B \<Rightarrow> {B, C1} | _ \<Rightarrow> {C1, C2, C3})"
  unfolding A.E_def using state.nchotomy by (auto simp: tau_def split: state.split)

lemma A_essential: "A.essential_class {C1, C2, C3}"
  by (rule A.essential_classI2) (auto intro!: A.accessible_single simp: A_E_eq)

lemma A_aperiodic: "A.aperiodic {C1, C2, C3}"
  unfolding A.aperiodic_def
proof safe
  have eq: "\<And>x'. (if x' = C1 then 1 else 0) = indicator {C1} x'" by auto

  show "{C1, C2, C3} \<in> UNIV // A.communicating"
    using A_essential by (simp add: A.essential_class_def)
  then have "A.period {C1, C2, C3} = Gcd (A.period_set C1)"
    by (rule A.period_eq) simp
  also have "\<dots> = 1"
    by (rule Gcd_eq_one) (simp add: A.period_set_def A.p_Suc A.p_0 eq tau_def)
  finally show "A.period {C1, C2, C3} = 1" .
qed

subsection {* The stationary distribution @{text n} *}

text {* Similar to @{text tau} we introduce @{text n} using the @{text finite_pmf}-locale. *}

definition n :: "state \<Rightarrow> real" where
  "n = (\<lambda>C1 \<Rightarrow> 0.3 | C2 \<Rightarrow> 0.3 | C3 \<Rightarrow> 0.4 | _ \<Rightarrow> 0)"

interpretation n: finite_pmf UNIV n
  by unfold_locales (simp_all add: n_def UNIV_state split: state.split)

lemma stationary_distribution_N: "A.stationary_distribution n.M"
  unfolding A.stationary_distribution_def
proof (rule measure_eqI_finite[where A=UNIV])
  fix a
  have sum_eq: "\<And>x. (\<Sum>y\<in>UNIV. tau y x * n y) = n x"
    by (simp add: UNIV_state n_def tau_def split: state.split)
  show "emeasure n.M {a} = emeasure (point_measure UNIV (\<lambda>x. \<integral>\<^sup>+ s. emeasure (tau.M s) {x} \<partial>n.M)) {a}"
    by (subst emeasure_point_measure_finite) (simp_all add: positive_integral_positive sum_eq)
qed (simp_all add: sets_point_measure)

lemma exclusive_N: "measure n.M (UNIV - {C1, C2, C3}) = 0"
  by (subst n.measure_M) (auto simp add: n_def UNIV_state setsum_diff)

lemma n_is_limit:
  assumes x: "x \<in> {C1, C2, C3}" and y: "y \<in> {C1, C2, C3}"
  shows "(A.p x y) ----> n y"
proof -
  have "(A.p x y) ----> measure n.M {y}"
    using A_aperiodic A_essential stationary_distribution_N n.prob_space_M exclusive_N x y
    by (rule A.stationary_distribution_imp_p_limit)
  then show ?thesis
    by simp
qed

lemma C_is_pos_recurrent: "x \<in> {C1, C2, C3} \<Longrightarrow> A.pos_recurrent x"
  using A.stationary_distributionD(1)[OF A_essential stationary_distribution_N n.prob_space_M exclusive_N]
  by auto

lemma C_recurrence_time: 
  assumes x: "x \<in> {C1, C2, C3}"
  shows "(\<integral>\<^sup>+\<omega>. A.t x \<omega> \<partial>A.paths x) = 1 / n x"
proof -
  from A.stationary_distributionD(2)[OF A_essential stationary_distribution_N n.prob_space_M exclusive_N]
  have "A.stat {C1, C2, C3} = n.M" ..
  with x have "1 / n x = 1 / emeasure (A.stat {C1, C2, C3}) {x}"
    by (simp add: divide_inverse one_ereal_def) (auto simp add: n_def)
  also have "\<dots> = (\<integral>\<^sup>+\<omega>. A.t x \<omega> \<partial>A.paths x)"
    unfolding A.stat_def using x
    by (subst emeasure_point_measure_finite)
       (simp_all add: inverse_nonneg_iff positive_integral_positive ereal_one_divide_invese)
  finally show ?thesis ..
qed

end

