theory Example_B
  imports Classifying_Markov_Chains
begin

lemma one_divide_ereal: "x \<noteq> 0 \<Longrightarrow> 1 / ereal x = ereal (1 / x)"
  by (simp add: divide_ereal_def divide_inverse)

section {* Example B *} text_raw {* \label{ex:B} *}

text {*

We now formalize the following Markov chain:

\begin{center}
\begin{tikzpicture}[thick]

  \begin{scope} [rotate = 45]
    \path [fill, color = gray!30] (7.5, -6) ellipse(3 and 1) ;
  \end{scope}
  
  \node (bot2)  at (7, -0.5) {} ;
  \node[draw, circle] (1) at ( 8, -0.5) {$0$} ;
  \node[draw, circle] (2) at ( 9,  0.5) {$1$} ;
  \node[draw, circle] (3) at (10,  1.5) {$2$} ;
  \node (inft) at (10.7, 2.6) {} ;
  \node (infb) at (11,   2) {} ;

  \node (inf1) at (10.5, 2) {} ;
  \node (inf2) at (11.5, 3) {} ;
  
  \path[->, >=latex]
    (bot2) edge (1)
    (1)    edge [loop below]   node [right] {$\frac{2}{3}$} (1)
           edge [bend left=30] node [above] {$\frac{1}{3}$} (2)
    (2)    edge [bend left=30] node [below] {$\frac{2}{3}$} (1)
           edge [bend left=30] node [above] {$\frac{1}{3}$} (3)
    (3)    edge [bend left=30] node [below] {$\frac{2}{3}$} (2)
           edge [bend left=30] node [above] {} (inft)
    (infb)  edge [bend left=30] node [above] {} (3) ;

  \path (inf1) edge [loosely dotted] (inf2) ;

\end{tikzpicture}
\end{center}

As state space we have the set of natural numbers, the transition function @{term tau} has three
cases:

*}

definition tau :: "nat \<Rightarrow> nat \<Rightarrow> real" where
  "tau x y = (if x + 1 = y then 1 / 3
         else if x - 1 = y then 2 / 3
         else 0)"

text {* For the special case when @{term "x = (0::real)"} we have @{term "x - 1 = (0::real)"} and hence
@{term "tau 0 0 = 2 / 3"}. *}

text {* We pack this transition function into a discrete Markov kernel, by showing that it forms a
  probability space. *}

subsection {* Discrete Markov Kernel *}

definition "K x = point_measure UNIV (tau x)"

lemma sets_K[simp]: "sets (K x) = UNIV" and space_K[simp]: "space (K x) = UNIV"
  by (simp_all add: K_def sets_point_measure space_point_measure)

lemma AE_K[simp]: "(AE x in K y. P x) \<longleftrightarrow> P (y + 1) \<and> P (y - 1)"
  by (subst K_def) (auto simp add: tau_def AE_point_measure)

lemma emeasure_K[simp]:
  "emeasure (K x) X = 1/3 * indicator X (x + 1) + 2/3 * indicator X (x - 1)"
proof -
  have eq: "{a \<in> X. 0 < tau x a} = (if x - 1 \<in> X \<and> x + 1 \<in> X then {x - 1, x + 1}
                                   else if x - 1 \<in> X then {x - 1}
                                   else if x + 1 \<in> X then {x + 1}
                                   else {})"
    by (auto simp: tau_def)
  show ?thesis
    unfolding K_def
    apply (subst emeasure_point_measure)
    apply (auto simp add: eq)
    apply (simp_all add: tau_def one_divide_ereal)
    done
qed

interpretation K: prob_space "K x" for x
  by rule (simp add: one_divide_ereal one_ereal_def[symmetric])

text {* We call the locale of the Markov chain @{text B}, hence all constants and theorems
  from this Markov chain get a @{text B} prefix. *}

interpretation B: Discrete_Markov_Kernel UNIV K
  by default simp_all

subsection {* Enabled, accessible and communicating states *}

text {* For each step the predecessor and the successor are enabled (in the @{term 0} case, the
predecessor is again @{term 0}. Hence every state is accessible from everywhere and every states is
communicating with each other state. Finally we know that the state space is an essential class. *}

lemma B_E_eq: "B.E x = {x - 1, x + 1}"
  unfolding B.E_def by (auto simp: one_divide_ereal)

lemma B_E_Suc: "Suc x \<in> B.E x" "x \<in> B.E (Suc x)"
  unfolding B_E_eq by auto

lemma B_accessible[intro]: "B.accessible i j"
proof (cases i j rule: linorder_le_cases)
  assume "i \<le> j" then show ?thesis
    by (induct rule: inc_induct) (auto intro: B.accessible.intros B.accessible_step_rev B_E_Suc)
next
  assume "j \<le> i" then show ?thesis
    by (induct rule: dec_induct) (auto intro: B.accessible.intros B.accessible_step_rev B_E_Suc)
qed

lemma B_communicating[intro]: "(i, j) \<in> B.communicating"
  by (simp add: B.communicating_def B_accessible)

lemma B_essential: "B.essential_class UNIV"
  by (rule B.essential_classI2) auto

subsection {* B is aperiodic *}

lemma B_aperiodic: "B.aperiodic UNIV"
  unfolding B.aperiodic_def
proof safe
  have eq: "\<And>x'. (if x' = 0 then 1 else 0) = indicator {0} x'" by auto

  show "UNIV \<in> UNIV // B.communicating"
    using B_essential by (simp add: B.essential_class_def)
  then have "B.period UNIV = Gcd (B.period_set 0)"
    by (rule B.period_eq) simp
  also have "\<dots> = 1"
    by (rule Gcd_eq_one) (simp add: B.period_set_def B.p_Suc B.p_0 eq)
  finally show "B.period UNIV = 1" .
qed

subsection {* The stationary distribution @{text N} *}

definition "N = point_measure UNIV (\<lambda>i. ereal (1 / 2) ^ Suc i)"

lemma borel_measurable_N[simp]: "f \<in> borel_measurable N"
  by (simp_all add: N_def sets_point_measure space_point_measure)

lemma sets_N[simp]: "sets N = UNIV" and space_N[simp]: "space N = UNIV"
  by (simp_all add: N_def sets_point_measure space_point_measure)

lemma emeasure_N_finite[simp]:
    "finite A \<Longrightarrow> emeasure N A = (\<Sum>i\<in>A. ereal (1 / 2)^Suc i)"
  unfolding N_def by (subst emeasure_point_measure_finite2) simp_all

lemma prob_space_N: "prob_space N"
proof
  have "(\<lambda>i. (1 / 2) ^ i :: real) sums 2"
    using geometric_sums[of "1 / 2 :: real"] by simp
  from sums_divide[OF this, of 2]
  have *: "(\<lambda>i. (1 / 2) ^ i / 2 :: real) sums 1"
    by simp
  
  show "emeasure N (space N) = 1"
    by (subst emeasure_countable_singleton)
       (auto simp add: positive_intergal_count_space_nat one_divide_ereal one_ereal_def *
             intro!: sums_suminf_ereal)
qed

lemma stationary_distribution_N: "B.stationary_distribution N"
  unfolding B.stationary_distribution_def
proof (rule measure_eqI_countable[where A=UNIV])
  fix a
  have eq: "\<And>s. 1 / ereal 3 * indicator {a} (Suc s) + ereal (2 / 3) * indicator {a} (s - Suc 0) =
     (if a = 0 then ereal (2 / 3) * indicator {0, 1} s else 1 / ereal 3 * indicator {a - 1} s + ereal (2 / 3) * indicator {a + 1} s)"
     by (auto split: split_indicator)
  show "emeasure N {a} = emeasure (point_measure UNIV (\<lambda>x. \<integral>\<^sup>+ s. emeasure (K s) {x} \<partial>N)) {a}"
    apply (subst emeasure_point_measure_finite2)
    apply (simp_all add: positive_integral_positive eq del: ereal_times_divide_eq_left)
    apply (cases "a = 0")
    apply (auto simp: positive_integral_cmult_indicator emeasure_N_finite one_divide_ereal)
    apply (subst positive_integral_add)
    apply (simp_all del: ereal_times_divide_eq_left)
    apply (subst (1 2) positive_integral_cmult_indicator)
    apply (simp_all add: one_divide_ereal)
    done
qed (simp_all add: sets_point_measure)

lemma exclusive_N: "measure N (UNIV - UNIV) = 0"
  by simp

subsection {* Limit behavior and recurrence times *}

lemma limit: "(B.p i j) ----> (1/2)^Suc j"
proof -
  have "(B.p i j) ----> measure N {j}"
    using B_aperiodic B_essential stationary_distribution_N prob_space_N exclusive_N UNIV_I UNIV_I
    by (rule B.stationary_distribution_imp_p_limit)
  then show ?thesis
    by (simp add: measure_def del: power_Suc)
qed

lemma pos_recurrent: "B.pos_recurrent i"
  using B.stationary_distributionD(1)[OF B_essential stationary_distribution_N prob_space_N exclusive_N]
  by auto

lemma recurrence_time: "(\<integral>\<^sup>+\<omega>. B.t i \<omega> \<partial>B.paths i) = 2^Suc i"
proof -
  from B.stationary_distributionD(2)[OF B_essential stationary_distribution_N prob_space_N exclusive_N]
  have "B.stat UNIV = N" ..
  then have "ereal 2^Suc i = 1 / emeasure (B.stat UNIV) {i}"
    by (simp add: one_divide_ereal field_simps power_mult_distrib[symmetric])
  also have "\<dots> = (\<integral>\<^sup>+\<omega>. B.t i \<omega> \<partial>B.paths i)"
    unfolding B.stat_def
    by (subst emeasure_point_measure_finite2)
       (simp_all add: inverse_nonneg_iff positive_integral_positive ereal_one_divide_invese)
  finally show ?thesis
    by (simp add: ereal_power)
qed

end

