theory Classifying_Markov_Chains
  imports Discrete_Markov_Kernel Constructing_Markov_Chain
begin

lemma int_cases': "(\<And>n. x = int n \<Longrightarrow> P) \<Longrightarrow> (\<And>n. x = - int n \<Longrightarrow> P) \<Longrightarrow> P"
  by (metis int_cases)

lemma nat_abs_int_diff: "nat \<bar>int a - int b\<bar> = (if a \<le> b then b - a else a - b)"
  by auto

lemma nat_int_add: "nat (int a + int b) = a + b"
  by auto

context Discrete_Markov_Kernel
begin

section {* Analyzing Markov Chains *} text_raw {* \label{sec:analyzing} *}

subsection {* Accessible *} text_raw {* \label{sec:accessible} *}

inductive accessible for s where
  start: "s \<in> S \<Longrightarrow> accessible s s"
| step: "accessible s t \<Longrightarrow> t' \<in> E t \<Longrightarrow> accessible s t'"

lemma accessible_in_S1: assumes "accessible s t" shows "s \<in> S"
  using assms by induct auto

lemma accessible_in_S2: assumes "accessible s t" shows "t \<in> S"
  using assms by induct auto

lemma accessible_trans:
  assumes "accessible s t" "accessible t t'" shows "accessible s t'"
  using assms(2,1) by induct (auto intro: accessible.step)

lemma accessible_single:
  "s \<in> S \<Longrightarrow> t \<in> E s \<Longrightarrow> accessible s t"
  by (blast intro: accessible.intros)

lemma accessible_induct_trans[consumes 1, case_names start step trans]:
  assumes t: "accessible s t"
  assumes 1: "\<And>s. s \<in> S \<Longrightarrow> P s s"
  assumes 2: "\<And>s t. s \<in> S \<Longrightarrow> t \<in> E s \<Longrightarrow> P s t"
  assumes 3: "\<And>t t' s. accessible s t' \<Longrightarrow> P s t' \<Longrightarrow> accessible t' t \<Longrightarrow> P t' t \<Longrightarrow> P s t"
  shows "P s t"
  using t by induct (blast intro: 1 2 3 accessible.intros accessible_in_S2)+

lemma accessible_step_rev:
  assumes "accessible s t" "s \<in> E s'" "s' \<in> S" shows "accessible s' t"
  using assms accessible_trans[of s s' t] by (auto intro: accessible_single accessible_trans)

lemma accessible_rev:
  assumes t: "accessible s t"
  obtains (start) "t = s" " s \<in> S" | (step) s' where "accessible s' t" "s' \<in> E s" "s \<in> S"
  using t by induct (auto intro: accessible.intros)

lemma accessible_induct_rev[consumes 1, case_names start step]:
  assumes t: "accessible s t"
  assumes 1: "P t"
  assumes 2: "\<And>t' s. t' \<in> E s \<Longrightarrow> accessible t' t \<Longrightarrow> P t' \<Longrightarrow> P s"
  shows "P s"
  using t
proof cases
  case (step s')
  then have "accessible s' t" "P s'"
    by (auto intro: accessible_single accessible_in_S2 2 1 accessible.intros)
  with step(1) show ?thesis
    by (induct rule: accessible_induct_trans) (auto intro: 1 2 accessible_trans)
qed (insert 1, simp)

lemma AE_accessible:
  assumes x: "x \<in> S"
  shows "AE \<omega> in paths x. \<forall>m. accessible x (nat_case x \<omega> m)"
  using AE_all_enabled[OF x] AE_space
proof (eventually_elim, safe)
  fix \<omega> m assume \<omega>: "\<forall>i. \<omega> i \<in> E (nat_case x \<omega> i)" "\<omega> \<in> space (paths x)"
  show "accessible x (nat_case x \<omega> m)"
  proof (induct m)
    case 0 then show ?case
      using \<omega>(2) by (auto intro!: accessible.start split: nat.split simp: space_PiM `x\<in>S`)
  next
    case (Suc m') with \<omega>(1)[THEN spec, of m'] show ?case
      by (auto intro: accessible.step)
  qed
qed

subsection {* Communicating *} text_raw {* \label{sec:communicating} *}

definition communicating :: "('s \<times> 's) set" where
  "communicating = {(s, t). accessible s t \<and> accessible t s}"

lemma irreducibleD:
  "C \<in> S // communicating \<Longrightarrow> a \<in> C \<Longrightarrow> b \<in> C \<Longrightarrow> (a, b) \<in> communicating"
  by (auto elim!: quotientE simp: communicating_def intro: accessible_trans)

lemma irreducibleD2:
  "C \<in> S // communicating \<Longrightarrow> a \<in> C \<Longrightarrow> (a, b) \<in> communicating \<Longrightarrow> b \<in> C"
  by (auto elim!: quotientE simp: communicating_def intro: accessible_trans)

lemma communicating_in_S:
  assumes "(x, y) \<in> communicating" shows "x \<in> S" "y \<in> S"
  using assms by (auto simp add: communicating_def dest: accessible_in_S1 accessible_in_S2)

lemma C_subset_S: "C \<in> S // communicating \<Longrightarrow> C \<subseteq> S"
  by (auto elim!: quotientE simp: communicating_in_S)

lemma equiv_communicating: "equiv S communicating"
  by (auto simp: equiv_def sym_def communicating_def refl_on_def
      accessible_in_S1 accessible_in_S2 accessible.start trans_def
      intro: accessible_trans)

lemma communicatingD1:
  "C \<in> S // communicating \<Longrightarrow> (a, b) \<in> communicating \<Longrightarrow> a \<in> C \<Longrightarrow> b \<in> C"
  by (auto elim!: quotientE) (auto simp add: communicating_def intro: accessible_trans)

lemma communicatingD2:
  "C \<in> S // communicating \<Longrightarrow> (a, b) \<in> communicating \<Longrightarrow> b \<in> C \<Longrightarrow> a \<in> C"
  by (auto elim!: quotientE) (auto simp add: communicating_def intro: accessible_trans)

lemma irreducible_MC: "(\<And>x y. x \<in> S \<Longrightarrow> y \<in> S \<Longrightarrow> accessible x y) \<Longrightarrow> S \<in> S // communicating"
  by (auto simp: quotient_def Image_def communicating_def accessible_in_S1 intro!: bexI[of _ s0] s0)

lemma not_empty_irreducible:
  "C \<in> S // communicating \<Longrightarrow> C \<noteq> {}"
  apply (auto simp: quotient_def Image_def)
  apply (erule_tac x=x in allE)
  apply (simp add: communicating_def accessible.start)
  done

subsection {* Time-Bounded Quantities *} text_raw {* \label{sec:timebound} *}

definition "p x y n = \<P>(\<omega> in paths x. nat_case x \<omega> n = y)"

lemma p_nonneg: "0 \<le> p x y n"
  by (simp add: p_def measure_nonneg)

lemma p_le_1: "p x y n \<le> 1"
  by (simp add: p_def)

lemma p_x_x_0[simp]: "p x x 0 = 1"
  by (simp add: p_def prob_space del: space_paths)

lemma p_0: "p x y 0 = (if x = y then 1 else 0)"
  by (simp add: p_def prob_space del: space_paths)

lemma p_Suc:
  assumes "x \<in> S"
  shows "p x y (Suc n) = (\<integral>x'. p x' y n \<partial>K x)"
  unfolding p_def by (subst prob_iterate_Collect[OF `x\<in>S`]) simp_all

lemma accessibleI_pos:
  "x \<in> S \<Longrightarrow> y \<in> S \<Longrightarrow> 0 < p x y n \<Longrightarrow> accessible x y"
proof (induct n arbitrary: x)
  case 0 then show ?case by (simp add: p_0 accessible.start split: split_if_asm)
next
  case (Suc n)
  then have less: "0 < (\<integral>x'. p x' y n \<partial>K x)"
    by (simp add: p_Suc)
  have "\<exists>x'\<in>E x. 0 < p x' y n"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then have "AE x' in K x. p x' y n = 0"
      using `x \<in> S` by (simp add: AE_K_iff p_nonneg less_le)
    then have "(\<integral>x'. p x' y n \<partial>K x) = (\<integral>x'. 0 \<partial>K x)"
      using `x \<in> S` by (intro integral_cong_AE) simp
    with less show False by simp
  qed
  with Suc show ?case
    by (auto intro: accessible_step_rev)
qed

lemma prob_reachable_le:
  assumes [simp]: "x \<in> S" "y \<in> S" "w \<in> S" and "m \<le> n"
  shows "p x y m * p y w (n - m) \<le> p x w n"
proof -
  have "emeasure (paths x) {\<omega>\<in>space (paths x). nat_case x \<omega> n = w} =
      (\<integral>\<^sup>+\<omega>. emeasure (paths (nat_case x \<omega> m)) {\<omega>'\<in>space (paths (nat_case x \<omega> m)). comb_seq (Suc m) (nat_case x \<omega>) \<omega>' n = w} \<partial>paths x)"
    by (subst emeasure_split_Collect[where i=m])  (simp_all add: nat_case_comb_seq')
  also have "\<dots> \<ge> (\<integral>\<^sup>+\<omega>. emeasure (paths y) {\<omega>'\<in>space (paths y). nat_case y \<omega>' (n - m) = w} *
      indicator {\<omega>\<in>space (paths x). nat_case x \<omega> m = y } \<omega> \<partial>paths x)" (is "_ \<ge> ?rhs")
    using `m \<le> n`
    apply (auto intro!: positive_integral_mono emeasure_mono simp del: space_paths split: split_indicator) 
    apply (simp add: comb_seq_def less_Suc_eq_le minus_nat.diff_Suc split: nat.split split_indicator)
    unfolding space_paths
    apply (intro predE pred_intros_logic pred_eq_const1)
    apply (rule measurable_component_singleton'[where I=UNIV and M="\<lambda>x. count_space UNIV"])
    apply (rule measurable_comb_seq')
    apply simp_all
    done
  also (xtrans) have "?rhs = emeasure (paths y) {\<omega>'\<in>space (paths y). nat_case y \<omega>' (n - m) = w} *
      emeasure (paths x) {\<omega>\<in>space (paths x). nat_case x \<omega> m = y }"
    by (subst positive_integral_cmult_indicator) auto
  finally show ?thesis
    by (simp add: emeasure_eq_measure ac_simps p_def)
qed

lemma accessibleD_pos:
  assumes "accessible x y"
  shows "\<exists>n. 0 < p x y n"
using assms proof induction
  case start with prob_space[of x] show ?case
    by (auto intro!: exI[of _ 0])
next
  case (step w y)
  then obtain n where "0 < p x w n" 
    and "0 < measure (K w) {y}"
    and [simp]: "x \<in> S" "y \<in> S" "w \<in> S"
    by (auto simp: E_def K.emeasure_eq_measure measure_nonneg less_le accessible_in_S1 accessible_in_S2)
  then have "0 < p x w n * measure (K w) {y}"
    by (intro mult_pos_pos)
  also have "\<dots> = p x w n * p w y (Suc 0)"
    using prob_paths_0[of w "\<lambda>s. s = y"] by (simp add: p_def cong: conj_cong)
  also have "\<dots> \<le> p x y (Suc n)"
    using prob_reachable_le[of x w y n "Suc n"] by simp
  finally show ?case ..
qed

subsubsection {* Accessibility as probability on traces *} text_raw {* \label{sec:accessibleiff} *}

lemma accessible_iff: "accessible x y \<longleftrightarrow> x \<in> S \<and> y \<in> S \<and> (\<exists>n. 0 < p x y n)"
  by (blast intro: accessibleD_pos accessibleI_pos accessible_in_S1 accessible_in_S2)

lemma communicating_iff: "(x, y) \<in> communicating \<longleftrightarrow> x \<in> S \<and> y \<in> S \<and> (\<exists>n. 0 < p x y n) \<and> (\<exists>n. 0 < p y x n)"
  by (auto simp add: accessible_iff communicating_def)

definition "u x y n = \<P>(\<omega> in paths x. (\<forall>i<n. \<omega> i \<noteq> y) \<and> \<omega> n = y)"

lemma shows u_nonneg: "0 \<le> u x y n" and u_le_1: "u x y n \<le> 1"
  by (simp_all add: u_def measure_nonneg)

lemma u_le_p: "u x y n \<le> p x y (Suc n)"
  unfolding u_def p_def by (auto intro!: finite_measure_mono)

lemma p_eq_setsum_p_u:
  assumes x[simp]: "x \<in> S" shows "p x y (Suc n) = (\<Sum>i\<le>n. p y y (n - i) * u x y i)"
proof -
  let ?S = "\<lambda>i. {\<omega> \<in> space (paths x). ((\<forall>j<i. \<omega> j \<noteq> y) \<and> \<omega> i = y) \<and> \<omega> n = y}"
  have "p x y (Suc n) = measure (paths x) (\<Union>i\<le>n. ?S i)"
    apply (simp add: p_def)
    apply (rule arg_cong2[where f=measure])
    apply auto
    apply (rule_tac x="LEAST i. x i = x n" in bexI)
    apply (auto dest: not_less_Least intro: Least_le LeastI)
    done
  also have "\<dots> = (\<Sum>i\<le>n. measure (paths x) (?S i))"
    by (rule finite_measure_finite_Union) (auto simp: disjoint_family_on_def neq_iff)
  also have "\<dots> = (\<Sum>i\<le>n. p y y (n - i) * u x y i)"
  proof (intro setsum_cong refl)
    fix i assume i: "i \<in> {.. n}"
    have "measure (paths x) (?S i) =
        (\<integral>\<omega>. \<P>(\<omega>' in paths (nat_case x \<omega> (Suc i)).
          ((\<forall>j<i. comb_seq (Suc i) \<omega> \<omega>' j \<noteq> y) \<and> comb_seq (Suc i) \<omega> \<omega>' i = y) \<and> comb_seq (Suc i) \<omega> \<omega>' n = y) \<partial>paths x)"
      by (rule prob_split_Collect) simp_all
    also have "\<dots> = (\<integral>\<omega>. p y y (n - i) *  indicator {\<omega>\<in>space (paths x). (\<forall>j<i. \<omega> j \<noteq> y) \<and> \<omega> i = y} \<omega> \<partial>paths x)"
      using i
      by (intro integral_cong)
         (auto simp: comb_seq_less comb_seq_Suc2 prob_eq_0 p_def split: split_indicator nat.split)
    also have "\<dots> = p y y (n - i) * u x y i"
      by (simp add: emeasure_eq_measure u_def)
    finally show "measure (paths x) (?S i) = p y y (n - i) * u x y i" .
  qed
  finally show ?thesis .
qed

definition "f x y n = \<P>(\<omega> in paths x. (\<forall>i<n. nat_case x \<omega> i \<noteq> y) \<and> nat_case x \<omega> n = y)"

lemma shows f_nonneg: "0 \<le> f x y n" and f_le_1: "f x y n \<le> 1"
  by (simp_all add: f_def measure_nonneg)

lemma f_le_p: "f x y n \<le> p x y n"
  unfolding f_def p_def by (auto intro!: finite_measure_mono)

lemma f_Suc: "x \<noteq> y \<Longrightarrow> f x y (Suc n) = u x y n"
  by (simp add: u_def f_def all_less_Suc_split)

lemma f_Suc_eq: "f x x (Suc n) = 0"
  by (simp add: f_def all_less_Suc_split)

lemma f_0: "f x y 0 = (if x = y then 1 else 0)"
  using prob_space by (simp add: f_def)

lemma atMost_Suc_insert_0: "{.. Suc n} = insert 0 (Suc ` {.. n})"
  apply (auto simp: image_iff Ball_def)
  apply (case_tac x)
  apply auto
  done

lemma p_eq_setsum_p_f:
  assumes x[simp]: "x \<in> S" shows "p x y n = (\<Sum>i\<le>n. p y y (n - i) * f x y i)"
proof cases
  assume "x = y" then show ?thesis
    unfolding lessThan_Suc_atMost[symmetric] lessThan_Suc_eq_insert_0
    by (simp add: f_Suc_eq zero_notin_Suc_image setsum_reindex f_0)
next
  assume "x \<noteq> y" then show ?thesis
    apply (cases n)
    apply (simp_all add: p_0 f_0 del: setsum_atMost_Suc)
    apply (subst p_eq_setsum_p_u[OF x])
    apply (subst atMost_Suc_insert_0)
    apply (simp add: zero_notin_Suc_image setsum_reindex f_0 f_Suc)
    done
qed

subsection {* Quantities @{text G}, @{text U}, and @{text F} *} text_raw {* \label{sec:GUF} *}

definition "convergence_G x y z \<longleftrightarrow> summable (\<lambda>n. p x y n * norm z ^ n)"

definition "G s t = (\<integral>\<^sup>+\<omega>. (\<Sum>i. indicator {t} (nat_case s \<omega> i)) \<partial>paths s)"

definition "U s t = \<P>(\<omega> in paths s. \<exists>n. \<omega> n = t)"

definition "F s t = \<P>(\<omega> in paths s. \<exists>n. nat_case s \<omega> n = t)"

lemma G_eq: "G s t = (\<integral>\<^sup>+\<omega>. emeasure (count_space UNIV) {i. nat_case s \<omega> i = t} \<partial>paths s)"
proof -
  have *: "\<And>i \<omega>. indicator {t} (nat_case s \<omega> i) = indicator {i. nat_case s \<omega> i = t} i"
    by (auto split: split_indicator)
  show ?thesis
    unfolding G_def
    by (rule positive_integral_cong)
       (simp add: * positive_intergal_count_space_nat[symmetric])
qed

subsubsection {* @{const G} equals infinite sum of @{const p} *} text_raw {* \label{sec:geqp} *}

lemma G_eq_suminf: "G x y = (\<Sum>i. ereal (p x y i))"
proof -
  have "G x y = (\<integral>\<^sup>+ \<omega>. (\<Sum>i. indicator {\<omega>\<in>space (paths x). nat_case x \<omega> i = y} \<omega>) \<partial>paths x)"
    unfolding G_def by (auto split: split_indicator intro!: positive_integral_cong arg_cong[where f=suminf])
  also have "\<dots> = (\<Sum>i. ereal (p x y i))"
    by (simp add: emeasure_eq_measure[symmetric] p_def emeasure_nonneg positive_integral_suminf)
  finally show ?thesis .
qed

lemma G_eq_real_suminf:
  "convergence_G x y (1::real) \<Longrightarrow> G x y = ereal (\<Sum>i. p x y i)"
  unfolding G_eq_suminf
  by (intro suminf_ereal p_nonneg suminf_ereal_finite) (auto simp: convergence_G_def p_def)

lemma convergence_norm_G:
  "convergence_G x y z \<Longrightarrow> summable (\<lambda>n. p x y n * norm z ^ n)"
  unfolding convergence_G_def .

lemma convergence_G:
  "convergence_G x y (z::'a::{banach, real_normed_div_algebra}) \<Longrightarrow> summable (\<lambda>n. p x y n *\<^sub>R z ^ n)"
  unfolding convergence_G_def
  by (rule summable_norm_cancel) (simp add: abs_mult p_nonneg norm_power)

lemma convergence_G_less_1:
  fixes z :: "_ :: {banach, real_normed_field}"
  assumes z: "norm z < 1" shows "convergence_G x y z"
  unfolding convergence_G_def
proof (rule summable_comparison_test)
  have "\<And>n. p x y n * norm (z ^ n) \<le> 1 * norm (z ^ n)"
    by (intro mult_right_mono p_le_1) simp_all
  then show "\<exists>N. \<forall>n\<ge>N. norm (p x y n * norm z ^ n) \<le> norm z ^ n"
    by (simp add: p_nonneg mult_nonneg_nonneg norm_power)
qed (simp add: z summable_geometric)

lemma convergence_norm_U:
  fixes z :: "_ :: real_normed_div_algebra"
  assumes z: "convergence_G x y z"
  shows "summable (\<lambda>n. u x y n * norm z ^ Suc n)"
  using summable_ignore_initial_segment[OF convergence_norm_G[OF z], of 1]
  by (rule summable_comparison_test[rotated])
     (auto simp add: u_nonneg p_nonneg abs_mult intro!: exI[of _ 0] mult_right_mono mult_nonneg_nonneg u_le_p)

lemma U_cases: "U s s = 1 \<or> U s s < 1"
  by (auto simp add: U_def intro!: antisym)

subsubsection {* @{const U} equals infinite sum of @{const u} *} text_raw {* \label{sec:uequ} *}

lemma u_sums_U: "u x y sums U x y"
proof -
  have "AE \<omega> in paths x. (\<exists>n. \<omega> n = y) \<longrightarrow> (\<exists>!n. (\<forall>i<n. \<omega> i \<noteq> y) \<and> \<omega> n = y)"
  proof (rule AE_I2)
    fix x :: "nat \<Rightarrow> 'a"
    { fix n
      from smallest[of "\<lambda>i. x i = x n", OF refl] guess i .
      then have "\<exists>i. (\<forall>ia<i. x ia \<noteq> x n) \<and> x i = x n"
        by auto }
    note smallest_Ex = this

    show "\<And>y. (\<exists>n. x n = y) \<longrightarrow> (\<exists>!n. (\<forall>i<n. x i \<noteq> y) \<and> x n = y)"
      by (auto intro: smallest_Ex) (metis linorder_cases)
  qed
  then show ?thesis 
    unfolding u_def[abs_def] U_def by (intro prob_sums) auto
qed

lemma accessibleI_U:
  assumes "x \<in> S" "0 < U x y" shows "accessible x y"
proof (rule ccontr)
  assume "\<not> accessible x y"
  
  have "U x y = 0"
    unfolding U_def
    apply (rule prob_eq_0_AE)
    using AE_all_enabled[OF `x \<in> S`]
  proof eventually_elim
    fix \<omega> assume \<omega>[THEN spec]: "\<forall>i. \<omega> i \<in> E (nat_case x \<omega> i)"
    { fix n have "accessible x (\<omega> n)"
      proof (induct n)
        case 0 with \<omega>[of 0] `x \<in> S` show ?case
          by (simp add: accessible_single)
      next
        case (Suc n) with \<omega>[of "Suc n"] show ?case
          by (simp add: accessible.step)
      qed }
    with `\<not> accessible x y` show "\<not> (\<exists>n. \<omega> n = y)"
      by auto
  qed
  with `0 < U x y` show False by auto
qed

lemma F_nonneg: "0 \<le> F x y" by (simp add: F_def measure_nonneg)

lemma F_le_1: "F x y \<le> 1" by (simp add: F_def)

lemma f_sums_F: "f x y sums F x y"
proof -
  have "AE \<omega> in paths x. (\<exists>n. nat_case x \<omega> n = y) \<longrightarrow> (\<exists>!n. (\<forall>i<n. nat_case x \<omega> i \<noteq> y) \<and> nat_case x \<omega> n = y)"
  proof (rule AE_I2)
    fix \<omega> assume "\<omega> \<in> space (paths x)"
    { fix n
      from smallest[of "\<lambda>i. nat_case x \<omega> i = nat_case x \<omega> n", OF refl] guess i .
      then have "\<exists>i. (\<forall>ia<i. nat_case x \<omega> ia \<noteq> nat_case x \<omega> n) \<and> nat_case x \<omega> i = nat_case x \<omega> n"
        by auto }
    note smallest_Ex = this

    show "(\<exists>n. nat_case x \<omega> n = y) \<longrightarrow> (\<exists>!n. (\<forall>i<n. nat_case x \<omega> i \<noteq> y) \<and> nat_case x \<omega> n = y)"
      by (auto intro: smallest_Ex) (metis linorder_cases)
  qed
  then show ?thesis 
    unfolding f_def[abs_def] F_def by (intro prob_sums) auto
qed

lemma U_le_F: "U x y \<le> F x y"
  apply (auto simp: U_def F_def intro!: finite_measure_mono)
  apply (rule_tac x="Suc n" in exI)
  apply simp
  done

lemma convergence_norm_F:
  fixes z :: "_ :: real_normed_div_algebra"
  assumes z: "convergence_G x y z"
  shows "summable (\<lambda>n. f x y n * norm z ^ n)"
  using convergence_norm_G[OF z]
  by (rule summable_comparison_test[rotated])
     (auto simp add: f_nonneg p_nonneg abs_mult intro!: exI[of _ 0] mult_right_mono mult_nonneg_nonneg f_le_p)

lemma U_bounded: "0 \<le> U x y" "U x y \<le> 1"
  unfolding U_def by (simp_all add: measure_nonneg)

subsection {* Generating functions of @{const G}, @{const U}, and @{const F} *} text_raw {* \label{sec:genfun} *}

definition "gf_U x y z = (\<Sum>n. u x y n *\<^sub>R z ^ Suc n)"

definition "gf_G x y z = (\<Sum>n. p x y n *\<^sub>R z ^ n)"

definition "gf_F x y z = (\<Sum>n. f x y n * z ^ n)"

lemma lim_gf_G: "((\<lambda>z. ereal (gf_G x y z)) ---> G x y) (at_left (1::real))"
  unfolding gf_G_def G_eq_suminf real_scaleR_def
  by (intro power_series_tendsto_at_left p_nonneg p_le_1 summable_power_series)

lemma gf_U_eq_U: "gf_U x y 1 = U x y"
  using u_sums_U[THEN sums_unique] by (simp add: gf_U_def U_def)

lemma gf_F_eq_F: "gf_F x y 1 = F x y"
  using f_sums_F[THEN sums_unique] by (simp add: gf_F_def F_def)

lemma gf_F_le_1:
  fixes z :: real
  assumes z: "0 \<le> z" "z \<le> 1"
  shows "gf_F x y z \<le> 1"
proof -
  have "gf_F x y z \<le> gf_F x y 1"
    using z unfolding gf_F_def
    by (intro summable_le[OF _ summable_comparison_test[OF _ sums_summable[OF f_sums_F[of x y]]]] mult_left_mono allI f_nonneg)
       (simp_all add: power_le_one mult_nonneg_nonneg f_nonneg mult_right_le_one_le f_le_1 sums_summable[OF f_sums_F[of x y]])
  also have "\<dots> \<le> 1"
    by (simp add: gf_F_eq_F F_def)
  finally show ?thesis .
qed

lemma gf_G_nonneg:
  fixes z :: real
  shows "0 \<le> z \<Longrightarrow> z < 1 \<Longrightarrow> 0 \<le> gf_G x y z"
  unfolding gf_G_def
  by (intro suminf_ge_zero convergence_G convergence_G_less_1) (simp_all add: mult_nonneg_nonneg p_nonneg)

lemma gf_F_nonneg:
  fixes z :: real
  shows "0 \<le> z \<Longrightarrow> z < 1 \<Longrightarrow> 0 \<le> gf_F x y z"
  unfolding gf_F_def
  using convergence_norm_F[OF convergence_G_less_1, of z x y]
  by (intro suminf_ge_zero) (simp_all add: mult_nonneg_nonneg f_nonneg)
  
lemma convergence_U:
  fixes z :: "_ :: banach"
  shows "convergence_G x y z \<Longrightarrow> summable (\<lambda>n. u x y n * z ^ Suc n)"
  by (rule summable_norm_cancel)
     (auto simp add: abs_mult u_nonneg power_abs dest!: convergence_norm_U)

lemma gf_G_eq_gf_F:
  assumes x[simp]: "x \<in> S"
  assumes z: "norm z < 1"
  shows "gf_G x y z = gf_F x y z * gf_G y y z"
proof -
  have "gf_G x y z = (\<Sum>n. \<Sum>i\<le>n. p y y (n - i) * f x y i * z^n)"
    by (simp add: gf_G_def p_eq_setsum_p_f[of x y] setsum_left_distrib)
  also have "\<dots> = (\<Sum>n. \<Sum>i=0 .. n. (f x y i * z^i) * (p y y (n - i) * z^(n - i)))"
    by (intro arg_cong[where f=suminf] setsum_cong ext atLeast0AtMost[symmetric])
       (simp_all add: power_add[symmetric])
  also have "\<dots> = (\<Sum>n. f x y n * z^n) * (\<Sum>n. p y y n * z^n)"
    using convergence_norm_F[OF convergence_G_less_1[OF z]] convergence_norm_G[OF convergence_G_less_1[OF z]]
    by (intro Cauchy_product[symmetric]) (auto simp: f_nonneg p_nonneg abs_mult power_abs)
  also have "\<dots> = gf_F x y z * gf_G y y z"
    by (simp add: gf_F_def gf_G_def)
  finally show ?thesis .
qed

subsubsection {* Relate @{const gf_G} and @{const gf_U} *} text_raw {* \label{sec:gfGgfU} *}

lemma gf_G_eq_gf_U:
  fixes z :: "'z :: {banach, real_normed_field}"
  assumes x[simp]: "x \<in> S"
  assumes z: "convergence_G x x z"
  shows "gf_G x x z = 1 / (1 - gf_U x x z)" "gf_U x x z \<noteq> 1"
proof -
  { fix n
    have "p x x (Suc n) *\<^sub>R z^Suc n = (\<Sum>i\<le>n. (p x x (n - i) * u x x i) *\<^sub>R z^Suc n)"
      unfolding scaleR_setsum_left[symmetric] by (simp add: p_eq_setsum_p_u)
    also have "\<dots> = (\<Sum>i\<le>n. (u x x i *\<^sub>R z^Suc i) * (p x x (n - i) *\<^sub>R z^(n - i)))"
      by (intro setsum_cong refl) (simp add: field_simps power_diff cong: disj_cong)
    finally have "p x x (Suc n) *\<^sub>R z^(Suc n) = (\<Sum>i=0 .. n. (u x x i *\<^sub>R z^Suc i) * (p x x (n - i) *\<^sub>R z^(n - i)))"
      unfolding atLeast0AtMost . }
  note gfs_Suc_eq = this

  have "gf_G x x z = 1 + (\<Sum>n. p x x (Suc n) *\<^sub>R z^(Suc n))"
    unfolding gf_G_def
    by (subst suminf_split_initial_segment[OF convergence_G[OF z], of 1]) simp
  also have "\<dots> = 1 + (\<Sum>n. \<Sum>i=0 .. n. (u x x i *\<^sub>R z^Suc i) * (p x x (n - i) *\<^sub>R z^(n - i)))"
    unfolding gfs_Suc_eq ..
  also have "\<dots> = 1 + gf_U x x z * gf_G x x z"
    unfolding gf_U_def gf_G_def
    by (subst Cauchy_product) 
       (auto simp: u_nonneg p_nonneg norm_power simp del: power_Suc 
                intro!:  z convergence_norm_G convergence_norm_U)
  finally show "gf_G x x z = 1 / (1 - gf_U x x z)" "gf_U x x z \<noteq> 1"
    apply -
    apply (cases "gf_U x x z = 1")
    apply (auto simp add: field_simps)
    done
qed

lemma gf_G_pos:
  fixes z :: real
  assumes z: "0 < z" "z < 1" and *: "accessible x y"
  shows "0 < gf_G x y z"
  unfolding gf_G_def
proof (subst suminf_gt_zero_iff)
  show "summable (\<lambda>n. p x y n *\<^sub>R z ^ n)"
    using z by (intro convergence_G convergence_G_less_1) simp
  show pos: "\<forall>n. 0 \<le> p x y n *\<^sub>R z ^ n"
    using z by (auto intro!: mult_nonneg_nonneg p_nonneg)
  show "\<exists>n. 0 < p x y n *\<^sub>R z ^ n"
  proof (rule ccontr)
    assume "\<not> (\<exists>n. 0 < p x y n *\<^sub>R z ^ n)"
    with pos have "\<forall>n. p x y n * z ^ n = 0"
      by (intro antisym allI) (simp_all add: not_less)
    with z have "\<And>n. p x y n = 0"
      by simp
    with *[THEN accessibleD_pos] show False
      by simp
  qed
qed

lemma gf_U: "(gf_U x y ---> U x y) (at_left 1)"
proof -
  have "((\<lambda>z. ereal (\<Sum>n. u x y n * z ^ n)) ---> (\<Sum>n. ereal (u x y n))) (at_left 1)"
    using u_le_1 u_nonneg by (intro power_series_tendsto_at_left summable_power_series)
  also have "(\<Sum>n. ereal (u x y n)) = ereal (suminf (u x y))"
    by (intro u_nonneg suminf_ereal suminf_ereal_finite sums_summable[OF u_sums_U])
  also have "suminf (u x y) = U x y"
    using u_sums_U by (rule sums_unique[symmetric])
  finally have "((\<lambda>z. z * (\<Sum>n. u x y n * z ^ n)) ---> 1 * U x y) (at_left 1)"
    by (intro tendsto_intros) simp
  then have "((\<lambda>z. \<Sum>n. u x y n * z ^ Suc n) ---> 1 * U x y) (at_left 1)"
    apply (rule filterlim_cong[OF refl refl, THEN iffD1, rotated])
    apply (rule eventually_at_left_1)
    apply (subst suminf_mult[symmetric])
    apply (auto intro!: summable_power_series u_le_1 u_nonneg)
    apply (simp add: field_simps)
    done
  then show ?thesis
    by (simp add: gf_U_def[abs_def] U_def)
qed

lemma gf_F: "(gf_F x y ---> F x y) (at_left 1)"
proof -
  have "((\<lambda>z. ereal (\<Sum>n. f x y n * z ^ n)) ---> (\<Sum>n. ereal (f x y n))) (at_left 1)"
    using f_le_1 f_nonneg by (intro power_series_tendsto_at_left summable_power_series)
  also have "(\<Sum>n. ereal (f x y n)) = ereal (suminf (f x y))"
    by (intro f_nonneg suminf_ereal suminf_ereal_finite sums_summable[OF f_sums_F])
  also have "suminf (f x y) = F x y"
    using f_sums_F by (rule sums_unique[symmetric])
  finally show ?thesis
    by (simp add: gf_F_def[abs_def] F_def)
qed

subsection {* Recurrent states and @{text H} *} text_raw {* \label{sec:recurrent} *}

definition recurrent :: "'s \<Rightarrow> bool" where
  "recurrent s \<longleftrightarrow> (AE \<omega> in paths s. \<exists>n. \<omega> n = s)"

lemma recurrent_iff_U_eq_1: "recurrent s \<longleftrightarrow> U s s = 1"
    unfolding recurrent_def U_def by (subst prob_Collect_eq_1) simp_all

definition "H s t = \<P>(\<omega> in paths s. infinite {n. \<omega> n = t})"

lemma H_eq:
  assumes "s \<in> S"
  shows "recurrent s \<longleftrightarrow> H s s = 1"
    "\<not> recurrent s \<longleftrightarrow> H s s = 0"
    "H s t = U s t * H t t"
proof -
  def H' \<equiv> "\<lambda>s t n. {\<omega>\<in>space (paths s). finite {n. \<omega> n = t} \<longrightarrow> n \<le> card {n. \<omega> n = t}}"
  have [measurable]: "\<And>x y n. H' x y n \<in> sets S_seq"
    unfolding H'_def space_paths
    apply (intro predE pred_intros_logic)
    apply simp
    apply (intro pred_const_le)
    apply (rule measurable_card)
    apply simp_all
    done
  let ?H' = "\<lambda>s t n. measure (paths s) (H' s t n)"
  { fix x y
    have "?H' x y 1 = U x y"
      unfolding H'_def U_def
      apply (intro arg_cong2[where f=measure] refl arg_cong[where f=Collect] ext conj_cong)
      apply (auto simp add: card_le_Suc_iff)
       apply (rule ccontr)
       apply simp
      apply (rule_tac x="n" in exI)
      apply (rule_tac x="{m. \<omega> m = \<omega> n} - {n}" in exI)
      apply auto
      done }
  note H'_1 = this

  { fix m and x y :: 's assume [simp]: "x \<in> S"
    def i == "\<lambda>\<omega>. if \<exists>n. \<omega> n = y then Suc (LEAST n. \<omega> n = y) else 0"
    have [measurable]: "i \<in> measurable S_seq (count_space UNIV)"
      by (simp add: i_def)
    have "emeasure (paths x) (H' x y (Suc m)) = (\<integral>\<^sup>+ \<omega>. indicator (H' x y (Suc m)) \<omega> \<partial>paths x)"
      by simp
    also have "\<dots> =
       (\<integral>\<^sup>+ \<omega>. indicator {\<omega>. i \<omega> \<noteq> 0} \<omega> *
          (\<integral>\<^sup>+ \<omega>'. indicator (H' x y (Suc m)) (comb_seq (i \<omega>) \<omega> \<omega>') \<partial>paths (nat_case x \<omega> (i \<omega>))) \<partial>paths x)"
    proof (rule positive_integral_prefixes)
      fix \<omega> \<omega>' assume i: "0 < i \<omega>" and le: "\<And>j. j < i \<omega> \<Longrightarrow> \<omega> j = \<omega>' j"
      def j \<equiv> "i \<omega> - 1"
      from i have "\<exists>i. \<omega> i = y"
        unfolding i_def by (auto split: split_if_asm)
      then have "j < i \<omega>" "\<omega> j = y" "\<forall>k<j. \<omega> k \<noteq> y"
        by (auto simp: i_def j_def intro: LeastI_ex dest: not_less_Least split: split_if_asm)
      with le have "i \<omega>' = Suc j"
        by (subst i_def) (auto simp add: not_less[symmetric] intro!: Least_equality)
      then show "i \<omega> = i \<omega>'"
        using `0 < i \<omega>` by (simp add: j_def)
    next
      fix \<omega> assume "i \<omega> = 0"
      then have "\<And>k. \<omega> k \<noteq> y" by (auto simp: i_def split: split_if_asm)
      then show "indicator (H' x y (Suc m)) \<omega> = (0::ereal)"
        by (auto simp add: H'_def split: split_indicator)
    qed simp_all
    also have "\<dots> = (\<integral>\<^sup>+ \<omega>. indicator {\<omega>\<in>space S_seq. \<exists>n. \<omega> n = y} \<omega> *
        (\<integral>\<^sup>+ \<omega>'. indicator (H' y y m) \<omega>' \<partial>paths y) \<partial>paths x)"
    proof (intro positive_integral_cong ereal_right_mult_cong)
      fix \<omega> assume "\<omega> \<in> space (paths x)"
      then show "indicator {\<omega>. i \<omega> \<noteq> 0} \<omega> = indicator {\<omega> \<in> space S_seq. \<exists>n. \<omega> n = y} \<omega>"
        by (auto simp add: i_def split: split_indicator)
    next
      fix \<omega> assume "indicator {\<omega> \<in> space S_seq. \<exists>n. \<omega> n = y} \<omega> \<noteq> 0" and \<omega>: "\<omega> \<in> space (paths x)"
      then have ex_y: "\<exists>n. \<omega> n = y" "i \<omega> \<noteq> 0"
        by (auto simp: indicator_def i_def split: split_if_asm)
      then have x_i_\<omega>: "nat_case x \<omega> (i \<omega>) = y"
        by (auto simp add: i_def intro: LeastI)
      then have paths_y: "paths (nat_case x \<omega> (i \<omega>)) = paths y"
        by simp
      { fix \<omega>' assume \<omega>': "\<omega>' \<in> space S_seq"
        have "\<And>n. Suc n < i \<omega> \<Longrightarrow> \<omega> n \<noteq> y"
          using ex_y by (auto simp add: i_def dest: not_less_Least split: if_splits)
        with x_i_\<omega> have "\<And>n. comb_seq (i \<omega>) \<omega> \<omega>' n = y \<longleftrightarrow>
            (i \<omega> \<le> Suc n \<and> nat_case y \<omega>' (Suc n - i \<omega>) = y)"
          by (auto simp add: comb_seq_def not_less le_imp_diff_is_add le_less split: nat.splits)
        then have A: "{n. comb_seq (i \<omega>) \<omega> \<omega>' n = y} = (\<lambda>n. n + (i \<omega> - 1)) ` {n. nat_case y \<omega>' n = y}"
          (is "?A = ?s ` ?B")
          using `i \<omega> \<noteq> 0`
          apply simp
          apply safe
          apply (rule_tac x="Suc x - i \<omega>" in image_eqI)
          apply simp_all
          done
        have s: "inj_on ?s ?B" by auto
        have B: "?B = {0} \<union> Suc ` {n. \<omega>' n = y}"
          by (simp add: set_eq_iff image_Collect split: nat.split)
        have "finite {n. \<omega>' n = y} \<Longrightarrow> card ?A = Suc (card {n. \<omega>' n = y})"
          unfolding A card_image[OF s] by (simp add: zero_notin_Suc_image card_image B)
        moreover have "finite ?A \<longleftrightarrow> finite {n. \<omega>' n = y}"
          unfolding A B by (auto dest!: finite_imageD)
        ultimately have "comb_seq (i \<omega>) \<omega> \<omega>' \<in> H' x y (Suc m) \<longleftrightarrow> \<omega>' \<in> H' y y m"
          using \<omega> \<omega>' by (auto simp add: H'_def space_PiM PiE_iff Pi_iff comb_seq_def)
        then have "indicator (H' x y (Suc m)) (comb_seq (i \<omega>) \<omega> \<omega>') = (indicator (H' y y m) \<omega>' :: ereal)"
          by (simp split: split_indicator) }
      then show "(\<integral>\<^sup>+ \<omega>'. indicator (H' x y (Suc m)) (comb_seq (i \<omega>) \<omega> \<omega>') \<partial>paths (nat_case x \<omega> (i \<omega>))) =
          integral\<^sup>P (paths y) (indicator (H' y y m))"
        unfolding paths_y by (intro positive_integral_cong) simp
    qed
    also have "\<dots> = U x y * ?H' y y m"
      by (simp add: positive_integral_multc measure_nonneg emeasure_eq_measure U_def)
    finally have "?H' x y (Suc m) = U x y * ?H' y y m"
      by (simp add: emeasure_eq_measure) }
  note H'_Suc = this

  { fix m x assume "x \<in> S"
    then have "?H' x x (Suc m) = U x x^Suc m"
      using H'_1 H'_Suc by (induct m) auto }
  note H'_eq = this

  { fix x y
    have "?H' x y ----> measure (paths x) (\<Inter>i. H' x y i)"
      apply (rule finite_Lim_measure_decseq)
      apply safe
      apply simp
      apply (auto simp add: decseq_Suc_iff subset_eq H'_def)
      done
    also have "(\<Inter>i. H' x y i) = {\<omega>\<in>space (paths x). infinite {n. \<omega> n = y}}"
     by (auto simp: H'_def not_less[symmetric])
    finally have "?H' x y ----> H x y"
      unfolding H_def . }
  note H'_lim = this  

  from H'_lim[of s s, THEN LIMSEQ_Suc] `s\<in>S`
  have "(\<lambda>n. U s s ^ Suc n) ----> H s s"
    by (simp add: H'_eq)
  then have lim_H: "(\<lambda>n. U s s ^ n) ----> H s s"
    by (rule LIMSEQ_imp_Suc)

  have "U s s < 1 \<Longrightarrow> (\<lambda>n. U s s ^ n) ----> 0"
    by (rule LIMSEQ_realpow_zero) (simp_all add: U_def measure_nonneg)
  with lim_H have "U s s < 1 \<Longrightarrow> H s s = 0"
    by (blast intro: LIMSEQ_unique)
  moreover have "U s s = 1 \<Longrightarrow> (\<lambda>n. U s s ^ n) ----> 1"
    by (simp add: tendsto_const)
  with lim_H have "U s s = 1 \<Longrightarrow> H s s = 1"
    by (blast intro: LIMSEQ_unique)
  moreover note recurrent_iff_U_eq_1 U_cases
  ultimately show "recurrent s \<longleftrightarrow> H s s = 1" "\<not> recurrent s \<longleftrightarrow> H s s = 0"
    by (metis one_neq_zero)+

  from H'_lim[of s t, THEN LIMSEQ_Suc] `s\<in>S` H'_Suc[OF `s\<in>S`]
  have "(\<lambda>n. U s t * ?H' t t n) ----> H s t"
    by simp
  moreover have "(\<lambda>n. U s t * ?H' t t n) ----> U s t * H t t"
    by (intro tendsto_intros H'_lim)
  ultimately show "H s t = U s t * H t t"
    by (blast intro: LIMSEQ_unique)
qed

subsubsection {* Relating @{const recurrent} and @{const G} *} text_raw {* \label{sec:recurrentG} *}

lemma recurrent_iff_G_infinite:
  assumes "x \<in> S"
  shows "recurrent x \<longleftrightarrow> G x x = \<infinity>"
proof -
  have "((\<lambda>z. ereal (gf_G x x z)) ---> G x x) (at_left 1)"
    by (rule lim_gf_G)
  then have G: "((\<lambda>z. ereal (1 / (1 - gf_U x x z))) ---> G x x) (at_left (1::real))"
    apply (rule filterlim_cong[OF refl refl, THEN iffD1, rotated])
    apply (rule eventually_at_left_1)
    apply (subst gf_G_eq_gf_U)
    apply (rule `x \<in> S`)
    apply (rule convergence_G_less_1)
    apply simp
    apply simp
    done

  { fix z :: real assume z: "0 < z" "z < 1"
    have 1: "summable (u x x)"
      using u_sums_U by (rule sums_summable)
    have "gf_U x x z \<noteq> 1"
      using gf_G_eq_gf_U[OF `x\<in>S` convergence_G_less_1[of z]] z by simp
    moreover
    have "gf_U x x z \<le> U x x"
      unfolding gf_U_def gf_U_eq_U[symmetric]
      using z
      by (intro summable_le)
         (auto simp add: 1 convergence_U convergence_G_less_1 u_nonneg simp del: power_Suc
               intro!: mult_right_le_one_le power_le_one)
    ultimately have "gf_U x x z < 1"
      using U_bounded[of x x] by simp }
  note strict = this

  { assume "U x x = 1"
    moreover have "((\<lambda>xa. 1 - gf_U x x xa :: real) ---> 1 - U x x) (at_left 1)"
      by (intro tendsto_intros gf_U)
    moreover have "eventually (\<lambda>z. gf_U x x z < 1) (at_left (1::real))"
      by (auto intro!: eventually_at_left_1 strict simp: `U x x = 1` gf_U_eq_U)
    ultimately have "((\<lambda>z. ereal (1 / (1 - gf_U x x z))) ---> \<infinity>) (at_left 1)"
      unfolding tendsto_PInfty_eq_at_top
      by (intro LIM_at_top_divide[where a=1] tendsto_const zero_less_one)
         (auto simp: field_simps)
    with G have "G x x = \<infinity>"
      by (rule tendsto_unique[rotated]) simp }
  moreover
  { assume "U x x < 1"
    then have "((\<lambda>xa. ereal (1 / (1 - gf_U x x xa))) ---> 1 / (1 - U x x)) (at_left 1)"
      by (intro tendsto_intros gf_U lim_ereal[THEN iffD2]) simp
    from tendsto_unique[OF _ G this] have "G x x \<noteq> \<infinity>"
      by simp }
  ultimately show ?thesis
    using U_cases recurrent_iff_U_eq_1 by blast    
qed

subsubsection {* @{const recurrent} is invariant *} text_raw {* \label{sec:invariantreccurrent} *}

lemma recurrent_iffI_communicating:
  assumes "(x, y) \<in> communicating"
  shows "recurrent x \<longleftrightarrow> recurrent y"
proof -
  from assms obtain n m where "0 < p x y n" "y \<in> S" "0 < p y x m" "x \<in> S"
    by (force simp: communicating_def dest: accessibleD_pos intro: accessible_in_S1)
  moreover
  { fix x y n m assume [simp]: "x \<in> S" "y \<in> S"
    assume "0 < p x y n" "0 < p y x m" "G y y = \<infinity>"
    then have "\<infinity> = ereal (p x y n * p y x m) * G y y"
      by (auto intro: mult_pos_pos)
    also have "ereal (p x y n * p y x m) * G y y = (\<Sum>i. ereal (p x y n * p y x m) * p y y i)"
      unfolding G_eq_suminf
      by (rule suminf_cmult_ereal[symmetric]) (auto simp: p_nonneg mult_nonneg_nonneg)
    also have "\<dots> \<le> (\<Sum>i. ereal (p x x (n + i + m)))"
    proof (intro suminf_le_pos)
      fix i
      have "(p x y n * p y y ((n + i) - n)) * p y x ((n + i + m) - (n + i)) \<le> p x y (n + i) * p y x ((n + i + m) - (n + i))"
        by (intro mult_right_mono prob_reachable_le) (simp_all add: p_nonneg)
      also have "\<dots> \<le> p x x (n + i + m)"
         by (intro prob_reachable_le) simp_all
      finally show "ereal (p x y n * p y x m) * p y y i \<le> ereal (p x x (n + i + m))"
        by (simp add: ac_simps)
    qed (simp add: p_nonneg mult_nonneg_nonneg)
    also have "\<dots> \<le> (\<Sum>i. ereal (p x x (i + (n + m))))"
      by (simp add: ac_simps)
    also have "\<dots> \<le> (\<Sum>i. ereal (p x x i))"
      by (rule suminf_ereal_offset_le) (simp add: p_nonneg)
    also have "\<dots> \<le> G x x"
      unfolding G_eq_suminf by (auto intro!: suminf_le_pos)
    finally have "G x x = \<infinity>"
      by simp }
  ultimately show ?thesis
    using recurrent_iff_G_infinite by blast
qed

subsubsection {* @{const recurrent} and @{const H} and @{const U} *} text_raw {* \label{sec:recurrentHU} *}

lemma recurrent_accessible:
  assumes "recurrent x" "accessible x y"
  shows "U y x = 1" "H y x = 1" "recurrent y" "(x, y) \<in> communicating"
proof -
  { fix w y assume step: "accessible x w" "y \<in> E w" "U w x = 1" "H w x = 1" "recurrent w" "x \<noteq> y"
    then have [simp]: "y \<in> S" "w \<in> S" "x \<in> S"
      by (auto intro: accessible_in_S2 accessible_in_S1)

    have "measure (K w) S = U w x"
      using step K.prob_space[of w] by simp
    also have "\<dots> = (\<integral>v. indicator {x} v + U v x * indicator (S - {x}) v \<partial>K w)"
      unfolding U_def
      apply (subst prob_iterate_Collect)
      apply (auto intro!: integral_cong arg_cong2[where f=measure] AE_I2
                  simp: exI[of _ 0] prob_eq_1 split: split_indicator)
      apply (rule_tac x="n - 1" in exI)
      apply (simp split: nat.splits)
      apply (rule_tac x="Suc n" in exI)
      apply (simp split: nat.splits)
      done
    also have "\<dots> = measure (K w) {x} + (\<integral>v. U v x * indicator (S - {x}) v \<partial>K w)"
      by (subst integral_add)
         (auto intro!: K.integrable_const_bound[where B=1]
               simp: abs_mult mult_le_one U_bounded K.emeasure_eq_measure)
    finally have "measure (K w) S - measure (K w) {x} = (\<integral>v. U v x * indicator (S - {x}) v \<partial>K w)"
      by simp
    also have "measure (K w) S - measure (K w) {x} = measure (K w) (S - {x})"
      by (subst K.finite_measure_Diff) auto
    finally have "0 = (\<integral>v. indicator (S - {x}) v \<partial>K w) - (\<integral>v. U v x * indicator (S - {x}) v \<partial>K w)"
      by (simp add: Diff_subset K.emeasure_eq_measure)
    also have "\<dots> = (\<integral>v. (1 - U v x) * indicator (S - {x}) v \<partial>K w)"
      by (subst integral_diff(2)[symmetric])
         (auto intro!: K.integrable_const_bound[where B=1] integral_cong
               simp: abs_mult mult_le_one U_bounded split: split_indicator)
    also have "\<dots> \<ge> (\<integral>v. (1 - U y x) * indicator {y} v \<partial>K w)" (is "_ \<ge> ?rhs")
      using `recurrent x`
      by (auto intro!: integral_mono K.integrable_const_bound[where B=1]
               simp: abs_mult mult_le_one U_bounded recurrent_iff_U_eq_1 field_simps simp del: `y\<in>S`
               split: split_indicator)
    also (xtrans) have "?rhs = (1 - U y x) * measure (K w) {y}"
      by (simp add: integral_cmul_indicator K.emeasure_eq_measure)
    finally have "(1 - U y x) * measure (K w) {y} = 0"
      by (auto intro!: antisym simp: U_bounded measure_nonneg mult_le_0_iff)
    with `y \<in> E w` have "U y x = 1"
      by (simp add: E_def field_simps K.emeasure_eq_measure)
    then have "U y x = 1" "H y x = 1"
      using H_eq(3)[of y x] H_eq(1)[of x] by (simp_all add: `recurrent x`)
    then have "accessible y x"
      by (intro accessibleI_U) auto
    with step have "(x, y) \<in> communicating"
      by (auto simp add: communicating_def intro: accessible.intros)
    with `recurrent x` have "recurrent y"
      by (simp add: recurrent_iffI_communicating)
    note this `U y x = 1` `H y x = 1` `(x, y) \<in> communicating` }
  note enabled = this

  from `accessible x y`
  show "U y x = 1" "H y x = 1" "recurrent y" "(x, y) \<in> communicating"
  proof induction
    case start then show "U x x = 1" "H x x = 1" "recurrent x" "(x, x) \<in> communicating"
      using `recurrent x` H_eq(1)[of x] by (auto simp: recurrent_iff_U_eq_1 communicating_def accessible.start)
  next
    case (step w y)
    with enabled[of w y] `recurrent x` H_eq(1)[of x]
    have "U y x = 1 \<and> H y x = 1 \<and> recurrent y \<and> (x, y) \<in> communicating"
      by (cases "x = y") (auto simp: recurrent_iff_U_eq_1 communicating_def accessible.start)
    then show "U y x = 1" "H y x = 1" "recurrent y" "(x, y) \<in> communicating"
      by auto
  qed
qed

lemma recurrent_class:
  assumes "recurrent x"
  shows "{y. accessible x y} = communicating `` {x}"
  using recurrent_accessible(4)[OF `recurrent x`] by (auto simp: communicating_def)

lemma irreduccible_recurrent_class:
  assumes "x \<in> S" "recurrent x" shows "{y. accessible x y} \<in> S // communicating"
  unfolding recurrent_class[OF `recurrent x`] by (rule quotientI[OF `x \<in> S`])

subsection {* Essential classes *} text_raw {* \label{sec:essential} *}

definition essential_class :: "'s set \<Rightarrow> bool" where
  "essential_class S' \<longleftrightarrow> S' \<in> S // communicating \<and> \<not> (\<exists>s\<in>S'. \<exists>t\<in>S - S'. accessible s t)"

lemma essential_classI:
  assumes C: "C \<in> S // communicating"
  assumes eq: "\<And>x y. x \<in> C \<Longrightarrow> accessible x y \<Longrightarrow> y \<in> C"
  shows "essential_class C"
  by (auto simp: essential_class_def intro: C) (metis eq)

lemma essential_classI2:
  assumes "X \<noteq> {}" "X \<subseteq> S"
  assumes accI: "\<And>x y. x \<in> X \<Longrightarrow> y \<in> X \<Longrightarrow> accessible x y"
  assumes ED: "\<And>x y. x \<in> X \<Longrightarrow> y \<in> E x \<Longrightarrow> y \<in> X"
  shows "essential_class X"
proof (rule essential_classI)
  { fix x y assume "accessible x y" "x \<in> X"
    then show "y \<in> X"
      by induct (auto dest: ED)}
  note accD = this

  from `X \<noteq> {}` obtain x where "x \<in> X" by auto
  from `x \<in> X` `X \<subseteq> S` show "X \<in> S // communicating"
    by (auto simp add: quotient_def Image_def communicating_def accI dest: accD intro!: bexI[of _ x])
qed

lemma essential_recurrent_class:
  assumes "x \<in> S" "recurrent x" shows "essential_class (communicating `` {x})"
  unfolding recurrent_class[OF `recurrent x`, symmetric]
  apply (rule essential_classI)
  apply (rule irreduccible_recurrent_class[OF assms])
  apply (auto simp: communicating_def intro: accessible_trans)
  done

lemma essential_classD1:
  "essential_class C \<Longrightarrow> x \<in> C \<Longrightarrow> x \<in> S"
  unfolding essential_class_def
  by (auto elim!: quotientE simp: communicating_in_S)

lemma essential_classD2:
  "essential_class C \<Longrightarrow> x \<in> C \<Longrightarrow> accessible x y \<Longrightarrow> y \<in> C"
  unfolding essential_class_def by (auto intro: accessible_in_S2)

lemma essential_classD3:
  "essential_class C \<Longrightarrow> x \<in> C \<Longrightarrow> y \<in> C \<Longrightarrow> (x, y) \<in> communicating"
  unfolding essential_class_def
  by (auto elim!: quotientE simp: communicating_def intro: accessible_trans)

lemma finite_essential_class_imp_recurrent:
  assumes C: "essential_class C" "finite C" and x: "x \<in> C"
  shows "recurrent x"
proof -
  note essential_classD1[OF C(1) x]
  have "AE \<omega> in paths x. \<exists>y\<in>C. infinite {n. \<omega> n = y}"
    using AE_accessible[OF `x \<in> S`]
  proof eventually_elim
    fix \<omega> assume *: "\<forall>m. accessible x (case m of 0 \<Rightarrow> x | Suc x \<Rightarrow> \<omega> x)"
    have "\<forall>m\<in>UNIV. \<exists>x\<in>C. \<omega> m = x"
    proof
      fix m
      from *[THEN spec, of "Suc m"] essential_classD2[OF C(1) x]
      show "\<exists>x\<in>C. \<omega> m = x"
        by auto
    qed
    from pigeonhole_infinite_rel[OF nat_infinite C(2) this]
    show "\<exists>y\<in>C. infinite {n. \<omega> n = y}"
      by simp
  qed
  then have "1 = \<P>(\<omega> in paths x. \<exists>y\<in>C. infinite {n. \<omega> n = y})"
    by (subst (asm) prob_Collect_eq_1[symmetric]) (auto simp: `finite C`)
  also have "\<dots> = measure (paths x) (\<Union>y\<in>C. {\<omega>\<in>space (paths x). infinite {n. \<omega> n = y}})"
    by (intro arg_cong2[where f=measure]) auto
  also have "\<dots> \<le> (\<Sum>y\<in>C. H x y)"
    unfolding H_def using `finite C` by (rule finite_measure_subadditive_finite) auto
  also have "\<dots> = (\<Sum>y\<in>C. U x y * H y y)"
    by (auto intro!: setsum_cong H_eq `x \<in> S`)
  finally have "\<exists>y\<in>C. recurrent y"
    by (rule_tac ccontr) (simp add: H_eq(2) essential_classD1[OF C(1)])
  then guess y ..
  from essential_classD3[OF C(1) x this(1)] recurrent_accessible(3)[OF this(2)]
  show "recurrent x"
    by (simp add: communicating_def)
qed

lemma essential_class_iff_recurrent:
  "finite C \<Longrightarrow> C \<in> S // communicating \<Longrightarrow> essential_class C \<longleftrightarrow> (\<forall>x\<in>C. recurrent x)"
  by (metis finite_essential_class_imp_recurrent irreducibleD2 recurrent_accessible(4) essential_classI)

subsection {* Hitting times *} text_raw {* \label{sec:hitting} *}

definition "t x \<omega> = (if \<exists>i. \<omega> i = x then ereal (Suc (LEAST i. \<omega> i = x)) else \<infinity>)"

definition "gf_U' x y z = (\<Sum>n. u x y n * Suc n * z ^ n)"

lemma summable_gf_U':
  assumes z: "norm z < 1"
  shows "summable (\<lambda>n. u x y n * Suc n * z ^ n)"
proof -
  have "summable (\<lambda>n. n * \<bar>z\<bar> ^ n)"
  proof (rule root_test_convergence)
    have "(\<lambda>n. root n n * \<bar>z\<bar>) ----> 1 * \<bar>z\<bar>"
      by (intro tendsto_intros LIMSEQ_root)
    then show "(\<lambda>n. root n (norm (n * \<bar>z\<bar> ^ n))) ----> \<bar>z\<bar>"
      by (rule filterlim_cong[THEN iffD1, rotated 3])
         (auto intro!: exI[of _ 1]
               simp add: abs_mult u_nonneg real_root_mult power_abs eventually_sequentially real_root_power)
  qed (insert z, simp add: abs_less_iff)
  note summable_mult[OF this, of "1 / \<bar>z\<bar>"]
  from summable_ignore_initial_segment[OF this, of 1]
  show "summable (\<lambda>n. u x y n * Suc n * z ^ n)"
    apply (rule summable_comparison_test[rotated])
    using z
    apply (auto intro!: exI[of _ 1]
                simp: abs_mult u_nonneg power_abs Suc_le_eq gr0_conv_Suc field_simps le_divide_eq u_le_1)
    done
qed

lemma DERIV_gf_U:
  fixes z :: real assumes z: "0 < z" "z < 1"
  shows "DERIV (gf_U x y) z :> gf_U' x y z"
  unfolding gf_U_def[abs_def]  gf_U'_def real_scaleR_def u_def[symmetric]
  using z by (intro DERIV_power_series'[where R=1] summable_gf_U') auto

lemma integral_t_eq:
  "recurrent x \<Longrightarrow> (\<integral>\<^sup>+\<omega>. t x \<omega>\<partial>paths x) = (\<integral>\<^sup>+\<omega>. ereal (Suc (LEAST n. \<omega> n = x)) \<partial>paths x)"
  by (auto simp: recurrent_def t_def intro!: positive_integral_cong_AE elim!: eventually_elim1)

lemma one_le_integral_t:
  assumes x: "recurrent x" shows "1 \<le> (\<integral>\<^sup>+\<omega>. t x \<omega> \<partial>paths x)" 
  unfolding integral_t_eq[OF x]
  by (auto intro!: positive_integral_ge_const simp: one_ereal_def)

subsubsection {* Integral equals infinite sum of @{const u} *} text_raw {* \label{sec:mequ} *}

lemma integral_t_eq_suminf:
  assumes x: "recurrent x" and [simp]: "x \<in> S"
  shows "(\<integral>\<^sup>+\<omega>. t x \<omega>\<partial>paths x) = (\<Sum>i. ereal (u x x i * Suc i))"
proof -
  let ?L = "\<lambda>\<omega>. LEAST n. \<omega> n = x"
  { fix t
    have "(\<lambda>n. \<P>(\<omega> in paths x. n \<ge> t \<and> (\<forall>i<n. \<omega> i \<noteq> x) \<and> \<omega> n = x)) sums \<P>(\<omega> in paths x. t \<le> ?L \<omega>)"
      apply (rule prob_sums[rotated 2])
      using x unfolding recurrent_def
      apply eventually_elim
    proof (intro iffI conjI allI impI)
      fix \<omega> and n assume ex: "\<exists>n. \<omega> n = x" and L: "t \<le> ?L \<omega>"
      then show "\<exists>!n. t \<le> n \<and> (\<forall>i<n. \<omega> i \<noteq> x) \<and> \<omega> n = x"
        apply (intro ex1I[of _ "?L \<omega>"])
        apply (auto dest: not_less_Least intro: LeastI)
        apply (rule LeastI2_wellorder)
        apply auto
        apply (metis linorder_cases not_le)
        done
    next
      fix \<omega> n assume ex: "\<exists>n. \<omega> n = x" and t: "t \<le> n \<and> (\<forall>i<n. \<omega> i \<noteq> x) \<and> \<omega> n = x"
      show "t \<le> ?L \<omega>"
        by (rule LeastI2_ex[OF ex]) (metis t not_less le_less_trans)
    qed simp_all
    also have "(\<lambda>n. \<P>(\<omega> in paths x. n \<ge> t \<and> (\<forall>i<n. \<omega> i \<noteq> x) \<and> \<omega> n = x)) = (\<lambda>n. u x x n * indicator {t ..} n)"
      by (auto simp: u_def split: split_indicator)
    finally have "\<P>(\<omega> in paths x. t \<le> ?L \<omega>) = (\<integral>\<^sup>+n. u x x n * indicator {t ..} n\<partial>count_space UNIV)"
      apply (subst positive_intergal_count_space_nat)
      apply (auto dest!: sums_unique intro!: mult_nonneg_nonneg u_nonneg)
      apply (intro suminf_ereal[symmetric] mult_nonneg_nonneg u_nonneg indicator_pos_le suminf_ereal_finite)
      apply (rule summable_comparison_test[OF _ u_sums_U[THEN sums_summable, of x x]])
      apply (auto intro!: exI[of _ 0] simp: u_nonneg split: split_indicator)
      done }
  note eq = this
  have "(\<integral>\<^sup>+\<omega>. t x \<omega>\<partial>paths x) = (\<integral>\<^sup>+\<omega>. ereal (Suc (?L \<omega>)) \<partial>paths x)"
    using x by (rule integral_t_eq)
  also have "\<dots> = (\<Sum>t. ereal (\<P>(\<omega> in paths x. t \<le> (?L \<omega>))))"
    unfolding real_eq_of_nat by (subst positive_integral_nat_function) (simp_all add: emeasure_eq_measure less_Suc_eq_le)
  also have "\<dots> = (\<Sum>t. \<integral>\<^sup>+n. u x x n * indicator {t ..} n\<partial>count_space UNIV)"
    unfolding eq ..
  also have "\<dots> = (\<integral>\<^sup>+n. (\<Sum>t. ereal (u x x n * indicator {t ..} n))\<partial>count_space UNIV)"
    by (subst positive_integral_suminf) (simp_all add: mult_nonneg_nonneg u_nonneg)
  also have "(\<lambda>n. (\<Sum>t. ereal (u x x n * indicator {t ..} n))) = (\<lambda>n. ereal (u x x n * Suc n))"
  proof
    fix n
    have "(\<Sum>t. ereal (u x x n * indicator {t ..} n)) = (\<Sum>t\<le>n. ereal (u x x n * indicator {t ..} n))"
      by (intro suminf_finite) (auto split: split_indicator)
    also have "\<dots> = (\<Sum>t\<le>n. ereal (u x x n))"
      by (intro setsum_cong) (auto split: split_indicator)
    also have "\<dots> = ereal (u x x n * Suc n)"
      by (simp add: real_eq_of_nat ac_simps)
    finally show "(\<Sum>t. ereal (u x x n * indicator {t ..} n)) = ereal (u x x n * Suc n)" .
  qed
  also have "(\<integral>\<^sup>+n. ereal (u x x n * Suc n) \<partial>count_space UNIV) = (\<Sum>n. ereal (u x x n * Suc n))"
    by (simp add: u_nonneg mult_nonneg_nonneg positive_intergal_count_space_nat)
  finally show ?thesis .
qed

subsubsection {* @{const gf_U'} approaches the integral of @{const t} *} text_raw {* \label{sec:gfM} *}

lemma gf_U'_tendsto_integral_t:
  assumes x: "recurrent x" "x \<in> S"
  shows "((\<lambda>z. ereal (gf_U' x x z)) ---> (\<integral>\<^sup>+\<omega>. t x \<omega>\<partial>paths x)) (at_left 1)"
  unfolding integral_t_eq_suminf[OF x] gf_U'_def
  by (auto intro!: power_series_tendsto_at_left summable_gf_U' mult_nonneg_nonneg u_nonneg)

lemma gf_U'_pos:
  fixes z :: real
  assumes z: "0 < z" "z < 1" and "U x y \<noteq> 0"
  shows "0 < gf_U' x y z"
  unfolding gf_U'_def
proof (subst suminf_gt_zero_iff)
  show "summable (\<lambda>n. u x y n * real (Suc n) * z ^ n)"
    using z by (intro summable_gf_U') simp
  show pos: "\<forall>n. 0 \<le> u x y n * real (Suc n) * z ^ n"
    using z by (auto intro!: mult_nonneg_nonneg u_nonneg)
  show "\<exists>n. 0 < u x y n * real (Suc n) * z ^ n"
  proof (rule ccontr)
    assume "\<not> (\<exists>n. 0 < u x y n * real (Suc n) * z ^ n)"
    with pos have "\<forall>n. u x y n * real (Suc n) * z ^ n = 0"
      by (intro antisym allI) (simp_all add: not_less)
    with z have "u x y = (\<lambda>n. 0)"
      by (intro ext) simp
    with u_sums_U[of x y, THEN sums_unique] `U x y \<noteq> 0` show False
      by simp
  qed
qed

lemma inverse_gf_U'_tendsto:
  assumes "recurrent y" "y \<in> S"
  shows "((\<lambda>x. - 1 / - gf_U' y y x) ---> real (1 / integral\<^sup>P (paths y) (t y))) (at_left (1::real))"
proof cases
  assume inf: "integral\<^sup>P (paths y) (t y) = \<infinity>"
  with gf_U'_tendsto_integral_t[of y] `recurrent y` `y \<in> S`
  have "LIM z (at_left 1). gf_U' y y z :> at_top"
    by (auto simp: tendsto_PInfty_eq_at_top)
  then have "LIM z (at_left 1). gf_U' y y z :> at_infinity"
    by (rule filterlim_mono) (auto simp: at_top_le_at_infinity)
  then show ?thesis
    by (auto intro!: tendsto_divide_0 simp add: inf)
next
  assume fin: "integral\<^sup>P (paths y) (t y) \<noteq> \<infinity>"
  then obtain r where r: "integral\<^sup>P (paths y) (t y) = ereal r"
    by (cases "integral\<^sup>P (paths y) (t y)") auto
  then have eq: "real (1 / integral\<^sup>P (paths y) (t y)) = - 1 / - r" and "1 \<le> r"
    using one_le_integral_t[OF `recurrent y`] by (auto simp add: one_ereal_def) 
  have gf_U': "(gf_U' y y ---> r) (at_left (1::real))"
    using gf_U'_tendsto_integral_t[OF `recurrent y` `y \<in> S`] r
    by simp
  show ?thesis
    using `1 \<le> r` unfolding eq by (intro tendsto_intros gf_U') simp
qed

subsection {* Positive recurrent *} text_raw {* \label{sec:posrecurrent} *}

definition "pos_recurrent x \<longleftrightarrow> recurrent x \<and> (\<integral>\<^sup>+\<omega>. t x \<omega>\<partial>paths x) \<noteq> \<infinity>"

lemma pos_recurrentI_communicating:
  assumes y: "pos_recurrent y" and x: "(y, x) \<in> communicating"
  shows "pos_recurrent x"
proof -
  let ?E = "\<lambda>x. integral\<^sup>P (paths x) (t x)"
  from y x have recurrent: "recurrent y" "recurrent x" and fin: "?E y \<noteq> \<infinity>"
    by (auto simp: pos_recurrent_def recurrent_iffI_communicating)
  have in_S[simp]: "x \<in> S" "y \<in> S"
    using x by (auto dest: communicating_in_S)
  have pos: "0 < real (1 / ?E y)"
    using one_le_integral_t[OF `recurrent y`]
    by (subst zero_less_real_of_ereal)
       (auto simp add: divide_ereal_def ereal_0_gt_inverse positive_integral_positive fin)
  
  from fin obtain r where r: "?E y = ereal r"
    by (cases "?E y") auto

  from x obtain n m where "0 < p x y n" "0 < p y x m"
    by (auto dest!: accessibleD_pos simp: communicating_def)

  let ?L = "at_left (1::real)"
  have le: "eventually (\<lambda>z. p x y n * p y x m * z^(n + m) \<le> (1 - gf_U y y z) / (1 - gf_U x x z)) ?L"
  proof (rule eventually_at_left_1)
    fix z :: real assume z: "0 < z" "z < 1"
    then have conv: "\<And>x. convergence_G x x z"
      by (intro convergence_G_less_1) simp
    have sums: "(\<lambda>i. (p x y n * p y x m * z^(n + m)) * (p y y i * z^i)) sums ((p x y n * p y x m * z^(n + m)) * gf_G y y z)"
      unfolding gf_G_def
      by (intro sums_mult summable_sums) (auto intro!: conv summable_sums convergence_G[where 'a=real, simplified])
    have "(\<Sum>i. (p x y n * p y x m * z^(n + m)) * (p y y i * z^i)) \<le> (\<Sum>i. p x x (i + (n + m)) * z^(i + (n + m)))"
    proof (intro allI summable_le sums_summable[OF sums] summable_ignore_initial_segment convergence_G[where 'a=real, simplified] convergence_G_less_1)
      show "norm z < 1" using z by simp
      fix i
      have "(p x y n * p y y ((n + i) - n)) * p y x ((n + i + m) - (n + i)) \<le> p x y (n + i) * p y x ((n + i + m) - (n + i))"
        by (intro mult_right_mono prob_reachable_le) (simp_all add: p_nonneg)
      also have "\<dots> \<le> p x x (n + i + m)"
         by (intro prob_reachable_le) simp_all
      finally show "p x y n * p y x m * z ^ (n + m) * (p y y i * z ^ i) \<le> p x x (i + (n + m)) * z ^ (i + (n + m))"
        using z by (auto simp add: ac_simps power_add intro!: mult_left_mono)
    qed
    also have "\<dots> \<le> gf_G x x z"
      unfolding gf_G_def
      using z
      apply (subst (2) suminf_split_initial_segment[where k="n + m"])
      apply (intro convergence_G conv)
      apply (simp add: setsum_nonneg mult_nonneg_nonneg p_nonneg)
      done
    finally have "(p x y n * p y x m * z^(n + m)) * gf_G y y z \<le> gf_G x x z"
      using sums_unique[OF sums] by simp 
    then have "(p x y n * p y x m * z^(n + m)) \<le> gf_G x x z / gf_G y y z"
      using z accessible.start[OF `y\<in>S`] gf_G_pos[of z y y] by (simp add: field_simps) 
    also have "\<dots> = (1 - gf_U y y z) / (1 - gf_U x x z)"
      unfolding gf_G_eq_gf_U(1)[OF `x\<in>S` conv] gf_G_eq_gf_U(1)[OF `y\<in>S` conv]
      using gf_G_eq_gf_U(2)[OF `x\<in>S` conv] gf_G_eq_gf_U(2)[OF `y\<in>S` conv]
      by (simp add: field_simps )
    finally show "p x y n * p y x m * z^(n + m) \<le> (1 - gf_U y y z) / (1 - gf_U x x z)" .
  qed

  have "?E x \<noteq> \<infinity>"
  proof
    assume "?E x = \<infinity>"
    have "((\<lambda>z. (1 - gf_U y y z) / (1 - gf_U x x z)) ---> 0) ?L"
    proof (rule lhopital_left)
      show "((\<lambda>z. 1 - gf_U y y z) ---> 0) ?L"
        using gf_U[of y] recurrent_iff_U_eq_1[of y] `recurrent y` by (auto intro!: tendsto_eq_intros)
      show "((\<lambda>z. 1 - gf_U x x z) ---> 0) ?L"
        using gf_U[of x] recurrent_iff_U_eq_1[of x] `recurrent x` by (auto intro!: tendsto_eq_intros)
      show "eventually (\<lambda>z. 1 - gf_U x x z \<noteq> 0) ?L"
        by (auto intro!: eventually_at_left_1 simp: gf_G_eq_gf_U(2) convergence_G_less_1)
      show "eventually (\<lambda>z. - gf_U' x x z \<noteq> 0) ?L"
        using gf_U'_pos[of _ x x] recurrent_iff_U_eq_1[of x] `recurrent x`
        by (auto intro!: eventually_at_left_1) (metis less_le)
      show "eventually (\<lambda>z. DERIV (\<lambda>xa. 1 - gf_U x x xa) z :> - gf_U' x x z) ?L"
        by (auto intro!: eventually_at_left_1 DERIV_intros DERIV_gf_U)
      show "eventually (\<lambda>z. DERIV (\<lambda>xa. 1 - gf_U y y xa) z :> - gf_U' y y z) ?L"
        by (auto intro!: eventually_at_left_1 DERIV_intros DERIV_gf_U)
      
      have "(gf_U' y y ---> ?E y) ?L"
        using `recurrent y` `y \<in> S` by (rule gf_U'_tendsto_integral_t)
      then have *: "(gf_U' y y ---> r) ?L"
        by (simp add: r)
      moreover
      have "(gf_U' x x ---> ?E x) ?L"
        using `recurrent x` `x \<in> S` by (rule gf_U'_tendsto_integral_t)
      then have "LIM z ?L. - gf_U' x x z :> at_bot"
        by (simp add: tendsto_PInfty_eq_at_top `?E x = \<infinity>` filterlim_uminus_at_top)
      then have "LIM z ?L. - gf_U' x x z :> at_infinity"
        by (rule filterlim_mono) (auto simp: at_bot_le_at_infinity)
      ultimately show "((\<lambda>z. - gf_U' y y z / - gf_U' x x z) ---> 0) ?L"
        by (intro tendsto_divide_0[where c="- r"] tendsto_intros)
    qed
    moreover
    have "((\<lambda>z. p x y n * p y x m * z^(n + m)) ---> p x y n * p y x m) ?L"
      by (auto intro!: tendsto_eq_intros)
    ultimately have "p x y n * p y x m \<le> 0"
      using le by (rule tendsto_le[OF trivial_limit_at_left_real])
    with `0 < p x y n` `0 < p y x m` show False
      by (auto simp add: mult_le_0_iff)
  qed
  with `recurrent x` show ?thesis
    by (simp add: pos_recurrent_def)
qed

lemma pos_recurrent_iffI_communicating:
  "(y, x) \<in> communicating \<Longrightarrow> pos_recurrent y \<longleftrightarrow> pos_recurrent x"
  using pos_recurrentI_communicating[of x y] pos_recurrentI_communicating[of y x]
  by (auto simp add: communicating_def)


subsection {* Stationary distribution *} text_raw {* \label{sec:stationary} *}

definition stationary_distribution :: "'s measure \<Rightarrow> bool" where
  "stationary_distribution N \<longleftrightarrow> N = point_measure S (\<lambda>x. \<integral>\<^sup>+s. emeasure (K s) {x} \<partial>N)"

definition "stat C = point_measure S (\<lambda>x. inverse (\<integral>\<^sup>+\<omega>. t x \<omega> \<partial>paths x) * indicator C x)"

lemma sets_stat[simp]: "sets (stat C) = Pow S"
  by (simp add: stat_def point_measure_def)

lemma space_stat[simp]: "space (stat C) = S"
  by (simp add: stat_def point_measure_def)

lemma stationary_distributionD_eq: "stationary_distribution N \<Longrightarrow>
   N = point_measure S (\<lambda>x. \<integral>\<^sup>+s. emeasure (K s) {x} \<partial>N)"
   unfolding stationary_distribution_def .

lemma stationary_distributionD_emeasure:
  assumes N: "stationary_distribution N" "prob_space N" and A: "A \<subseteq> S"
  shows "emeasure N A = (\<integral>\<^sup>+s. emeasure (K s) A \<partial>N)"
proof -
  interpret N: prob_space N by fact

  have sets_N: "sets N = Pow S"
    by (subst stationary_distributionD_eq[OF N(1)]) (simp add: sets_point_measure)
  
  have space_N: "space N = S"
    using sets_eq_imp_space_eq[of N "count_space S"] by (simp_all add: sets_N)

  then have [simp]: "\<And>f. f \<in> borel_measurable N"
    unfolding measurable_def by (simp add: sets_N space_N)

  have [simp]: "countable A" "\<And>x. x \<in> A \<Longrightarrow> x \<in> S"
    using A by (auto simp add: countable_subset)
  then have [simp]: "sets (count_space A \<Otimes>\<^sub>M N) = Pow (A \<times> S)"
    by (simp add: sets_N space_N sets_pair_countable)

  then have [simp]: "\<And>f. f \<in> borel_measurable (count_space A \<Otimes>\<^sub>M N)"
    using sets_eq_imp_space_eq[of "count_space A \<Otimes>\<^sub>M N" "count_space (A \<times> S)"]
    by (simp add: measurable_def)

  interpret A: sigma_finite_measure "count_space A"
    by (rule sigma_finite_measure_count_space) simp

  have "(\<integral>\<^sup>+ s. emeasure (K s) A \<partial>N) = (\<integral>\<^sup>+ s. \<integral>\<^sup>+a. emeasure (K s) {a} \<partial>count_space A \<partial>N)"
    by (intro positive_integral_cong emeasure_countable_singleton) (simp_all add: space_N)
  also have "\<dots> = (\<integral>\<^sup>+a. \<integral>\<^sup>+s. emeasure (K s) {a} \<partial>N \<partial>count_space A)"
    apply (rule pair_sigma_finite.Fubini')
    apply unfold_locales
    apply simp
    done
  also have "\<dots> = (\<integral>\<^sup>+a. emeasure N {a} \<partial>count_space A)"
    apply (intro positive_integral_cong)
    apply (subst (2) stationary_distributionD_eq[OF N(1)])
    apply (simp add: point_measure_def emeasure_density)
  proof -
    fix a assume "a \<in> A"
    have eq: "\<And>xa. (\<integral>\<^sup>+ s. emeasure (K s) {xa} \<partial>N) * indicator {a} xa = (\<integral>\<^sup>+ s. emeasure (K s) {a} \<partial>N) * indicator {a} xa"
      by (simp split: split_indicator)
    show "(\<integral>\<^sup>+ s. emeasure (K s) {a} \<partial>N) = (\<integral>\<^sup>+ xa. (\<integral>\<^sup>+ s. emeasure (K s) {xa} \<partial>N) * indicator {a} xa \<partial>count_space S)"
      unfolding eq using `a \<in> A`
      by (subst positive_integral_cmult) (simp_all add: positive_integral_positive one_ereal_def[symmetric])
  qed
  also have "\<dots> = emeasure N A"
    by (intro emeasure_countable_singleton[symmetric]) (simp_all add: sets_N)
  finally show ?thesis ..
qed

lemma stationary_distributionI:
  assumes "finite_measure N"
  assumes N[simp]: "sets N = Pow S"
  assumes le: "\<And>y. y \<in> S \<Longrightarrow> (\<integral>x. measure (K x) {y} \<partial>N) \<le> measure N {y}"
  shows "stationary_distribution N"
  unfolding stationary_distribution_def
proof (rule measure_eqI_countable[OF _ _ countable_space])
  fix a assume [simp]: "a \<in> S"
  interpret N: finite_measure N by fact
  interpret S: sigma_finite_measure "count_space S"
    using countable_space by (rule sigma_finite_measure_count_space)
  interpret NS: pair_sigma_finite N "count_space S" ..

  have space_N[simp]: "space N = S"
    using sets_eq_imp_space_eq[of N "count_space S"] by simp

  have [measurable]: "\<And>f. f \<in> borel_measurable N"
    unfolding measurable_def using  sets_eq_imp_space_eq[of N "count_space S"]
    by (auto simp add: N)

  have "sets (N \<Otimes>\<^sub>M count_space S) = Pow (S \<times> S)" 
    by (intro sets_pair_countable countable_space N) simp
  then have [measurable]: "\<And>f. f \<in> borel_measurable (N \<Otimes>\<^sub>M count_space S)" 
    by (auto simp add: measurable_def)

  have "(\<integral>\<^sup>+ t. ereal (\<integral>x. measure (K x) {t} \<partial>N) \<partial>count_space S)
    = (\<integral>\<^sup>+ t. (\<integral>\<^sup>+x. emeasure (K x) {t} \<partial>N) \<partial>count_space S)"
    apply (simp add: K.emeasure_eq_measure)
    apply (subst (2) positive_integral_eq_integral)
    apply (rule N.integrable_const_bound[where B=1])
    apply (simp_all add: measure_nonneg)
    done
  also have "\<dots> = (\<integral>\<^sup>+ x. (\<integral>\<^sup>+t. emeasure (K x) {t} \<partial>count_space S) \<partial>N)"
    by (rule NS.Fubini') simp
  also have "\<dots> = (\<integral>\<^sup>+ x. emeasure (K x) (space (K x)) \<partial>N)"
    apply (intro positive_integral_cong)
    apply simp
    apply (subst emeasure_countable_singleton[symmetric])
    apply simp_all
    done
  also have "\<dots> = emeasure N S"
    unfolding K.emeasure_space_1 by simp
  finally have comm: "(\<integral>\<^sup>+ t. ereal (\<integral>x. measure (K x) {t} \<partial>N) \<partial>count_space S) = ereal (measure N S)"
    by (simp add: N.emeasure_eq_measure)

  { assume "(\<integral>\<^sup>+ x. emeasure (K x) {a} \<partial>N) < emeasure N {a}" 
    then have "0 < ereal (measure N {a} - (\<integral>x. measure (K x) {a} \<partial>N))"
      apply (simp add: K.emeasure_eq_measure N.emeasure_eq_measure)
      apply (subst (asm) positive_integral_eq_integral)
      apply (rule N.integrable_const_bound[where B=1])
      apply (simp_all add: measure_nonneg)
      done
    also have "\<dots> = (\<integral>\<^sup>+t. ereal (measure N {a} - (\<integral>x. measure (K x) {a} \<partial>N)) * indicator {a} t \<partial>count_space S)"
      by (subst positive_integral_cmult_indicator) (simp_all add: le_diff_eq le)
    also have "\<dots> \<le> (\<integral>\<^sup>+t. ereal (measure N {t}) - ereal (\<integral>x. measure (K x) {t} \<partial>N) \<partial>count_space S)"
      by (intro positive_integral_mono) (auto split: split_indicator simp: le_diff_eq le)
    also have "\<dots> = (\<integral>\<^sup>+t. emeasure N {t} \<partial>count_space S) - emeasure N S"
      by (subst positive_integral_diff)
         (auto intro!: AE_I2 lebesgue_integral_nonneg measure_nonneg le simp: comm N.emeasure_eq_measure)
    also have "\<dots> = 0"
      by (subst emeasure_countable_singleton[symmetric]) (simp_all add: N.emeasure_eq_measure)
    finally have False by simp }
  note is_eq = this

  show "emeasure N {a} = emeasure (point_measure S (\<lambda>x. \<integral>\<^sup>+ s. emeasure (K s) {x} \<partial>N)) {a}"
    apply (simp add: emeasure_point_measure_finite2 positive_integral_positive)
    apply (intro antisym le `a \<in> S`)
    apply (rule ccontr)
    apply (simp add: not_le)
    apply (auto dest: is_eq)
    using le[OF `a \<in> S`]
    apply (simp add: K.emeasure_eq_measure N.emeasure_eq_measure)
    apply (subst positive_integral_eq_integral)
    apply (rule N.integrable_const_bound[where B=1])
    apply (simp_all add: measure_nonneg)
    done
qed (simp_all add: sets_point_measure)

lemma stationary_distribution_iterate:
  assumes N: "stationary_distribution N" and "finite_measure N" and [simp]: "y \<in> S"
  shows "measure N {y} = (\<integral>x. p x y n \<partial>N)"
proof -
  interpret N: finite_measure N by fact
  interpret S: sigma_finite_measure "count_space S"
    using countable_space by (rule sigma_finite_measure_count_space)
  interpret NS: pair_sigma_finite N "count_space S" ..

  from N have N_eq: "N = point_measure S (\<lambda>x. \<integral>\<^sup>+ s. emeasure (K s) {x} \<partial>N)"
    by (simp add: stationary_distribution_def)
  have sets_N[simp]: "sets N = Pow S"
    by (subst N_eq) (simp add: sets_point_measure)
  have space_N[simp]: "space N = S"
    using sets_eq_imp_space_eq[of N "count_space S"] by simp
  have [simp]: "\<And>f. f \<in> borel_measurable N"
    by (auto simp: measurable_def)

  have "sets (N \<Otimes>\<^sub>M count_space S) = Pow (S \<times> S)" 
    by (intro sets_pair_countable countable_space sets_N) simp
  then have [measurable]: "\<And>f. f \<in> borel_measurable (N \<Otimes>\<^sub>M count_space S)" 
    by (auto simp add: measurable_def)

  have integrable_K: "\<And>x. x \<in> S \<Longrightarrow> integrable N (\<lambda>s. Sigma_Algebra.measure (K s) {x})"
    by (rule N.integrable_const_bound[where B=1]) (simp_all add: measure_nonneg)

  have integrable_p: "\<And>n y. y \<in> S \<Longrightarrow> integrable N (\<lambda>x. p x y n)"
    by (rule N.integrable_const_bound[where B=1]) (simp_all add: p_nonneg p_le_1)

  have integrable_p_K: "\<And>n x y. x \<in> S \<Longrightarrow> y \<in> S \<Longrightarrow> integrable (K x) (\<lambda>x. p x y n)"
    by (rule K.integrable_const_bound[where B=1]) (simp_all add: p_nonneg p_le_1)

  note N_eq also have "point_measure S (\<lambda>x. \<integral>\<^sup>+ s. emeasure (K s) {x} \<partial>N) =
    point_measure S (\<lambda>x. ereal (\<integral>s. measure (K s) {x} \<partial>N))"
    unfolding point_measure_def
    by (intro density_cong)
       (auto simp: AE_count_space K.emeasure_eq_measure integrable_K positive_integral_eq_integral measure_nonneg)
  finally have N_eq': "N = point_measure S (\<lambda>x. ereal (\<integral>s. measure (K s) {x} \<partial>N))" .

  show ?thesis
  proof (induct n)
    case 0
    have "measure N {y} = (\<integral>x. indicator {y} x \<partial>N)"
      by (simp add: N.emeasure_eq_measure)
    also have "\<dots> = (\<integral>x. p x y 0 \<partial>N)"
      by (intro integral_cong) (simp add: p_0)
    finally show ?case .
  next
    case (Suc n)
    then have "Sigma_Algebra.measure N {y} = (\<integral>\<^sup>+ x. p x y n \<partial>N)"
      by (simp add: positive_integral_eq_integral integrable_p p_nonneg)
    also have "\<dots> = (\<integral>\<^sup>+s. (\<integral>\<^sup>+x. p x y n * emeasure (K s) {x} \<partial>count_space S) \<partial>N)"
      by (subst N_eq)
         (simp add: point_measure_def positive_integral_density positive_integral_positive
                       positive_integral_cmult[symmetric] p_nonneg NS.Fubini' ac_simps)
    also have "\<dots> = (\<integral>\<^sup>+s. (\<integral>\<^sup>+x. p x y n \<partial>K s) \<partial>N)"
      apply (intro positive_integral_cong)
      apply (subst (2) K_eq_density)
      apply (simp_all add: emeasure_nonneg positive_integral_density ac_simps)
      done
    also have "\<dots> = (\<integral>\<^sup>+s. p s y (Suc n) \<partial>N)"
      apply (intro positive_integral_cong)
      apply (subst p_Suc)
      apply (simp_all add: p_nonneg integrable_p_K positive_integral_eq_integral)
      done
    also have "\<dots> = (\<integral>s. p s y (Suc n) \<partial>N)"
      by (simp add: p_nonneg integrable_p positive_integral_eq_integral)
    finally show ?case by simp
  qed
qed

lemma stat_subprob:
  assumes C: "essential_class C" and pos: "\<forall>c\<in>C. pos_recurrent c"
  shows "emeasure (stat C) C \<le> 1"
proof -
  let ?E = "\<lambda>x. integral\<^sup>P (paths x) (t x)"
  let ?L = "at_left (1::real)"
  have "C \<subseteq> S"
    using C by (auto simp add: essential_class_def quotient_def communicating_in_S)
  from this countable_space have "countable C"
    by (rule countable_subset)
  from finite_sequence[OF this] guess A . note A = this
  from A `C \<subseteq> S` have "(\<lambda>n. emeasure (stat C) (A n)) ----> emeasure (stat C) (\<Union>i. A i)"
    by (intro Lim_emeasure_incseq) auto
  then have "emeasure (stat C) (\<Union>i. A i) \<le> 1"
  proof (rule LIMSEQ_le[OF _ tendsto_const], intro exI allI impI)
    fix n
    from A(1,3) `C \<subseteq> S` have A_n: "finite (A n)" "A n \<subseteq> S"
      by auto

    from C have "C \<noteq> {}"
      by (simp add: essential_class_def not_empty_irreducible)
    then obtain x where "x \<in> C" by auto

    have "((\<lambda>z. (\<Sum>y\<in>A n. gf_F x y z * ((1 - z) / (1 - gf_U y y z)))) ---> (\<Sum>y\<in>A n. F x y * real (1 / ?E y))) ?L"
    proof (intro tendsto_intros gf_F, rule lhopital_left)
      fix y assume "y \<in> A n"
      with `A n \<subseteq> C` `C \<subseteq> S` have "y \<in> S" "y \<in> C"
        by auto
      show "(op - 1 ---> 0) ?L"
        by (intro tendsto_eq_intros) simp_all
      have "recurrent y"
        using pos[THEN bspec, OF `y\<in>C`] by (simp add: pos_recurrent_def)
      then have "U y y = 1"
        by (simp add: recurrent_iff_U_eq_1)
        
      show "((\<lambda>x. 1 - gf_U y y x) ---> 0) ?L"
        using gf_U[of y y] `U y y = 1` by (intro tendsto_eq_intros) auto
      show "eventually (\<lambda>x. 1 - gf_U y y x \<noteq> 0) ?L"
        using gf_G_eq_gf_U(2)[OF `y \<in> S` convergence_G_less_1, where 'z=real] by (auto intro!: eventually_at_left_1)
      have "eventually (\<lambda>x. 0 < gf_U' y y x) ?L"
        by (intro eventually_at_left_1 gf_U'_pos) (simp_all add: `U y y = 1`)
      then show "eventually (\<lambda>x. - gf_U' y y x \<noteq> 0) ?L"
        by eventually_elim simp
      show "eventually (\<lambda>x. DERIV (\<lambda>x. 1 - gf_U y y x) x :> - gf_U' y y x) ?L"
        by (auto intro!: eventually_at_left_1 DERIV_intros DERIV_gf_U)
      show "eventually (\<lambda>x. DERIV (op - 1) x :> - 1) ?L"
        by (auto intro!: eventually_at_left_1 DERIV_intros)
      show "((\<lambda>x. - 1 / - gf_U' y y x) ---> real (1 / integral\<^sup>P (paths y) (t y))) ?L"
        using `recurrent y` `y \<in> S` by (rule inverse_gf_U'_tendsto)
    qed
    also have "(\<Sum>y\<in>A n. F x y * real (1 / integral\<^sup>P (paths y) (t y))) = (\<Sum>y\<in>A n. real (1 / integral\<^sup>P (paths y) (t y)))"
    proof (intro setsum_cong refl)
      fix y assume "y \<in> A n"
      with `A n \<subseteq> C` have "y \<in> C" by auto
      with `x \<in> C` have "(x, y) \<in> communicating"
        by (rule essential_classD3[OF C])
      with `y\<in>C` have "recurrent y" "accessible y x"
        using pos[THEN bspec, of y] by (auto simp add: pos_recurrent_def communicating_def)
      then have "U x y = 1"
        by (rule recurrent_accessible)
      with F_le_1[of x y] U_le_F[of x y] have "F x y = 1" by simp
      then show "F x y * real (1 / integral\<^sup>P (paths y) (t y)) = real (1 / integral\<^sup>P (paths y) (t y))"
        by simp
    qed
    finally have le: "(\<Sum>y\<in>A n. real (1 / ?E y)) \<le> 1"
    proof (rule tendsto_le[OF trivial_limit_at_left_real tendsto_const], intro eventually_at_left_1)
      fix z :: real assume z: "0 < z" "z < 1"
      with `C \<subseteq> S` `x \<in> C` have "norm z < 1" "x \<in> S"
        by auto
      then have conv: "\<And>x y. convergence_G x y z"
        by (simp add: convergence_G_less_1)
      have "(\<Sum>y\<in>A n. gf_F x y z / (1 - gf_U y y z)) = (\<Sum>y\<in>A n. gf_G x y z)"
        using `norm z < 1` `x \<in> S`
        apply (auto intro!: setsum_cong simp: gf_G_eq_gf_F)
        using `A n \<subseteq> S`
        apply (subst gf_G_eq_gf_U(1)[OF _ conv])
        apply auto
        done
      also have "\<dots> = (\<Sum>y\<in>A n. \<Sum>n. p x y n * z^n)"
        by (simp add: gf_G_def)
      also have "\<dots>  = (\<Sum>i. \<Sum>y\<in>A n. p x y i *\<^sub>R z^i)"
        by (subst suminf_setsum[OF convergence_G[OF conv]]) simp
      also have "\<dots>  \<le> (\<Sum>i. z^i)"
      proof (intro summable_le summable_setsum convergence_G conv summable_geometric allI)
        fix l
        have "(\<Sum>y\<in>A n. p x y l *\<^sub>R z ^ l) = (\<Sum>y\<in>A n. p x y l) * z ^ l"
          by (simp add: setsum_left_distrib)
        also have "\<dots> \<le> z ^ l"
        proof (intro mult_left_le_one_le)
          have "(\<Sum>y\<in>A n. p x y l) = \<P>(\<omega> in paths x. nat_case x \<omega> l \<in> A n)"
            unfolding p_def using `finite (A n)` `A n \<subseteq> S`
            by (subst finite_measure_finite_Union[symmetric])
               (auto simp: disjoint_family_on_def intro!: arg_cong2[where f=measure])
          then show "(\<Sum>y\<in>A n. p x y l) \<le> 1"
            by simp
        qed (insert z, auto simp: setsum_nonneg p_nonneg)
        finally show "(\<Sum>y\<in>A n. p x y l *\<^sub>R z ^ l) \<le> z ^ l" .
      qed fact
      also have "\<dots> = 1 / (1 - z)"
        using sums_unique[OF geometric_sums, OF `norm z < 1`] ..
      finally have "(\<Sum>y\<in>A n. gf_F x y z / (1 - gf_U y y z)) \<le> 1 / (1 - z)" .
      then have "(\<Sum>y\<in>A n. gf_F x y z / (1 - gf_U y y z)) * (1 - z) \<le> 1"
        using z by (simp add: field_simps)
      then have "(\<Sum>y\<in>A n. gf_F x y z / (1 - gf_U y y z) * (1 - z)) \<le> 1"
        by (simp add: setsum_left_distrib)
      then show "(\<Sum>y\<in>A n. gf_F x y z * ((1 - z) / (1 - gf_U y y z))) \<le> 1"
        by simp
    qed

    from A_n have "emeasure (stat C) (A n) = (\<Sum>y\<in>A n. emeasure (stat C) {y})"
      by (intro emeasure_eq_setsum_singleton) (simp_all, blast) 
    also have "\<dots> = (\<Sum>y\<in>A n. inverse (?E y))"
      unfolding stat_def using `A n \<subseteq> S` A(1)[of n]
      apply (intro setsum_cong refl)
      apply (subst emeasure_point_measure_finite2)
      apply (auto simp: inverse_nonneg_iff positive_integral_positive)
      done
    also have "\<dots> = ereal (\<Sum>y\<in>A n. real (1 / ?E y))"
      apply (subst setsum_ereal[symmetric])
    proof (intro setsum_cong refl)
      fix y assume "y \<in> A n"
      with `A n \<subseteq> C` pos have "pos_recurrent y"
        by auto
      with one_le_integral_t[of y] obtain r where "?E y = ereal r" "1 \<le> ?E y"
        by (cases "?E y") (auto simp: pos_recurrent_def)
      then show "inverse (?E y) = ereal (real (1 / ?E y))"
        by (simp add: one_ereal_def inverse_eq_divide)
    qed
    also have "\<dots> \<le> 1"
      using le by (simp add: one_ereal_def)
    finally show "emeasure (stat C) (A n) \<le> 1" .
  qed
  with A show ?thesis 
    by simp
qed

lemma emeasure_stat_not_C:
  assumes "y \<notin> C"
  shows "emeasure (stat C) {y} = 0"
proof cases
  assume "y \<in> S"
  show ?thesis
    unfolding stat_def using `y \<in> S` `y \<notin> C`
    by (subst emeasure_point_measure_finite2) auto
next
  assume "y \<notin> S"
  then have "{y} \<notin> sets (stat C)" by simp
  then show ?thesis
    by (rule emeasure_notin_sets)
qed

subsubsection {* @{const stationary_distribution} implies @{const pos_recurrent} *} text_raw {* \label{sec:sdimppr} *}

lemma stationary_distributionD:
  assumes C: "essential_class C"
  assumes N: "stationary_distribution N" and "prob_space N" and null: "measure N (S - C) = 0"
  shows "\<forall>x\<in>C. pos_recurrent x" "N = stat C"
proof -
  interpret N: prob_space N by fact
  have N_finite: "finite_measure N" by unfold_locales
  from N have N_eq: "N = point_measure S (\<lambda>x. \<integral>\<^sup>+ s. emeasure (K s) {x} \<partial>N)"
    by (simp add: stationary_distribution_def)
  have sets_N[simp]: "sets N = Pow S"
    by (subst N_eq) (simp add: sets_point_measure)
  have space_N[simp]: "space N = S"
    using sets_eq_imp_space_eq[of N "count_space S"] by simp
  have [simp]: "\<And>f. f \<in> borel_measurable N"
    by (auto simp: measurable_def)
  have "C \<subseteq> S"
    using C by (auto simp add: essential_class_def quotient_def communicating_in_S)
  then have measure_C: "measure N C = 1"
    using N.finite_measure_Diff[of S C] N.prob_space null by simp
  from null `C \<subseteq> S` have ae_C: "AE x in N. x \<in> C"
    using N.prob_eq_0[of "S - C"] by (auto simp: Diff_subset)

  have integrable_K: "\<And>x. x \<in> S \<Longrightarrow> integrable N (\<lambda>s. Sigma_Algebra.measure (K s) {x})"
    by (rule N.integrable_const_bound[where B=1]) (simp_all add: measure_nonneg)

  note N_eq also have "point_measure S (\<lambda>x. \<integral>\<^sup>+ s. emeasure (K s) {x} \<partial>N) =
    point_measure S (\<lambda>x. ereal (\<integral>s. measure (K s) {x} \<partial>N))"
    unfolding point_measure_def
    by (intro density_cong)
       (auto simp: AE_count_space K.emeasure_eq_measure integrable_K positive_integral_eq_integral measure_nonneg)
  finally have N_eq': "N = point_measure S (\<lambda>x. ereal (\<integral>s. measure (K s) {x} \<partial>N))" .

  have integral_N_finite: "\<And>A f. finite A \<Longrightarrow> A \<subseteq> S \<Longrightarrow> (\<integral>x. f x * indicator A x \<partial>N) = (\<Sum>a\<in>A. (\<integral>s. measure (K s) {a} \<partial>N) * f a)"
    apply (subst N_eq')
    apply (subst lebesgue_integral_point_measure_finite_indicator)
    apply (simp_all add: measure_nonneg lebesgue_integral_nonneg)
    done

  have integrable_p: "\<And>n y. y \<in> S \<Longrightarrow> integrable N (\<lambda>x. p x y n)"
    by (rule N.integrable_const_bound[where B=1]) (simp_all add: p_nonneg p_le_1)

  { fix e :: real assume "0 < e"
    have "\<exists>A\<subseteq>C. finite A \<and> 1 - e < measure N A"
    proof (rule ccontr)
      assume contr: "\<not> (\<exists>A \<subseteq> C. finite A \<and> 1 - e < measure N A)"
      have "countable C"
        using `C \<subseteq> S` countable_space by (rule countable_subset)
      from finite_sequence[OF this] guess F . note F = this
      with `C \<subseteq> S` have *: "(\<lambda>n. measure N (F n)) ----> measure N (\<Union>i. F i)"
        by (intro N.finite_Lim_measure_incseq) auto
      with F contr have "measure N (\<Union>i. F i) \<le> 1 - e"
        by (intro LIMSEQ_le[OF * tendsto_const]) (auto simp: not_less)
      with F `measure N C = 1` `0 < e` show False
        by simp
    qed
    then obtain A where "A \<subseteq> C" "finite A" and e: "1 - e < measure N A" by auto
    
    { fix y n assume "y \<in> C"
      with `C \<subseteq> S` have "y \<in> S" by auto
      with N N_finite have "measure N {y} = (\<integral>x. p x y n \<partial>N)"
        by (rule stationary_distribution_iterate)
      also have "\<dots> \<le> (\<integral>x. p x y n * indicator A x + indicator (C - A) x \<partial>N)"
        using ae_C `C \<subseteq> S` `A \<subseteq> C`
        by (intro integral_mono_AE)
           (auto elim!: eventually_elim1 
                 intro!: integral_add integral_indicator p_le_1 integrable_mult_indicator
                 split: split_indicator simp: integrable_p `y\<in>S`)
      also have "\<dots> = (\<integral>x. p x y n * indicator A x \<partial>N) + measure N (C - A)"
        using ae_C `C \<subseteq> S` `A \<subseteq> C`
        apply (subst integral_add)
        apply (auto elim!: eventually_elim1 
                    intro!: integral_add integral_indicator p_le_1 integrable_mult_indicator
                    split: split_indicator simp: integrable_p `y\<in>S`)
        apply (subst integral_indicator)
        apply (auto simp add: N.emeasure_eq_measure)
        done
      also have "\<dots> \<le> (\<integral>x. p x y n * indicator A x \<partial>N) + e"
        using e `C \<subseteq> S` `A \<subseteq> C` measure_C by (simp add: N.finite_measure_Diff)
      finally have "measure N {y} \<le> (\<integral>x. p x y n * indicator A x \<partial>N) + e" .
      then have "emeasure N {y} \<le> ereal (\<integral>x. p x y n * indicator A x \<partial>N) + e"
        by (simp add: N.emeasure_eq_measure)
      also have "\<dots> = (\<integral>\<^sup>+x. ereal (p x y n) * indicator A x \<partial>N) + e"
        by (subst positive_integral_eq_integral[symmetric])
           (auto intro!: N.integrable_const_bound[where B=1]
                 simp: abs_mult mult_nonneg_nonneg p_nonneg p_le_1 mult_le_one times_ereal.simps[symmetric] ereal_indicator
                 simp del: times_ereal.simps)
      finally have "emeasure N {y} \<le> (\<integral>\<^sup>+x. ereal (p x y n) * indicator A x \<partial>N) + e" . }
    note v_le = this

    { fix y and z :: real assume y: "y \<in> C" and z: "0 < z" "z < 1"
      have summable_int_p: "summable (\<lambda>n. (\<integral> x. p x y n * indicator A x \<partial>N) * (1 - z) * z ^ n)"
        using `y\<in>C` z `C \<subseteq> S` `A \<subseteq> C`
        by (auto intro!: summable_comparison_test[OF _ summable_mult[OF summable_geometric[of z], of 1]] exI[of _ 0] mult_le_one
                            N.integral_le_const integrable_mult_indicator integrable_p AE_I2 p_le_1
                    simp: abs_mult lebesgue_integral_nonneg p_nonneg mult_nonneg_nonneg)

      from y z have sums_y: "(\<lambda>n. measure N {y} * (1 - z) * z ^ n) sums measure N {y}"
        using sums_mult[OF geometric_sums[of z], of "measure N {y} * (1 - z)"] by simp
      then have "emeasure N {y} = ereal (\<Sum>n. (measure N {y} * (1 - z)) * z ^ n)"
        by (simp add: sums_unique N.emeasure_eq_measure)
      also have "\<dots> = (\<Sum>n. emeasure N {y} * (1 - z) * z ^ n)"
        using z
        by (subst suminf_ereal[symmetric]) 
           (auto intro!: mult_nonneg_nonneg measure_nonneg suminf_ereal_finite
              summable_mult[OF summable_geometric[of z]] simp: N.emeasure_eq_measure)
      also have "\<dots> \<le> (\<Sum>n. ((\<integral>\<^sup>+x. ereal (p x y n) * indicator A x \<partial>N) + e) * (1 - z) * z ^ n)"
        using `y\<in>C` z `C \<subseteq> S` `A \<subseteq> C`
        by (intro suminf_le_pos ereal_mult_right_mono v_le)
           (auto simp: N.emeasure_eq_measure mult_nonneg_nonneg measure_nonneg)
      also have "\<dots> = (\<Sum>n. (\<integral>\<^sup>+x. ereal (p x y n) * indicator A x \<partial>N) * (1 - z) * z ^ n) + e"
        using `0 < e` z sums_mult[OF geometric_sums[of z], of "e * (1 - z)"]
        by (simp add: ereal_left_distrib positive_integral_positive mult_nonneg_nonneg suminf_add_ereal sums_suminf_ereal)
      also have "\<dots> = (\<Sum>n. (1 - z) * ((\<integral>\<^sup>+x. ereal (p x y n) * indicator A x \<partial>N) * z ^ n)) + e"
        by (simp add: ac_simps)
      also have "\<dots> = (1 - z) * (\<Sum>n. ((\<integral>\<^sup>+x. ereal (p x y n) * indicator A x \<partial>N) * z ^ n)) + e"
        using z by (subst suminf_cmult_ereal) (simp_all add: ereal_0_le_mult positive_integral_positive )
      also have "(\<Sum>n. ((\<integral>\<^sup>+x. ereal (p x y n) * indicator A x \<partial>N) * z ^ n)) =
          (\<Sum>n. (\<integral>\<^sup>+x. ereal (p x y n * z ^ n) * indicator A x \<partial>N))"
        using z by (simp add: ac_simps positive_integral_cmult[symmetric])
      also have "\<dots> = (\<integral>\<^sup>+x. ereal (gf_G x y z) * indicator A x \<partial>N)"
        using z
        apply (subst positive_integral_suminf[symmetric])
        apply (auto simp add: ereal_0_le_mult mult_nonneg_nonneg p_nonneg gf_G_def
                    intro!: positive_integral_cong suminf_ereal' split: split_indicator)
        apply (intro summable_comparison_test[OF _ summable_mult[OF summable_geometric[of z], of 1]] impI)
        apply (simp_all add: abs_mult p_nonneg p_le_1 mult_le_one power_le_one split: split_indicator)
        done
      also have "\<dots> = (\<integral>\<^sup>+x. ereal (gf_F x y z * gf_G y y z) * indicator A x \<partial>N)"
        using z by (intro positive_integral_cong) (simp add:  gf_G_eq_gf_F[symmetric])
      also have "\<dots> = gf_G y y z * (\<integral>\<^sup>+x. ereal (gf_F x y z) * indicator A x \<partial>N)"
        using z by (subst positive_integral_cmult[symmetric]) (simp_all add: gf_G_nonneg ac_simps)
      also have "\<dots> = (1 / (1 - gf_U y y z)) * (\<integral>\<^sup>+x. ereal (gf_F x y z) * indicator A x \<partial>N)"
        using z `y \<in> C` `C \<subseteq> S` by (subst gf_G_eq_gf_U) (auto intro!: convergence_G_less_1)
      finally have "emeasure N {y} \<le> ereal ((1 - z) / (1 - gf_U y y z)) * (\<integral>\<^sup>+x. gf_F x y z * indicator A x \<partial>N) + e"
        by (subst (asm) mult_assoc[symmetric]) (simp add: ereal_indicator[symmetric])
      then have "measure N {y} \<le> (1 - z) / (1 - gf_U y y z) * (\<integral>x. gf_F x y z * indicator A x \<partial>N) + e"
        using z
        by (subst (asm) positive_integral_eq_integral[OF N.integrable_const_bound[where B=1]])
           (auto simp: gf_F_nonneg mult_nonneg_nonneg N.emeasure_eq_measure gf_F_le_1 mult_le_one) }
    then have "\<exists>A \<subseteq> C. finite A \<and> (\<forall>y\<in>C. \<forall>z. 0 < z \<longrightarrow> z < 1 \<longrightarrow> measure N {y} \<le> (1 - z) / (1 - gf_U y y z) * (\<integral>x. gf_F x y z * indicator A x \<partial>N) + e)"
      using `A \<subseteq> C` `finite A` by auto }
  note eps = this

  { fix y A assume "y \<in> C" "finite A" "A \<subseteq> C"
    with `C \<subseteq> S` have "((\<lambda>z. \<integral>x. gf_F x y z * indicator A x \<partial>N) ---> \<integral>x. F x y * indicator A x \<partial>N) (at_left 1)"
      by (subst (1 2) integral_N_finite) (auto intro!: tendsto_intros gf_F) }
  note int_gf_F = this

  have all_recurrent: "\<forall>y\<in>C. recurrent y"
  proof (rule ccontr)
    assume "\<not> (\<forall>y\<in>C. recurrent y)"
    then obtain x where "x \<in> C" "\<not> recurrent x" by auto
    then have transient: "\<And>x. x \<in> C \<Longrightarrow> \<not> recurrent x"
      using C by (auto simp: essential_class_def recurrent_iffI_communicating[symmetric] elim!: quotientE)
       
    { fix y assume "y \<in> C"
      with transient have "U y y < 1"
        by (metis recurrent_iff_U_eq_1 U_cases)
      have "measure N {y} \<le> 0"
      proof (rule dense_ge)
        fix e :: real assume "0 < e"
        from eps[OF this] `y \<in> C` obtain A where 
          A: "finite A" "A \<subseteq> C" and
          le: "\<And>z. 0 < z \<Longrightarrow> z < 1 \<Longrightarrow> measure N {y} \<le> (1 - z) / (1 - gf_U y y z) * (\<integral>x. gf_F x y z * indicator A x \<partial>N) + e"
          by auto
        have "((\<lambda>z. (1 - z) / (1 - gf_U y y z) * (\<integral>x. gf_F x y z * indicator A x \<partial>N) + e) ---> 
          (1 - 1) / (1 - U y y) * (\<integral>x. F x y * indicator A x \<partial>N) + e) (at_left (1::real))"
          using A `U y y < 1` `y \<in> C` by (intro tendsto_intros gf_U int_gf_F) auto
        then have 1: "((\<lambda>z. (1 - z) / (1 - gf_U y y z) * (\<integral>x. gf_F x y z * indicator A x \<partial>N) + e) ---> e) (at_left (1::real))"
          by simp
        with le show "measure N {y} \<le> e"
          by (intro tendsto_le[OF trivial_limit_at_left_real _ tendsto_const])
             (auto simp: eventually_at_left_1)
      qed
      then have "measure N {y} = 0"
        by (intro antisym measure_nonneg) }
    then have "emeasure N C = 0"
      using `C \<subseteq> S` countable_subset[OF `C \<subseteq> S` countable_space]
      by (subst emeasure_countable_singleton) (auto simp: N.emeasure_eq_measure positive_integral_0_iff_AE)
    then show False
      using `measure N C = 1` by (simp add: N.emeasure_eq_measure)
  qed
  then have "\<And>x. x \<in> C \<Longrightarrow> U x x = 1"
    by (metis recurrent_iff_U_eq_1)
  
  { fix y assume "y \<in> C"
    then have "y \<in> S" "U y y = 1" "recurrent y"
      using `y \<in> C \<Longrightarrow> U y y = 1` `C \<subseteq> S` all_recurrent by auto
    have "measure N {y} \<le> real (1 / integral\<^sup>P (paths y) (t y))"
    proof (rule field_le_epsilon)
      fix e :: real assume "0 < e"
      from eps[OF `0 < e`] `y \<in> C` obtain A where 
        A: "finite A" "A \<subseteq> C" and
        le: "\<And>z. 0 < z \<Longrightarrow> z < 1 \<Longrightarrow> measure N {y} \<le> (1 - z) / (1 - gf_U y y z) * (\<integral>x. gf_F x y z * indicator A x \<partial>N) + e"
        by auto
      let ?L = "at_left (1::real)"
      have "((\<lambda>z. (1 - z) / (1 - gf_U y y z) * (\<integral>x. gf_F x y z * indicator A x \<partial>N) + e) ---> 
          real (1 / integral\<^sup>P (paths y) (t y)) * (\<integral>x. F x y * indicator A x \<partial>N) + e) ?L"
      proof (intro tendsto_add tendsto_const tendsto_mult int_gf_F,
             rule lhopital_left[where f'="\<lambda>x. - 1" and g'="\<lambda>z. - gf_U' y y z"])
        show "(op - 1 ---> 0) ?L" "((\<lambda>x. 1 - gf_U y y x) ---> 0) ?L"
          using gf_U[of y y] by (auto intro!: tendsto_eq_intros simp: `U y y = 1`)
        show "y \<in> C" "finite A" "A \<subseteq> C" by fact+
        show "eventually (\<lambda>x. 1 - gf_U y y x \<noteq> 0) ?L"
          using gf_G_eq_gf_U(2)[OF `y \<in> S` convergence_G_less_1, where 'z=real] by (auto intro!: eventually_at_left_1)
        show "((\<lambda>x. - 1 / - gf_U' y y x) ---> real (1 / integral\<^sup>P (paths y) (t y))) ?L"
          using `recurrent y` `y \<in> S` by (rule inverse_gf_U'_tendsto)
        have "eventually (\<lambda>x. 0 < gf_U' y y x) ?L"
          by (intro eventually_at_left_1 gf_U'_pos) (simp_all add: `U y y = 1`)
        then show "eventually (\<lambda>x. - gf_U' y y x \<noteq> 0) ?L"
          by eventually_elim simp
        show "eventually (\<lambda>x. DERIV (\<lambda>x. 1 - gf_U y y x) x :> - gf_U' y y x) ?L"
          by (auto intro!: eventually_at_left_1 DERIV_intros DERIV_gf_U)
        show "eventually (\<lambda>x. DERIV (op - 1) x :> - 1) ?L"
          by (auto intro!: eventually_at_left_1 DERIV_intros)
      qed
      then have "measure N {y} \<le> real (1 / integral\<^sup>P (paths y) (t y)) * (\<integral>x. F x y * indicator A x \<partial>N) + e"
        by (rule tendsto_le[OF trivial_limit_at_left_real _ tendsto_const]) (intro eventually_at_left_1 le)
      then have "measure N {y} - e \<le> real (1 / integral\<^sup>P (paths y) (t y)) * (\<integral>x. F x y * indicator A x \<partial>N)"
        by simp
      also have "\<dots> \<le> real (1 / integral\<^sup>P (paths y) (t y))"
        using A `C \<subseteq> S`
        by (intro mult_left_le real_of_ereal_pos zero_le_divide_ereal positive_integral_positive
                  N.integral_le_const N.integrable_const_bound[where B=1])
           (auto simp: F_nonneg mult_nonneg_nonneg mult_le_one F_le_1)
      finally show "measure N {y} \<le> real (1 / integral\<^sup>P (paths y) (t y)) + e"
        by simp
    qed }
  note measure_y_le = this

  show pos: "\<forall>y\<in>C. pos_recurrent y"
  proof (rule ccontr)
    assume "\<not> (\<forall>y\<in>C. pos_recurrent y)"
    then obtain x where x: "x \<in> C" "\<not> pos_recurrent x" by auto
    { fix y assume "y \<in> C"
      with x have "\<not> pos_recurrent y"
        using C by (auto simp: essential_class_def pos_recurrent_iffI_communicating[symmetric] elim!: quotientE)
      with all_recurrent `y \<in> C` have "real (1 / (\<integral>\<^sup>+\<omega>. t y \<omega>\<partial>paths y)) = 0"
        by (simp add: pos_recurrent_def)
      with measure_y_le[OF `y \<in> C`] have "measure N {y} = 0"
        by (auto intro!: antisym simp: measure_nonneg pos_recurrent_def) }
    then have "emeasure N C = 0"
      using `C \<subseteq> S` countable_subset[OF `C \<subseteq> S` countable_space]
      by (subst emeasure_countable_singleton) (auto simp: N.emeasure_eq_measure positive_integral_0_iff_AE)
    then show False
      using `measure N C = 1` by (simp add: N.emeasure_eq_measure)
  qed

  let ?E = "\<lambda>x. integral\<^sup>P (paths x) (t x)"
  { fix A assume "A \<subseteq> S"
    from this countable_space have [simp]: "countable A" by (rule countable_subset)
    have "emeasure N A = (\<integral>\<^sup>+x. emeasure N {x} \<partial>count_space A)"
      using `A \<subseteq> S` by (intro emeasure_countable_singleton) auto
    also have "\<dots> \<le> (\<integral>\<^sup>+x. emeasure (stat C) {x} \<partial>count_space A)"
    proof (intro positive_integral_mono)
      fix y assume "y \<in> space (count_space A)"
      with `A \<subseteq> S` have "y \<in> S" by auto
      show "emeasure N {y} \<le> emeasure (stat C) {y}"
      proof cases
        assume "y \<in> C"
        with pos have "pos_recurrent y"
          by auto
        with one_le_integral_t[of y] obtain r where r: "?E y = ereal r" "1 \<le> ?E y"
          by (cases "?E y") (auto simp: pos_recurrent_def)
    
        from measure_y_le[OF `y \<in> C`]
        have "emeasure N {y} \<le> ereal (real (1 / integral\<^sup>P (paths y) (t y)))"
          by (simp add: N.emeasure_eq_measure)
        also have "\<dots> = emeasure (stat C) {y}"
          unfolding stat_def using `C \<subseteq> S` `y \<in> C` r
          by (subst emeasure_point_measure_finite2)
             (auto simp: inverse_nonneg_iff positive_integral_positive divide_ereal_def one_ereal_def)
        finally show "emeasure N {y} \<le> emeasure (stat C) {y}"
          by simp
      next
        assume "y \<notin> C" 
        with `y \<in> S` ae_C have "emeasure N {y} = 0"
          by (subst AE_iff_measurable[symmetric, where P="\<lambda>x. x \<noteq> y"]) (auto elim!: eventually_elim1)
        moreover have "emeasure (stat C) {y} = 0"
          using emeasure_stat_not_C[OF `y \<notin> C`] .
        ultimately show ?thesis by simp
      qed
    qed
    also have "\<dots> = emeasure (stat C) A"
      using `A \<subseteq> S` by (intro emeasure_countable_singleton[symmetric]) auto
    finally have "emeasure N A \<le> emeasure (stat C) A" . }
  note N_le_C = this

  from stat_subprob[OF C pos] N_le_C[OF `C \<subseteq> S`] `measure N C = 1`
  have "emeasure (stat C) C = 1"
    by (auto simp add: N.emeasure_eq_measure one_ereal_def)
  moreover have "emeasure (stat C) (S - C) = 0"
    using `C \<subseteq> S`
    by (subst emeasure_countable_singleton)
       (auto simp: sets_point_measure emeasure_stat_not_C intro!: positive_integral_0_iff_AE[THEN iffD2])
  ultimately have "emeasure (stat C) (space (stat C)) = 1"
    using plus_emeasure[of C "stat C" "S - C"] `C \<subseteq> S` by (simp add: Diff_subset Un_absorb1)
  interpret stat: prob_space "stat C"    
    by default fact
  
  show "N = stat C"
  proof (rule measure_eqI_countable)
    show "sets N = Pow S" "sets (stat C) = Pow S" "countable S"
      by auto
    
    { assume "\<exists>x\<in>S. emeasure N {x} \<noteq> emeasure (stat C) {x}"
      then obtain x where [simp]: "x \<in> S" and "emeasure N {x} \<noteq> emeasure (stat C) {x}" by auto
      with N_le_C[of "{x}"] have x: "emeasure N {x} < emeasure (stat C) {x}"
        by auto
      have "1 = emeasure N {x} + emeasure N (S - {x})"
        using N.emeasure_space_1 by (subst plus_emeasure) (auto simp: insert_absorb)
      also have "\<dots> < emeasure (stat C) {x} + emeasure (stat C) (S - {x})"
        using x N_le_C[of "S - {x}"] by (simp add: stat.emeasure_eq_measure N.emeasure_eq_measure Diff_subset)
      also have "\<dots> = 1"
        using stat.emeasure_space_1 by (subst plus_emeasure) (auto simp: insert_absorb)
      finally have False by simp }
    then show "!!x. x \<in> S \<Longrightarrow> emeasure N {x} = emeasure (stat C) {x}" by metis
  qed
qed

lemma stationary_distribution_imp_int_t:
  assumes C: "essential_class C" "stationary_distribution N" "prob_space N" "measure N (S - C) = 0"
  assumes x: "x \<in> C" shows "(\<integral>\<^sup>+\<omega>. t x \<omega> \<partial>paths x) = 1 / measure N {x}"
proof -
  have "C \<subseteq> S"
    using C by (auto simp: essential_class_def C_subset_S)
  with x have [simp]: "x \<in> S" by auto
  from stationary_distributionD[OF C] have "N = stat C" and *: "\<forall>x\<in>C. pos_recurrent x" by auto
  show ?thesis
    unfolding `N = stat C` stat_def
    using *[THEN bspec, OF x] x
    apply (cases "(\<integral>\<^sup>+\<omega>. t x \<omega> \<partial>paths x)")
    apply (auto simp add: measure_def emeasure_point_measure_finite2 inverse_nonneg_iff
                     mult_nonneg_nonneg positive_integral_positive pos_recurrent_def
                     divide_inverse)
    done
qed

subsection {* Aperiodic classes *} text_raw {* \label{sec:aperiodic} *}

definition "period_set x = {i. 0 < i \<and> 0 < p x x i }"

definition "period C = (SOME d. \<forall>x\<in>C. d = Gcd (period_set x))"

definition "aperiodic C \<longleftrightarrow> C \<in> S // communicating \<and> period C = 1"

definition "not_ephemeral C \<longleftrightarrow> C \<in> S // communicating \<and> \<not> (\<exists>x. C = {x} \<and> p x x 1 = 0)"

end

inductive_set monoid_closure :: "nat set \<Rightarrow> int set" for S where
  base: "s \<in> S \<Longrightarrow> int s \<in> monoid_closure S"
| plus: "s \<in> monoid_closure S \<Longrightarrow> t \<in> monoid_closure S \<Longrightarrow> s + t \<in> monoid_closure S"
| diff: "s \<in> monoid_closure S \<Longrightarrow> t \<in> monoid_closure S \<Longrightarrow> s - t \<in> monoid_closure S"

lemma Gcd_nonneg: "0 \<le> Gcd (S :: int set)"
  by (simp add: Gcd_int_def)

lemma Gcd_eq_Gcd_monoid_closure:
  fixes S :: "nat set"
  shows "Gcd S = Gcd (monoid_closure S)"
  apply (simp add: Gcd_int_def)
proof (safe intro!: dvd.antisym gcd_lcm_complete_lattice_nat.Inf_greatest)
  fix s assume "s \<in> monoid_closure S"
  then show "Gcd S dvd nat \<bar>s\<bar>"
  proof induct
    case (plus s t)
    then show ?case
      by (cases s t rule: int_cases'[case_product int_cases'])
         (simp_all add: uminus_add_conv_diff nat_abs_int_diff nat_int_add)
  next
    case (diff s t)
    then show ?case
      by (cases s t rule: int_cases'[case_product int_cases'])
         (simp_all add: uminus_add_conv_diff nat_abs_int_diff nat_int_add)
  qed simp
next
  fix s assume "s \<in> S"
  then show "Gcd ((\<lambda>x. nat \<bar>x\<bar>) ` monoid_closure S) dvd s" 
    by (intro Gcd_dvd_nat image_eqI[where x="int s"] monoid_closure.base) simp
qed

lemma monoid_closure_nat_mult:
  "s \<in> monoid_closure S \<Longrightarrow> int n * s \<in> monoid_closure S"
proof (induct n)
  case 0 with monoid_closure.diff[of s S s] show ?case by simp
next
  case (Suc n) then show ?case
    by (simp add: field_simps monoid_closure.plus)
qed

lemma monoid_closure_uminus:
  assumes s: "s \<in> monoid_closure S" shows "-s \<in> monoid_closure S"
  using monoid_closure.diff[OF monoid_closure.diff[OF s s] s] by simp

lemma monoid_closure_int_mult:
  "s \<in> monoid_closure S \<Longrightarrow> i * s \<in> monoid_closure S"
  by (cases i rule: int_cases') (auto intro: monoid_closure_uminus monoid_closure_nat_mult)

lemma monoid_closure_setsum:
  assumes X: "finite X" "X \<noteq> {}" "X \<subseteq> S"
  shows "(\<Sum>x\<in>X. a x * int x) \<in> monoid_closure S"
  using X
  apply (induct X rule: finite_ne_induct)
  apply (auto intro!: monoid_closure_int_mult monoid_closure.base monoid_closure.plus)
  done

lemma Gcd_monoid_closure_in_monoid_closure:
  fixes S :: "nat set"
  assumes "s \<in> S" "s \<noteq> 0"
  shows "Gcd (monoid_closure S) \<in> monoid_closure S"
proof -
  def m \<equiv> "LEAST n. n \<noteq> 0 \<and> int n \<in> monoid_closure S"
 
  from `s \<in> S` have s: "int s \<in> monoid_closure S"
    by (rule monoid_closure.base)
  moreover from `s \<noteq> 0` have "0 < s"
    by (metis neq0_conv)
  ultimately have s': "s \<noteq> 0 \<and> int s \<in> monoid_closure S"
    by auto

  have m: "m \<noteq> 0 \<and> m \<in> monoid_closure S"
    unfolding m_def
    apply (rule LeastI[of _ s])
    using s'
    by simp
  then have m': "m \<in> monoid_closure S" "0 < int m"
    by auto

  have "Gcd (monoid_closure S) = int m"
    apply (simp add: Gcd_int_def)
    apply (safe intro!: gcd_lcm_complete_lattice_nat.Inf_eqI)
  proof -
    fix y assume "\<forall>i. i \<in> (\<lambda>x. nat \<bar>x\<bar>) ` monoid_closure S \<longrightarrow> y dvd i"
    moreover have "m \<in> (\<lambda>x. nat \<bar>x\<bar>) ` monoid_closure S"
      using m apply (intro image_eqI[where x="int m"]) by auto
    ultimately show "y dvd m" by auto
  next
    fix i s assume s: "s \<in> monoid_closure S"
    from zmod_zdiv_equality'[of s m]
    have "s mod int m = s - s div int m * int m"
      by auto
    also have "s - s div int m * int m \<in> monoid_closure S"
      by (intro monoid_closure.diff s monoid_closure_int_mult m')
    finally have mod_in: "s mod int m \<in>monoid_closure S" by auto 

    have "int m dvd s"
    proof cases
      assume "s mod int m = 0"
      with dvd_eq_mod_eq_0[of "int m" s] show ?thesis by simp
    next
      assume "s mod int m \<noteq> 0"
      moreover have "0 \<le> s mod int m"
        using m'(2) by (rule pos_mod_sign[of "int m" s])
      ultimately have pos: "0 < s mod int m" by auto
      moreover have "s mod int m < int m" 
        using m'(2) by (rule pos_mod_bound)
      moreover have "m \<le> nat (s mod int m)"
        using pos
        apply (subst m_def)
        apply (intro Least_le)
        apply (auto intro: mod_in)
        done
     ultimately show ?thesis
       by auto
    qed
    then show "m dvd nat \<bar>s\<bar>"
      unfolding int_dvd_iff[symmetric] .
  qed
  with m' show ?thesis by simp
qed

lemma Gcd_in_monoid_closure:
  fixes S :: "nat set"
  shows "s \<in> S \<Longrightarrow> s \<noteq> 0 \<Longrightarrow> Gcd S \<in> monoid_closure S"
  unfolding Gcd_eq_Gcd_monoid_closure by (rule Gcd_monoid_closure_in_monoid_closure)

lemma eventually_mult_Gcd:
  fixes S :: "nat set"
  assumes S: "\<And>s t. s \<in> S \<Longrightarrow> t \<in> S \<Longrightarrow> s + t \<in> S"
  assumes s: "s \<in> S" "s \<noteq> 0"
  shows "eventually (\<lambda>m. m * Gcd S \<in> S) sequentially"
proof -
  have "Gcd S \<in> monoid_closure S"
    using s by (rule Gcd_in_monoid_closure)
  also have "monoid_closure S = {s - t | s t. (s = 0 \<or> s \<in> S) \<and> (t = 0 \<or> t \<in> S) }"
  proof auto
    fix x assume "x \<in> monoid_closure S"
    then show "\<exists>s t. x = int s - int t \<and> (s = 0 \<or> s \<in> S) \<and> (t = 0 \<or> t \<in> S)"
    proof induct
      case (base x) then show ?case
        apply (rule_tac x=x in exI)
        apply (rule_tac x=0 in exI)
        apply auto
        done
    next
      case (plus x y)
      then obtain a b c d where
        "a = 0 \<or> a \<in> S" "b = 0 \<or> b \<in> S" "x = int a - b"
        "c = 0 \<or> c \<in> S" "d = 0 \<or> d \<in> S" "y = int c - d"
        by auto
      then show ?case
        apply (rule_tac x="a + c" in exI)
        apply (rule_tac x="b + d" in exI)
        apply simp
        apply (metis S add_0_right add_0_left)
        done
    next
      case (diff x y)
      then obtain a b c d where
        "a = 0 \<or> a \<in> S" "b = 0 \<or> b \<in> S" "x = int a - b"
        "c = 0 \<or> c \<in> S" "d = 0 \<or> d \<in> S" "y = int c - d"
        by auto
      then show ?case
        apply (rule_tac x="a + d" in exI)
        apply (rule_tac x="b + c" in exI)
        apply simp
        apply (metis S add_0_right add_0_left)
        done
    qed
  next
    show "0 \<in> monoid_closure S"
      using monoid_closure.diff[of "int s" S "int s"] monoid_closure.base[OF `s \<in> S`] by simp
  qed (auto intro: monoid_closure.base monoid_closure_uminus monoid_closure.diff)
  finally obtain s t :: nat where st: "s = 0 \<or> s \<in> S" "t = 0 \<or> t \<in> S" "Gcd S = int s - int t" by auto
  then have Gcd_S: "Gcd S = s - t"
    by auto
  with gcd_lcm_complete_lattice_nat.Inf_top_conv[of S] s
  have "t < s"
    by (rule_tac ccontr) auto

  { fix s n have "0 < n \<Longrightarrow> s \<in> S \<Longrightarrow> n * s \<in> S"
    proof (induct n)
      case (Suc n) then show ?case
        by (cases n) (auto intro: S)
    qed simp }
  note cmult_S = this

  show ?thesis
    unfolding eventually_sequentially
  proof cases
    assume "s = 0 \<or> t = 0"
    with st s have *: "Gcd S \<in> S"
      by (auto simp: int_eq_iff)
    then show "\<exists>N. \<forall>n\<ge>N. n * Gcd S \<in> S" by (auto intro!: exI[of _ 1] cmult_S)
  next
    assume "\<not> (s = 0 \<or> t = 0)"
    with st have "s \<in> S" "t \<in> S" "t \<noteq> 0" by auto
    then have "Gcd S dvd t" by auto
    then obtain a where a: "t = Gcd S * a" ..
    with `t \<noteq> 0` have "0 < a" by auto

    show "\<exists>N. \<forall>n\<ge>N. n * Gcd S \<in> S"
    proof (safe intro!: exI[of _ "a * a"])
      fix n 
      def m \<equiv> "(n - a * a) div a"
      def r \<equiv> "(n - a * a) mod a"
      with `0 < a` have "r < a" by simp
      moreover def am \<equiv> "a + m"
      ultimately have "r < am" by simp
      assume "a * a \<le> n" then have n: "n = a * a + (m * a + r)"
        unfolding m_def r_def by simp
      have "n * Gcd S = am * t + r * Gcd S"
        unfolding n a by (simp add: field_simps am_def)
      also have "\<dots> = r * s + (am - r) * t"
        unfolding `Gcd S = s - t` 
        using `t < s` `r < am` by (simp add: field_simps diff_mult_distrib2)
      also have "\<dots> \<in> S"
        using `s \<in> S` `t \<in> S` `r < am` 
        by (cases "r = 0") (auto intro!: cmult_S S)
      finally show "n * Gcd S \<in> S" .
    qed
  qed
qed

context Discrete_Markov_Kernel
begin 

lemma Gcd_period_set_invariant:
  assumes c: "(x, y) \<in> communicating"
  shows "Gcd (period_set x) = Gcd (period_set y)"
proof -
  { fix x y n assume c: "(x, y) \<in> communicating" "x \<noteq> y" and n: "n \<in> period_set x"
    from c obtain l k where "0 < p x y l" "0 < p y x k"
      by (auto simp: communicating_def dest!: accessibleD_pos)
    moreover with `x \<noteq> y` have "l \<noteq> 0 \<and> k \<noteq> 0"
      by (intro notI conjI) (auto simp: p_0)
    ultimately have pos: "0 < l" "0 < k" and l: "0 < p x y l" and k: "0 < p y x k"
      by auto

    from mult_pos_pos[OF k l] prob_reachable_le[of y x y k "k + l"] c 
    have k_l: "0 < p y y (k + l)"
      by (simp add: communicating_in_S)
    then have "Gcd (period_set y) dvd k + l"
      using pos by (auto intro!: Gcd_dvd_nat simp: period_set_def)
    moreover
    from n have "0 < p x x n" "0 < n" by (auto simp: period_set_def)
    from mult_pos_pos[OF k this(1)] prob_reachable_le[of y x x k "k + n"] c 
    have "0 < p y x (k + n)"
      by (simp add: communicating_in_S)
    from mult_pos_pos[OF this(1) l] prob_reachable_le[of y x y "k + n" "(k + n) + l"] c 
    have "0 < p y y (k + n + l)"
      by (simp add: communicating_in_S)
    then have "Gcd (period_set y) dvd (k + l) + n"
      using pos by (auto intro!: Gcd_dvd_nat simp: period_set_def ac_simps)
    ultimately have "Gcd (period_set y) dvd n"
      by (blast elim: dvd_plusE) }
  note this[of x y] this[of y x] c
  moreover have "(y, x) \<in> communicating"
    using c by (simp add: communicating_def)
  ultimately show ?thesis
    by (metis dvd.antisym gcd_lcm_complete_lattice_nat.Inf_greatest)
qed

lemma period_eq:
  assumes "C \<in> S // communicating" "x \<in> C"
  shows "period C = Gcd (period_set x)"
  unfolding period_def
  using assms
  by (rule_tac someI2[where a="Gcd (period_set x)"])
     (auto intro!: Gcd_period_set_invariant irreducibleD)

lemma not_ephemeralD:
  assumes C: "not_ephemeral C" "x \<in> C"
  shows "\<exists>n>0. 0 < p x x n"
proof cases
  assume "\<exists>x. C = {x}"
  with `x \<in> C` have "C = {x}" by auto
  with C p_nonneg[of x x 1] have "0 < p x x 1"
    by (auto simp: not_ephemeral_def)
  with `C = {x}` show ?thesis by auto
next
  from C have irr: "C \<in> S // communicating"
    by (auto simp: not_ephemeral_def)
  assume "\<not>(\<exists>x. C = {x})"
  then have "\<forall>x. C \<noteq> {x}" by auto
  with `x \<in> C` obtain y where "y \<in> C" "x \<noteq> y"
    by blast
  with irreducibleD[OF irr, of x y] C `x \<in> C` have c: "(x, y) \<in> communicating" by auto
  with accessibleD_pos[of x y] accessibleD_pos[of y x]
  obtain k l where pos: "0 < p x y k" "0 < p y x l"
    by (auto simp: communicating_def)
  with `x \<noteq> y` have "l \<noteq> 0"
    by (intro notI) (auto simp: p_0)
  have "0 < p x y k * p y x (k + l - k)"
    using pos by (auto simp: mult_pos_pos)
  also have "p x y k * p y x (k + l - k) \<le> p x x (k + l)"
    using prob_reachable_le[of x y x "k" "k + l"] c by (auto simp: communicating_in_S)
  finally show ?thesis
    using `l \<noteq> 0` `x \<in> C` by (auto intro!: exI[of _ "k + l"])
qed

lemma not_ephemeralD_pos_period:
  assumes C: "not_ephemeral C"
  shows "0 < period C"
proof -
  from C not_empty_irreducible[of C] obtain x where "x \<in> C"
    by (auto simp: not_ephemeral_def)
  from not_ephemeralD[OF C this]
  obtain n where n: "0 < p x x n" "0 < n" by auto
  have C': "C \<in> S // communicating"
    using C by (auto simp: not_ephemeral_def)

  have "period C \<noteq> 0"
    unfolding period_eq[OF C' `x \<in> C`] gcd_lcm_complete_lattice_nat.Inf_top_conv
    using n by (auto simp: period_set_def)
  then show ?thesis by auto
qed

lemma period_posD:
  assumes C: "C \<in> S // communicating" and "0 < period C" "x \<in> C"
  shows "\<exists>n>0. 0 < p x x n"
proof -
  from `0 < period C` have "period C \<noteq> 0"
    by auto
  then show ?thesis
    unfolding period_eq[OF C `x \<in> C`] gcd_lcm_complete_lattice_nat.Inf_top_conv 
    unfolding period_set_def by simp
qed

lemma eventually_periodic:
  assumes C: "C \<in> S // communicating" "0 < period C" "x \<in> C"
  shows "eventually (\<lambda>m. 0 < p x x (m * period C)) sequentially"
proof -
  from period_posD[OF assms] obtain n where n: "0 < p x x n" "0 < n" by auto
  have C': "C \<in> S // communicating"
    using C by auto

  have "period C \<noteq> 0"
    unfolding period_eq[OF C' `x \<in> C`] gcd_lcm_complete_lattice_nat.Inf_top_conv
    using n by (auto simp: period_set_def)
  moreover have "eventually (\<lambda>m. m * Gcd (period_set x) \<in> (period_set x)) sequentially"
  proof (rule eventually_mult_Gcd)
    show "n \<noteq> 0" "n \<in> period_set x"
      using n by (auto simp add: period_set_def)
    fix k l  assume "k \<in> period_set x" "l \<in> period_set x"
    then have "0 < p x x k * p x x l" "0 < l" "0 < k" by (auto simp: period_set_def mult_pos_pos)
    moreover have "p x x k * p x x l \<le> p x x (k + l)"
      using prob_reachable_le[of x x x k "k + l"] C_subset_S[OF C'] `x \<in> C`
      by auto
    ultimately show "k + l \<in> period_set x"
      using `0 < l` by (auto simp: period_set_def)
  qed
  ultimately show "eventually (\<lambda>m. 0 < p x x (m * period C)) sequentially"
    unfolding period_eq[OF C' `x \<in> C`, symmetric]
    using eventually_ge_at_top[of 1]
    by (auto simp add: period_set_def elim!: eventually_elim2)
qed

subsubsection {* @{const aperiodic} is equal to non-zero @{const p} *} text_raw {* \label{sec:aperiodicp} *}

lemma aperiodic_eventually_recurrent:
  "aperiodic C \<longleftrightarrow> C \<in> S // communicating \<and> (\<forall>x\<in>C. eventually (\<lambda>m. 0 < p x x m) sequentially)"
proof safe
  fix x assume "x \<in> C" "aperiodic C"
  with eventually_periodic[of C x]
  show "eventually (\<lambda>m. 0 < p x x m) sequentially"
    by (auto simp add: aperiodic_def)
next
  assume "\<forall>x\<in>C. eventually (\<lambda>m. 0 < p x x m) sequentially" and C: "C \<in> S // communicating"
  moreover from not_empty_irreducible[OF C] obtain x where "x \<in> C" by auto
  ultimately obtain N where "\<And>M.  M\<ge>N \<Longrightarrow> 0 < p x x M"
    by (auto simp: eventually_sequentially)
  then have "{N <..} \<subseteq> period_set x"
    by (auto simp: period_set_def)
  show "aperiodic C"
    unfolding period_eq[OF C `x \<in> C`] aperiodic_def
  proof (intro conjI gcd_lcm_complete_lattice_nat.Inf_eqI)
    fix y assume "\<And>i. i \<in> period_set x \<Longrightarrow> y dvd i"
    from this[of "Suc N"]  this[of "Suc (Suc N)"]  `{N <..} \<subseteq> period_set x`
    have "y dvd (Suc (Suc N) - Suc N)"
      by (intro dvd_diff_nat) (auto simp: subset_eq)
    then show "y dvd 1" by simp
  qed (simp_all add: C)
qed (simp add: aperiodic_def)

end

subsection {* Product Construction for Markov Chains *}

locale Pair_Discrete_Markov_Kernel =
  K1!: Discrete_Markov_Kernel S1 K1 + K2!: Discrete_Markov_Kernel S2 K2 for S1 S2 K1 K2
begin

definition "Sp = S1 \<times> S2"
definition "Kp = (\<lambda>(s1, s2). K1 s1 \<Otimes>\<^sub>M K2 s2)"

lemma SpI[simp]: "x1 \<in> S1 \<Longrightarrow> x2 \<in> S2 \<Longrightarrow> (x1, x2) \<in> Sp"
  by (simp add: Sp_def)

lemma measurable_K1K2[measurable]: "x1 \<in> S1 \<Longrightarrow> x2 \<in> S2 \<Longrightarrow> f \<in> borel_measurable (K1 x1 \<Otimes>\<^sub>M K2 x2)"
  unfolding measurable_def
  by (simp add: space_pair_measure sets_pair_countable[OF K1.countable_space K2.countable_space])

sublocale K1K2!: pair_prob_space "K1 s1" "K2 s2" for s1 s2 by default
sublocale Kp!: prob_space "Kp s" for s
  apply (cases s)
  apply (simp add: Kp_def)
  apply unfold_locales
  done

sublocale P!: Discrete_Markov_Kernel Sp Kp
proof
  show "countable Sp"
    by (simp add: Sp_def)
  show "Sp \<noteq> {}"
    using K1.non_empty_space K2.non_empty_space by (simp add: Sp_def)
  { fix s show "sets (Kp s) = Pow Sp"
      by (cases s) (simp add: Sp_def Kp_def sets_pair_countable) }
qed

lemma sets_Sp: "sets (count_space Sp) = sets (count_space S1 \<Otimes>\<^sub>M count_space S2)"
  by (subst sets_pair_countable[of S1 S2]) (simp_all add: Sp_def)

lemma sets_S_seq_Sp: "sets P.S_seq = sets (\<Pi>\<^sub>M i\<in>UNIV. count_space S1 \<Otimes>\<^sub>M count_space S2)"
  by (intro sets_PIM_cong refl sets_Sp)

lemma space_S_seq_Sp: "space P.S_seq = space (\<Pi>\<^sub>M i\<in>UNIV. count_space S1 \<Otimes>\<^sub>M count_space S2)"
  using sets_eq_imp_space_eq[OF sets_S_seq_Sp] .

lemma prod_measurable: "(\<lambda>(\<omega>1, \<omega>2) n. (\<omega>1 n, \<omega>2 n)) \<in> measurable (K1.paths x1 \<Otimes>\<^sub>M K2.paths x2) P.S_seq"
proof -
  have space_Sp: "space (count_space Sp) = space (count_space S1 \<Otimes>\<^sub>M count_space S2)"
    by (simp add: space_pair_measure Sp_def)
  have eq: "\<And>i. (\<lambda>\<omega>. (case \<omega> of (\<omega>1, \<omega>2) \<Rightarrow> \<lambda>n. (\<omega>1 n, \<omega>2 n)) i) = (\<lambda>\<omega>. (fst \<omega> i, snd \<omega> i))"
    by auto
  have measurable_Sp: "\<And>f M. f \<in> measurable M (count_space Sp) \<longleftrightarrow> f \<in> measurable M (count_space S1 \<Otimes>\<^sub>M count_space S2)"
    unfolding measurable_def space_Sp sets_Sp ..
  
  have PiE: "(\<lambda>\<omega>. case \<omega> of (\<omega>1, \<omega>2) \<Rightarrow> \<lambda>n. (\<omega>1 n, \<omega>2 n)) \<in> space (K1.paths x1) \<times> space (K2.paths x2) \<rightarrow> UNIV \<rightarrow>\<^sub>E Sp"
    by (auto simp: space_PiM PiE_iff)

  show ?thesis
    apply (intro measurable_PiM_single')
    unfolding eq measurable_Sp
    apply (intro measurable_Pair)
    apply (simp_all add: space_pair_measure PiE del: K1.space_paths K2.space_paths)
    done
qed

lemma P_paths_eq_prod:
  assumes [simp]: "x1 \<in> S1" "x2 \<in> S2"
  shows "P.paths (x1, x2) = distr (K1.paths x1 \<Otimes>\<^sub>M K2.paths x2) P.S_seq (\<lambda>(\<omega>1, \<omega>2) n. (\<omega>1 n, \<omega>2 n))"
  (is "?L = ?R")
proof (rule nat_eq_cyl)
  show "sets ?L = sets P.S_seq" "sets ?R = sets P.S_seq" by simp_all

  have sp_eq: "(\<lambda>(\<omega>1, \<omega>2) n. (\<omega>1 n, \<omega>2 n)) -` space (P.S_seq) \<inter> space (K1.paths x1) \<times> space (K2.paths x2) =
    space (K1.paths x1) \<times> space (K2.paths x2)"
    by (auto simp: space_PiM PiE_iff)
    
  show "emeasure ?R (space P.S_seq) \<noteq> \<infinity>"
    apply (subst emeasure_distr[OF prod_measurable])
    apply (simp_all add: space_pair_measure sp_eq del: K1.space_paths K2.space_paths)
    apply (subst K2.path.emeasure_pair_measure_Times)
    apply simp_all
    done
  show "countable Sp" by simp

  fix n :: nat and \<omega>' :: "nat \<Rightarrow> _" assume \<omega>': "\<And>i. \<omega>' i \<in> Sp"
  { fix i from \<omega>'[of i] have "fst (\<omega>' i) \<in> S1" "snd (\<omega>' i) \<in> S2"
      by (auto simp: Sp_def) }
  note this[simp] \<omega>'[simp]
  { fix n 
    have eq: "((\<lambda>(\<omega>1, \<omega>2) n. (\<omega>1 n, \<omega>2 n)) -` {\<omega> \<in> space ?R. \<forall>i<n. \<omega> i = \<omega>' i} \<inter> space (K1.paths x1 \<Otimes>\<^sub>M K2.paths x2)) = 
      {\<omega> \<in> space (K1.paths x1). \<forall>i<n. \<omega> i = fst (\<omega>' i)} \<times> {\<omega> \<in> space (K2.paths x2). \<forall>i<n. \<omega> i = snd (\<omega>' i)}"
      by (auto simp: space_PiM space_pair_measure PiE_iff eq_commute)

    { fix i
      let ?p = "nat_case (x1, x2) \<omega>' i"
      let ?x1 = "case i of 0 \<Rightarrow> x1 | Suc x \<Rightarrow> fst (\<omega>' x)" and ?x2 = "case i of 0 \<Rightarrow> x2 | Suc x \<Rightarrow> snd (\<omega>' x)"
      have "emeasure (Kp ?p) {\<omega>' i} = emeasure (K1 ?x1) {fst (\<omega>' i)} * emeasure (K2 ?x2) {snd (\<omega>' i)}"
        unfolding Kp_def using K2.K.emeasure_pair_measure_Times[of "{fst (\<omega>' i)}" "K1 ?x1" "{snd (\<omega>' i)}" "?x2"]
        by (auto split: prod.split nat.splits) }
    note prod = this

    have "emeasure ?L {\<omega> \<in> space ?L. \<forall>i<n. \<omega> i = \<omega>' i} = emeasure ?R {\<omega> \<in> space ?R. \<forall>i<n. \<omega> i = \<omega>' i}"
      apply (subst emeasure_distr[OF prod_measurable])
      apply simp
      unfolding eq
      apply (subst K2.path.emeasure_pair_measure_Times)
      apply simp
      apply simp
      apply (subst P.emeasure_path)
      apply simp
      apply simp
      apply (subst K1.emeasure_path)
      apply simp
      apply simp
      apply (subst K2.emeasure_path)
      apply simp
      apply simp
      apply (simp add: prod setprod.distrib)
      done }
  from this[of "Suc n"] show "emeasure ?L {\<omega> \<in> space ?L. \<forall>i\<le>n. \<omega> i = \<omega>' i} = emeasure ?R {\<omega> \<in> space ?R. \<forall>i\<le>n. \<omega> i = \<omega>' i}"
    unfolding less_Suc_eq_le .
qed

lemma prod_eq_prob_paths:
  assumes x[simp]: "x1 \<in> S1" "x2 \<in> S2" and [measurable]: "Measurable.pred K1.S_seq P1" "Measurable.pred K2.S_seq P2"
  shows "\<P>(\<omega> in K1.paths x1. P1 \<omega>) * \<P>(\<omega> in K2.paths x2. P2 \<omega>) = \<P>(\<omega> in P.paths (x1, x2). P1 (fst \<circ> \<omega>) \<and> P2 (snd \<circ> \<omega>))"
proof -
  interpret PROD: pair_prob_space "K1.paths x1" "K2.paths x2"
    by default
  let ?P1 = "{\<omega>\<in>space K1.S_seq. P1 \<omega>}" and ?P2 = "{\<omega>\<in>space K2.S_seq. P2 \<omega>}"
  let ?P = "{\<omega> \<in> space (P.paths (x1, x2)). P1 (fst \<circ> \<omega>) \<and> P2 (snd \<circ> \<omega>)}"
  have eq: "((\<lambda>(\<omega>1, \<omega>2) n. (\<omega>1 n, \<omega>2 n)) -` ?P \<inter> space (K1.paths x1 \<Otimes>\<^sub>M K2.paths x2)) = ?P1 \<times> ?P2"
    by (auto simp: space_PiM space_pair_measure PiE_iff o_def)
  show ?thesis
    apply (subst P_paths_eq_prod[OF x])
    apply (subst measure_distr[OF prod_measurable])
    apply (simp add: comp_def sets_S_seq_Sp sets_eq_imp_space_eq[OF sets_S_seq_Sp])
    unfolding eq
    using K2.path.emeasure_pair_measure_Times[of ?P1 "K1.paths x1" ?P2 x2]
    apply (simp add: K2.path.emeasure_eq_measure K1.path.emeasure_eq_measure PROD.emeasure_eq_measure)
    done
qed

lemma p_eq_p1_p2:
  "x1 \<in> S1 \<Longrightarrow> x2 \<in> S2 \<Longrightarrow> y1 \<in> S1 \<Longrightarrow> y2 \<in> S2 \<Longrightarrow> P.p (x1, x2) (y1, y2) n = K1.p x1 y1 n * K2.p x2 y2 n"
  unfolding P.p_def K1.p_def K2.p_def
  apply (subst prod_eq_prob_paths)
  apply (auto intro!: arg_cong2[where f=measure] split: nat.splits)
  done

lemma Sp_iff: "(a, b) \<in> Sp \<longleftrightarrow> a \<in> S1 \<and> b \<in> S2"
  by (simp add: Sp_def)

lemma P_accessibleD:
  assumes "P.accessible (x1, x2) (y1, y2)"shows "K1.accessible x1 y1" "K2.accessible x2 y2"
  using assms by (auto simp: P.accessible_iff K1.accessible_iff K2.accessible_iff Sp_iff p_eq_p1_p2
                             zero_less_mult_iff not_le[of 0, symmetric] K1.p_nonneg K2.p_nonneg
                      cong: conj_cong)

lemma aperiodicI_pair:
  assumes C1: "K1.aperiodic C1" and C2: "K2.aperiodic C2"
  shows "P.aperiodic (C1 \<times> C2)"
  unfolding P.aperiodic_eventually_recurrent
proof safe
  from C1[unfolded K1.aperiodic_eventually_recurrent] C2[unfolded K2.aperiodic_eventually_recurrent]
  have C1: "C1 \<in> S1 // K1.communicating" and C2: "C2 \<in> S2 // K2.communicating" and
    ev: "\<And>x. x \<in> C1 \<Longrightarrow> eventually (\<lambda>m. 0 < K1.p x x m) sequentially" "\<And>x. x \<in> C2 \<Longrightarrow> eventually (\<lambda>m. 0 < K2.p x x m) sequentially"
    by auto
  then have subset: "C1 \<subseteq> S1" "C2 \<subseteq> S2"
    using K1.C_subset_S[OF C1] K2.C_subset_S[OF C2] by auto
  { fix x1 x2 assume x: "x1 \<in> C1" "x2 \<in> C2"
    with subset have xS: "x1 \<in> S1" "x2 \<in> S2" by auto
    from ev(1)[OF x(1)] ev(2)[OF x(2)]
    show "eventually (\<lambda>m. 0 < P.p (x1, x2) (x1, x2) m) sequentially"
       by eventually_elim  (simp add: p_eq_p1_p2 x xS mult_pos_pos) }
  
  { fix x1 x2 y1 y2
    assume acc: "K1.accessible x1 y1" "K2.accessible x2 y2" "x1 \<in> C1" "y1 \<in> C1" "x2 \<in> C2" "y2 \<in> C2"
    then obtain k l where "0 < K1.p x1 y1 l" "0 < K2.p x2 y2 k"
      by (auto dest!: K1.accessibleD_pos K2.accessibleD_pos)
    with acc ev(1)[of y1] ev(2)[of y2]
    have "eventually (\<lambda>m. 0 < K1.p x1 y1 l * K1.p y1 y1 m \<and> 0 < K2.p x2 y2 k * K2.p y2 y2 m) sequentially"
      by (auto simp: K1.accessible_in_S2 K2.accessible_in_S2 elim: eventually_elim2 intro: mult_pos_pos)
    then have "eventually (\<lambda>m. 0 < K1.p x1 y1 (m + l) \<and> 0 < K2.p x2 y2 (m + k)) sequentially"
    proof eventually_elim
      fix m assume "0 < K1.p x1 y1 l * K1.p y1 y1 m \<and> 0 < K2.p x2 y2 k * K2.p y2 y2 m"
      with acc
        K1.prob_reachable_le[of x1 y1 y1 l "l + m"]
        K2.prob_reachable_le[of x2 y2 y2 k "k + m"]
      show "0 < K1.p x1 y1 (m + l) \<and> 0 < K2.p x2 y2 (m + k)"
        by (auto simp add: K1.accessible_in_S1 K2.accessible_in_S1 K1.accessible_in_S2 K2.accessible_in_S2 ac_simps)
    qed
    then have "eventually (\<lambda>m. 0 < K1.p x1 y1 m \<and> 0 < K2.p x2 y2 m) sequentially"
      unfolding  eventually_conj_iff by (subst (asm) (1 2) eventually_sequentially_seg) (auto elim: eventually_elim2)
    then obtain N where "0 < K1.p x1 y1 N" "0 < K2.p x2 y2 N"
      by (auto simp: eventually_sequentially)
    with acc have "0 < P.p (x1, x2) (y1, y2) N"
      by (auto simp add: K1.accessible_in_S1 K2.accessible_in_S1 K1.accessible_in_S2 K2.accessible_in_S2 p_eq_p1_p2 mult_pos_pos)
    with acc have "P.accessible (x1, x2) (y1, y2)"
      by (auto simp add: K1.accessible_in_S1 K2.accessible_in_S1 K1.accessible_in_S2 K2.accessible_in_S2
               intro!: P.accessibleI_pos) }
  note 1 = this

  { fix x1 x2 y1 y2
    assume acc:"P.accessible (x1, x2) (y1, y2)"
    then have "(x1, x2) \<in> Sp" "(y1, y2) \<in> Sp"
      by (auto simp: P.accessible_in_S1 P.accessible_in_S2)
    moreover from acc obtain k where "0 < P.p (x1, x2) (y1, y2) k" by (auto dest!: P.accessibleD_pos)
    ultimately have "K1.accessible x1 y1 \<and> K2.accessible x2 y2" using subset
      by (subst (asm) p_eq_p1_p2)
         (auto intro!: K1.accessibleI_pos K2.accessibleI_pos
               simp: Sp_def zero_less_mult_iff K1.p_nonneg K2.p_nonneg not_le[of 0, symmetric]) }
  note 2 = this

  from K1.not_empty_irreducible[OF C1] K2.not_empty_irreducible[OF C2] obtain x1 x2 where xC: "x1 \<in> C1" "x2 \<in> C2" by auto
  show "C1 \<times> C2 \<in> Sp // P.communicating"
    apply (simp add: quotient_def Image_def)
    apply (safe intro!: bexI[of _ "(x1, x2)"])
  proof -
    fix y1 y2 assume yC: "y1 \<in> C1" "y2 \<in> C2"
    from K1.irreducibleD[OF C1 `x1 \<in> C1` `y1 \<in> C1`] K2.irreducibleD[OF C2 `x2 \<in> C2` `y2 \<in> C2`]
    show "((x1, x2), (y1, y2)) \<in> P.communicating"
      using 1[of x1 y1 x2 y2] 1[of y1 x1 y2 x2] xC yC
      by (auto simp: P.communicating_def K1.communicating_def K2.communicating_def)
  next
    fix y1 y2 assume "((x1, x2), (y1, y2)) \<in> P.communicating"
    with 2[of x1 x2 y1 y2] 2[of y1 y2 x1 x2]
    have "(x1, y1) \<in> K1.communicating" "(x2, y2) \<in> K2.communicating"
      by (auto simp: P.communicating_def K1.communicating_def K2.communicating_def)
    with xC show "y1 \<in> C1" "y2 \<in> C2"
      using K1.communicatingD1[OF C1] K2.communicatingD1[OF C2] by auto
  next
    from subset xC show "(x1, x2) \<in> Sp" by (auto simp: Sp_def)
  qed
qed

lemma stationary_distributionI_pair:
  assumes N1: "K1.stationary_distribution N1" "prob_space N1"
  assumes N2: "K2.stationary_distribution N2" "prob_space N2"
  shows "P.stationary_distribution (N1 \<Otimes>\<^sub>M N2)"
  unfolding P.stationary_distribution_def
proof (rule pair_measure_eqI)
  interpret N1: prob_space N1 by fact
  show "sigma_finite_measure N1" by default

  interpret N2: prob_space N2 by fact
  show "sigma_finite_measure N2" by default

  interpret N1N2: pair_prob_space N1 N2 by default

  interpret SpC: sigma_finite_measure "count_space Sp"
    using P.countable_space by (rule sigma_finite_measure_count_space)

  have sets_N1: "sets N1 = Pow S1"
    by (subst K1.stationary_distributionD_eq[OF N1(1)]) (simp add: sets_point_measure)
  moreover have sets_N2: "sets N2 = Pow S2"
    by (subst K2.stationary_distributionD_eq[OF N2(1)]) (simp add: sets_point_measure)
  ultimately have sets_N1N2: "sets (N1 \<Otimes>\<^sub>M N2) = Pow Sp"
    by (simp add: Sp_def sets_pair_countable)
  then show "sets (N1 \<Otimes>\<^sub>M N2) = sets (point_measure Sp (\<lambda>x. \<integral>\<^sup>+ s. emeasure (Kp s) {x} \<partial>(N1 \<Otimes>\<^sub>M N2)))"
    by (simp add: sets_point_measure)

  have [simp]: "\<And>f. f \<in> borel_measurable (N1 \<Otimes>\<^sub>M N2)"
    unfolding measurable_def using sets_eq_imp_space_eq[of "N1 \<Otimes>\<^sub>M N2" "count_space Sp"]
    by (simp add: sets_N1N2)
  
  have space_N12: "space N1 = S1" "space N2 = S2"
    using sets_eq_imp_space_eq[of N1 "count_space S1"] sets_eq_imp_space_eq[of N2 "count_space S2"]
    by (simp_all add: sets_N1 sets_N2)

  have sets_N1N2C: "sets ((N1 \<Otimes>\<^sub>M N2) \<Otimes>\<^sub>M count_space Sp) = Pow (Sp \<times> Sp)"
    by (simp add: sets_pair_countable sets_N1N2)
  then have [simp]: "\<And>f. f \<in> borel_measurable ((N1 \<Otimes>\<^sub>M N2) \<Otimes>\<^sub>M count_space Sp)"
    unfolding measurable_def  using sets_eq_imp_space_eq[of "(N1 \<Otimes>\<^sub>M N2) \<Otimes>\<^sub>M count_space Sp" "count_space (Sp \<times> Sp)"]
    by simp

  have [simp]: "\<And>f. f \<in> borel_measurable N1"
    unfolding measurable_def using sets_eq_imp_space_eq[of "N1" "count_space S1"]
    by (simp add: sets_N1)

  have [simp]: "\<And>f. f \<in> borel_measurable N2"
    unfolding measurable_def using sets_eq_imp_space_eq[of "N2" "count_space S2"]
    by (simp add: sets_N2)

  fix A B assume A_and_B: "A \<in> sets N1" "B \<in> sets N2"
  then have AB: "A \<times> B \<subseteq> Sp" and "A \<subseteq> S1" "B \<subseteq> S2"
    by (auto simp add: sets_N1 sets_N2 Sp_def)
  have 1: "\<And>x s. indicator (A \<times> B) x * indicator {s} x = (indicator (if s \<in> A \<times> B then {s} else {}) x :: ereal)"
    by (auto simp: indicator_def)
  have "(\<integral>\<^sup>+ x. (\<integral>\<^sup>+ y. (\<integral>\<^sup>+ ya. emeasure (Kp (x, y)) {ya} * indicator (A \<times> B) ya \<partial>count_space Sp) \<partial>N2) \<partial>N1) =
    (\<integral>\<^sup>+ x. (\<integral>\<^sup>+ y. (\<integral>\<^sup>+ ya. emeasure (density (Kp (x, y)) (indicator (A \<times> B))) {ya} \<partial>count_space Sp) \<partial>N2) \<partial>N1)"
    by (intro positive_integral_cong) (auto simp: emeasure_density 1 space_N12)
  also have "\<dots> = (\<integral>\<^sup>+ x. (\<integral>\<^sup>+ y. emeasure (density (Kp (x, y)) (indicator (A \<times> B))) Sp \<partial>N2) \<partial>N1)"
    apply (intro positive_integral_cong)
    apply (subst emeasure_countable_singleton[symmetric])
    apply (simp_all add: space_N12 )
    done
  also have "\<dots> = (\<integral>\<^sup>+ x. (\<integral>\<^sup>+ y. emeasure (Kp (x, y)) (A \<times> B) \<partial>N2) \<partial>N1)"
    apply (intro positive_integral_cong)
    apply (subst emeasure_density)
    apply (simp_all add: space_N12 AB)
    apply (subst positive_integral_cong)
    apply simp
    apply (simp add: AB)
    done
  also have "\<dots> = (\<integral>\<^sup>+ x. (\<integral>\<^sup>+ y. emeasure (K1 x) A * emeasure (K2  y) B \<partial>N2) \<partial>N1)"
    apply (intro positive_integral_cong)
    apply (simp_all add: space_N12 Kp_def)
    apply (subst K2.K.emeasure_pair_measure_Times)
    using A_and_B sets_N1 sets_N2
    apply simp_all
    done
  also have "\<dots> = (\<integral>\<^sup>+ x. emeasure (K1 x) A * (\<integral>\<^sup>+ y. emeasure (K2  y) B \<partial>N2) \<partial>N1)"
    apply (intro positive_integral_cong)
    apply (subst positive_integral_cmult)
    apply (simp_all add: emeasure_nonneg)
    done
  also have "\<dots> = (\<integral>\<^sup>+ x. emeasure (K1 x) A \<partial>N1) * (\<integral>\<^sup>+ y. emeasure (K2  y) B \<partial>N2)"
    apply (subst positive_integral_multc)
    apply (simp_all add: positive_integral_positive)
    done
  also have "\<dots> = emeasure N1 A * emeasure N2 B"
    apply (subst K1.stationary_distributionD_emeasure[OF N1 `A \<subseteq> S1`])
    apply (subst K2.stationary_distributionD_emeasure[OF N2 `B \<subseteq> S2`]) ..
  finally have 2: "(\<integral>\<^sup>+ x. (\<integral>\<^sup>+ y. (\<integral>\<^sup>+ ya. emeasure (Kp (x, y)) {ya} * indicator (A \<times> B) ya \<partial>count_space Sp) \<partial>N2) \<partial>N1) =
   emeasure N1 A * emeasure N2 B" .

  show "emeasure N1 A * emeasure N2 B =
    emeasure (point_measure Sp (\<lambda>x. \<integral>\<^sup>+ s. emeasure (Kp s) {x} \<partial>(N1 \<Otimes>\<^sub>M N2))) (A \<times> B)"
    apply (simp add: point_measure_def emeasure_density AB)
    apply (subst positive_integral_multc[symmetric])
    apply simp_all
    apply (subst pair_sigma_finite.Fubini')
    apply unfold_locales
    apply simp    
    apply (subst N2.positive_integral_fst[symmetric])
    apply (simp_all add: positive_integral_positive 2)
    done
qed

end

subsection {* Stationary Distribution is the Limit of the Marginal Distributions *} text_raw {* \label{sec:limit} *}

context Discrete_Markov_Kernel
begin

lemma stationary_distribution_imp_limit:
  assumes "aperiodic C" "essential_class C" and N: "stationary_distribution N" "prob_space N" "measure N (S - C) = 0"
  assumes [simp]: "y \<in> C"
  shows "(\<lambda>n. \<integral>x. \<bar>p y x n - measure N {x}\<bar> \<partial>count_space C) ----> 0"
    (is "?L ----> 0")
proof -
  interpret N: prob_space N by fact
  have sets_N[simp]: "sets N = Pow S"
    by (subst stationary_distributionD_eq[OF N(1)]) (simp add: sets_point_measure)
  
  have space_N[simp]: "space N = S"
    using sets_eq_imp_space_eq[of N "count_space S"] by (simp_all add: sets_N)

  then have [simp]: "\<And>f. f \<in> borel_measurable N"
    unfolding measurable_def by (simp add: sets_N space_N)

  have [simp]: "\<And>f M. f \<in> measurable N M \<longleftrightarrow> f \<in> measurable (count_space S) M"
    unfolding measurable_def by (simp add: sets_N space_N)

  def S' \<equiv> "insert None (Some ` S)"
  def K' \<equiv> "\<lambda>x. distr (case x of Some x \<Rightarrow> K x | None \<Rightarrow> N) (count_space S') Some"
  have [simp]: "Some \<in> S \<rightarrow> S'"
    by (simp add: S'_def)
  have [simp]: "(Some -` S' \<inter> S) = S"
    by (auto simp: S'_def)

  from `essential_class C` have C: "C \<in> S // communicating"
    by (simp add: essential_class_def)
  from C_subset_S[OF this] have [simp]: "\<And>x. x \<in> C \<Longrightarrow> x \<in> S" by auto

  have [simp]: "\<And>x. Some x \<in> S' \<longleftrightarrow> x \<in> S" by (auto simp: S'_def)

  interpret K': prob_space "K' s" for s
    apply rule
    apply (cases s)
    using N.emeasure_space_1 K.emeasure_space_1
    apply (simp_all add: K'_def emeasure_distr K_measurable1)
    done
  interpret K2: Discrete_Markov_Kernel S' K'
    proof qed (simp_all add: S'_def K'_def)
  interpret KN: Pair_Discrete_Markov_Kernel S S' K K'
    by default

  from stationary_distributionD[OF `essential_class C` N]
  have pos: "\<And>x. x \<in> C \<Longrightarrow> pos_recurrent x" and "N = stat C" by auto

  have pos: "\<And>x. x \<in> C \<Longrightarrow> 0 < emeasure N {x}"
    using pos C_subset_S[OF C] unfolding stat_def `N = stat C`
    by (subst emeasure_point_measure_finite2)
       (auto simp: ereal_0_gt_inverse positive_integral_positive inverse_nonneg_iff pos_recurrent_def)
  then have rpos: "\<And>x. x \<in> C \<Longrightarrow> 0 < measure N {x}"
    by (simp add: N.emeasure_eq_measure)

  have eq: "\<And>x y. (if x = y then 1 else 0) = indicator {y} x" by auto

  have intK: "\<And>f x. (\<integral>x. f x \<partial>K' (Some x)) = (\<integral>x. f (Some x) \<partial>K x)"
    by (simp add: K'_def integral_distr K_measurable1)

  { fix m x y assume "x \<in> S" "y \<in> S"
    then have "K2.p (Some x) (Some y) m = p x y m"
    proof (induct m arbitrary: x)
      case 0 then show ?case by (simp add: K2.p_0 p_0)
    next
      case (Suc m)
      then have [simp]: "x \<in> S" "y \<in> S" "Some x \<in> S'" "Some y \<in> S'" by (auto simp: S'_def)
      show ?case
        by (auto intro!: integral_cong simp add: K2.p_Suc p_Suc intK Suc.hyps)
    qed }
  note K_p_eq = this

  have in_S': "\<And>x. x \<in> S' \<longleftrightarrow> x = None \<or> (\<exists>y\<in>S. x = Some y)"
    by (auto simp: S'_def)

  have None_S'[simp]: "None \<in> S'"
    by (auto simp: S'_def)

  { fix n x assume x: "x \<in> S"
    then have "K2.p (Some x) None n = 0"
    proof (induct n arbitrary: x)
      case (Suc n x) then show ?case
        apply (simp add: K2.p_Suc intK cong: integral_cong)
        apply (subst integral_cong)
        apply (rule Suc)
        apply simp_all
        done
    qed (simp add: K2.p_0) }
  note K_S_None = this

  from not_empty_irreducible[OF C] obtain c0 where c0: "c0 \<in> C" by auto

  have K2_accessible: "\<And>x y. K2.accessible (Some x) y \<longleftrightarrow> (\<exists>z\<in>S. y = Some z \<and> accessible x z)"
    apply (auto simp: K2.accessible_iff accessible_iff K_p_eq)
    apply (case_tac y)
    apply (auto simp: K_p_eq K_S_None)
    done

  have K2_communicating: "\<And>c x. c \<in> C \<Longrightarrow> (Some c, x) \<in> K2.communicating \<longleftrightarrow> (\<exists>c'\<in>C. x = Some c')"
  proof safe
    fix x c assume "c \<in> C" "(Some c, x) \<in> K2.communicating"
    then show "\<exists>c'\<in>C. x = Some c'"
      by (cases x)
         (auto simp: communicating_iff K2.communicating_iff K_p_eq K_S_None intro!: irreducibleD2[OF C `c\<in>C`])
  next
    fix c c' x assume "c \<in> C" "c' \<in> C" 
    with irreducibleD[OF C this] show "(Some c, Some c') \<in> K2.communicating"
      by (auto simp: K2.communicating_iff communicating_iff K_p_eq)
  qed

  have "Some ` C \<in> S' // K2.communicating"
    by (auto simp add: quotient_def Image_def c0 K2_communicating
             intro!: bexI[of _ "Some c0"])
  then have "K2.essential_class (Some ` C)"
    by (rule K2.essential_classI)
       (auto simp: K2_accessible essential_classD2[OF `essential_class C`])

  have "K2.aperiodic (Some ` C)"
    unfolding K2.aperiodic_eventually_recurrent
  proof safe
    fix x assume "x \<in> C" then show "eventually (\<lambda>m. 0 < K2.p (Some x) (Some x) m) sequentially"
      using `aperiodic C` unfolding aperiodic_eventually_recurrent
      by (auto elim!: eventually_elim1 simp: K_p_eq)
  qed fact
  then have aperiodic: "KN.P.aperiodic (C \<times> Some ` C)"
    by (rule KN.aperiodicI_pair[OF `aperiodic C`])

  have KN_essential: "KN.P.essential_class (C \<times> Some ` C)"
  proof (rule KN.P.essential_classI)
    show "C \<times> Some ` C \<in> KN.Sp // KN.P.communicating"
      using aperiodic by (simp add: KN.P.aperiodic_def)
  next
    fix x y assume "x \<in> C \<times> Some ` C" "KN.P.accessible x y"
    with KN.P_accessibleD[of "fst x" "snd x" "fst y" "snd y"]
    show "y \<in> C \<times> Some ` C"
      by (cases x y rule: prod.exhaust[case_product prod.exhaust])
         (auto simp: K2_accessible essential_classD2[OF `essential_class C`])
  qed

  have fm_N: "finite_measure N" 
    by unfold_locales

  { fix n x y assume [simp]: "y \<in> S"
    have "measure N {y} = \<P>(\<omega> in K2.paths None. nat_case None \<omega> (Suc n) = Some y)"
      unfolding stationary_distribution_iterate[OF N(1) fm_N `y \<in> S`, of n]
      apply (subst K2.p_def[symmetric])
      apply (subst K2.p_Suc)
      apply simp
      apply (subst K'_def)
      apply simp
      apply (subst integral_distr)
      apply simp
      apply simp
      apply (auto intro!: integral_cong simp: K_p_eq)
      done
    then have "measure N {y} = \<P>(\<omega> in K2.paths None. \<omega> n = Some y)"
      by simp }
  note measure_y_eq = this

  def D \<equiv> "{x\<in>S \<times> S'. Some (fst x) = snd x}"

  { fix n x assume [simp]: "x \<in> S"
    have "\<P>(\<omega> in KN.P.paths (y, None). (\<exists>i<n. snd (\<omega> n) = Some x \<and> smallest \<omega> D i)) =
      (\<Sum>i<n. \<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> smallest \<omega> D i))"
      apply (subst KN.P.finite_measure_finite_Union[symmetric])
      apply (auto simp add: KN.space_S_seq_Sp KN.sets_S_seq_Sp
                            disjoint_family_on_def
                  intro: smallest_unique
                  intro!: arg_cong2[where f=measure])
      done
    also have "\<dots> = (\<Sum>i<n. \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> smallest \<omega> D i))"
    proof (intro setsum_cong refl)
      fix i assume i: "i \<in> {..< n}"
      from i show "\<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> smallest \<omega> D i) = 
        \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> smallest \<omega> D i)"
        apply (subst (1 2) KN.P.prob_split_Collect[where i="Suc i"])
        apply (simp add: KN.Sp_def)
        apply (simp_all add: KN.space_S_seq_Sp KN.sets_S_seq_Sp smallest_comb_seq comb_seq_Suc_less)
      proof (intro integral_cong)
        fix \<omega> assume \<omega>: "\<omega> \<in> space (KN.P.paths (y, None))"
        then have "\<omega> i \<in> S \<times> S'" 
          by (auto simp: space_PiM PiE_iff) (simp add: KN.Sp_def)
        then have [simp]: "fst (\<omega> i) \<in> S" "snd (\<omega> i) \<in> S'"  
          by auto
        show "measure (KN.P.paths (\<omega> i)) {\<omega>' \<in> space (Pi\<^sub>M UNIV (\<lambda>i. count_space S \<Otimes>\<^sub>M count_space S')).
            snd (\<omega>' (n - Suc i)) = Some x \<and> smallest \<omega> D i} =
          measure (KN.P.paths (\<omega> i)) {\<omega>' \<in> space (Pi\<^sub>M UNIV (\<lambda>i. count_space S \<Otimes>\<^sub>M count_space S')).
            fst (\<omega>' (n - Suc i)) = x \<and> smallest \<omega> D i}"
        proof cases
          assume "smallest \<omega> D i"
          from smallestD1[OF this]
          have eq: "snd (\<omega> i) = Some (fst (\<omega> i))" by (simp add: D_def)

          have "\<P>(\<omega>' in KN.P.paths (\<omega> i). fst (\<omega>' (n - Suc i)) = x) = 
            \<P>(\<omega>' in paths (fst (\<omega> i)). \<omega>' (n - Suc i) = x) * \<P>(\<omega>' in K2.paths (snd (\<omega> i)). True)"
            using \<omega> by (subst KN.prod_eq_prob_paths) simp_all
          also have "\<dots> = p (fst (\<omega> i)) x (Suc (n - Suc i))"
            using K2.prob_space by (simp add: p_def)
          also have "\<dots> = K2.p (snd (\<omega> i)) (Some x) (Suc (n - Suc i))"
            by (simp add: K_p_eq eq)
          also have "\<dots> = \<P>(\<omega>' in paths (fst (\<omega> i)). True) * \<P>(\<omega>' in K2.paths (snd (\<omega> i)). \<omega>' (n - Suc i) = Some x)"
            using prob_space by (simp add: K2.p_def)
          also have "\<dots> = \<P>(\<omega>' in KN.P.paths (\<omega> i). snd (\<omega>' (n - Suc i)) = Some x)"
            using \<omega> by (subst KN.prod_eq_prob_paths) simp_all
          finally show ?thesis using `smallest \<omega> D i` by (simp add: KN.space_S_seq_Sp KN.sets_S_seq_Sp)
        qed simp
      qed
    qed
    also have "\<dots> = \<P>(\<omega> in KN.P.paths (y, None). (\<exists>i<n. fst (\<omega> n) = x \<and> smallest \<omega> D i))"
      apply (subst KN.P.finite_measure_finite_Union[symmetric])
      apply (auto simp add: KN.space_S_seq_Sp KN.sets_S_seq_Sp
                            disjoint_family_on_def
                  intro: smallest_unique
                  intro!: arg_cong2[where f=measure])
      done
    finally have eq: "\<P>(\<omega> in KN.P.paths (y, None). (\<exists>i<n. snd (\<omega> n) = Some x \<and> smallest \<omega> D i)) =
      \<P>(\<omega> in KN.P.paths (y, None). (\<exists>i<n. fst (\<omega> n) = x \<and> smallest \<omega> D i))" .

    have "p y x (Suc n) - measure N {x} = \<P>(\<omega> in paths y. \<omega> n = x) - \<P>(\<omega> in K2.paths None. \<omega> n = Some x)"
      unfolding p_def by (subst measure_y_eq) simp_all
    also have "\<P>(\<omega> in paths y. \<omega> n = x) = \<P>(\<omega> in paths y. \<omega> n = x) * \<P>(\<omega> in K2.paths None. True)"
      using K2.prob_space by simp
    also have "\<dots> = \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x)"
      by (subst KN.prod_eq_prob_paths) auto
    also have "\<dots> = \<P>(\<omega> in KN.P.paths (y, None). (\<exists>i<n. fst (\<omega> n) = x \<and> smallest \<omega> D i)) +
      \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> \<not> (\<exists>i<n. smallest \<omega> D i))"
      apply (subst KN.P.finite_measure_Union[symmetric])
      unfolding KN.P.space_paths KN.P.sets_paths
      apply (simp_all add: KN.space_S_seq_Sp KN.sets_S_seq_Sp)
      apply (auto intro!: arg_cong2[where f=measure])
      done
    also have "\<P>(\<omega> in K2.paths None. \<omega> n = Some x) = \<P>(\<omega> in paths y. True) * \<P>(\<omega> in K2.paths None. \<omega> n = Some x)"
      using prob_space by simp
    also have "\<dots> = \<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x)"
      by (subst KN.prod_eq_prob_paths) auto
    also have "\<dots> = \<P>(\<omega> in KN.P.paths (y, None). (\<exists>i<n. snd (\<omega> n) = Some x \<and> smallest \<omega> D i)) +
      \<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> \<not> (\<exists>i<n. smallest \<omega> D i))"
      apply (subst KN.P.finite_measure_Union[symmetric])
      unfolding KN.P.space_paths KN.P.sets_paths
      apply (simp_all add: KN.space_S_seq_Sp KN.sets_S_seq_Sp)
      apply (auto intro!: arg_cong2[where f=measure])
      done
    finally have "\<bar> p y x (Suc n) - measure N {x} \<bar> =
      \<bar> \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)) -
      \<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)) \<bar>"
      unfolding eq by (simp add: field_simps)
    also have "\<dots> \<le> \<bar> \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)) \<bar> +
      \<bar> \<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)) \<bar>"
      by (rule abs_triangle_ineq4)
    also have "\<dots> \<le> \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)) +
      \<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> \<not> (\<exists>i<n. smallest \<omega> D i))"
      by (simp add: measure_nonneg)
    finally have "\<bar> p y x (Suc n) - measure N {x} \<bar> \<le> \<dots>" . }
  note mono = this

  have Sp_iff: "\<And>x. x \<in> KN.Sp \<longleftrightarrow> fst x \<in> S \<and> snd x \<in> S'"
    by (auto simp add: KN.Sp_def)

  from `C \<subseteq> S` countable_space have [simp]: "countable C" by (rule countable_subset)

  { fix n :: nat
    have "(\<integral>\<^sup>+x. \<bar> p y x (Suc n) - measure N {x} \<bar> \<partial>count_space C) \<le>
      (\<integral>\<^sup>+x. ereal (\<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> \<not> (\<exists>i<n. smallest \<omega> D i))) +
      ereal (\<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> \<not> (\<exists>i<n. smallest \<omega> D i))) \<partial>count_space C)"
      using mono by (intro positive_integral_mono) simp
    also have "\<dots> = (\<integral>\<^sup>+x. \<P>(\<omega> in KN.P.paths (y, None). fst (\<omega> n) = x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)) \<partial>count_space C) +
      (\<integral>\<^sup>+x. \<P>(\<omega> in KN.P.paths (y, None). snd (\<omega> n) = Some x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)) \<partial>count_space C)"
      by (subst positive_integral_add) (simp_all add: measure_nonneg)
    also have "\<dots> = emeasure (KN.P.paths (y, None)) (\<Union>x\<in>C. {\<omega>\<in>space (KN.P.paths (y, None)). fst (\<omega> n) = x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)}) +
      emeasure (KN.P.paths (y, None)) (\<Union>x\<in>C. {\<omega>\<in>space (KN.P.paths (y, None)). snd (\<omega> n) = Some x \<and> \<not> (\<exists>i<n. smallest \<omega> D i)})"
      by (subst (1 2) emeasure_countable_UNION)
         (auto simp add: KN.space_S_seq_Sp KN.sets_S_seq_Sp disjoint_family_on_def KN.P.emeasure_eq_measure)
    also have "\<dots> \<le> ereal (\<P>(\<omega> in KN.P.paths (y, None). \<not> (\<exists>i<n. smallest \<omega> D i))) + ereal (\<P>(\<omega> in KN.P.paths (y, None). \<not> (\<exists>i<n. smallest \<omega> D i)))"
      unfolding KN.P.emeasure_eq_measure
      by (intro add_mono ereal_less_eq(3)[THEN iffD2] KN.P.finite_measure_mono)
         (auto simp: Sp_iff KN.space_S_seq_Sp KN.sets_S_seq_Sp)
    also have "\<dots> \<le> 2 * \<P>(\<omega> in KN.P.paths (y, None). \<not> (\<exists>i<n. smallest \<omega> D i))"
      by simp
    finally have "?L (Suc n) \<le> 2 * \<P>(\<omega> in KN.P.paths (y, None). \<not> (\<exists>i<n. smallest \<omega> D i))"
      by (intro integral_bound) simp_all }
  note le_2 = this

  have c0_D: "(c0, Some c0) \<in> D"
    by (simp add: D_def c0)

  let ?N' = "distr N (count_space S') Some"
  interpret N': prob_space ?N'
    by (rule N.prob_space_distr) simp
  interpret NP: pair_prob_space N ?N'
    by default

  have *: "Some -` {None} \<inter> S = {}" by auto
  have pos_recurrent: "\<forall>x\<in>C \<times> Some ` C. KN.P.pos_recurrent x"
  proof (rule KN.P.stationary_distributionD(1)[OF KN_essential KN.stationary_distributionI_pair[OF N(1,2)]])
    show "prob_space ?N'" by unfold_locales
    show "K2.stationary_distribution ?N'"
    proof (rule K2.stationary_distributionI)
      show "finite_measure ?N'" by unfold_locales
      show "sets ?N' = Pow S'" by simp
      fix y assume y: "y \<in> S'"
      show "(\<integral> x. measure (K' x) {y} \<partial>?N') \<le> measure ?N' {y}"
      proof (cases y)
        case None then show ?thesis
          apply simp
          apply (subst measure_distr)
          apply simp_all
          apply (subst integral_distr)
          apply (simp_all add: K'_def)
          apply (subst measure_distr)
          apply (simp_all add: K_measurable1 *)
          done
      next
        case (Some x)
        with y have *: "(Some -` {Some x} \<inter> S) = {x}" by auto
        from Some y show ?thesis
          apply simp
          apply (subst measure_distr)
          apply simp_all
          apply (subst integral_distr)
          apply (simp_all add: K'_def)
          apply (subst measure_distr)
          apply (simp_all add: K_measurable1 *)
          using stationary_distributionD_emeasure[OF N(1,2), of "{x}"]
          apply (simp add: N.emeasure_eq_measure K.emeasure_eq_measure)
          apply (subst (asm) positive_integral_eq_integral)
          apply (rule N.integrable_const_bound[where B=1])
          apply (auto simp: measure_nonneg)
          done
      qed
    qed
    show "prob_space (N \<Otimes>\<^sub>M distr N (count_space S') Some)" by unfold_locales
    have eq: "KN.Sp - C \<times> Some ` C = S \<times> (Some ` (S - C) \<union> {None}) \<union> (S - C) \<times> S'"
      unfolding KN.Sp_def by (auto simp add: S'_def inj_image_mem_iff)
    have eq2: "Some -` insert None (Some ` (S - C)) \<inter> S = S - C"
      by auto
    have "measure (N \<Otimes>\<^sub>M ?N') (KN.Sp - C \<times> Some ` C) \<le>  
      measure (N \<Otimes>\<^sub>M ?N') (S \<times> (Some ` (S - C) \<union> {None})) + measure (N \<Otimes>\<^sub>M ?N') ((S - C) \<times> S')"
      unfolding eq
      by (intro NP.finite_measure_subadditive pair_measureI) (auto simp add: S'_def)
    also have "measure (N \<Otimes>\<^sub>M ?N') (S \<times> (Some ` (S - C) \<union> {None})) = 0"
      using N(3) unfolding measure_def
      apply (subst N'.emeasure_pair_measure_Times)
      apply (auto simp add: emeasure_distr * S'_def)
      apply (subst (asm) emeasure_distr)
      apply (auto simp: eq2 N)
      done
    also have "measure (N \<Otimes>\<^sub>M ?N') ((S - C) \<times> S') = 0"
      using N(3) unfolding measure_def
      apply (subst N'.emeasure_pair_measure_Times)
      apply (auto simp add: * S'_def)
      done
    finally show "measure (N \<Otimes>\<^sub>M ?N') (KN.Sp - C \<times> Some ` C) = 0"
      by (simp add: antisym measure_nonneg)
  qed
  
  from c0_D have "measure (KN.P.paths (y, None)) {\<omega>\<in>space (KN.P.paths (y, None)). \<forall>i. \<omega> i \<notin> D} \<le>
    measure (KN.P.paths (y, None)) {\<omega>\<in>space (KN.P.paths (y, None)). \<forall>i. \<omega> i \<noteq> (c0, Some c0)}"
    apply (auto intro!: KN.P.finite_measure_mono)
    apply (erule_tac x=i in allE)
    apply simp
    done
  also have "\<dots> = 0"
    apply (rule KN.P.prob_eq_0_AE)
    apply (subst KN.P.AE_iterate)
    apply simp
    apply simp
    apply (subst KN.P.AE_K_iff)
    apply simp
  proof 
    fix t assume t: "t \<in> KN.P.E (y, None)"
    obtain a b where t_eq: "t = (a, b)" by (cases t) auto
    with t have ab: "a \<in> S" "b \<in> S'" and eq_0: "emeasure (KN.Kp (y, None)) {(a, b)} \<noteq> 0"
      unfolding KN.P.E_def by (auto simp add: KN.Sp_def)
    have "b \<noteq> None"
    proof
      assume "b = None"
      from ab have "emeasure (KN.Kp (y, None)) ({a} \<times> {b}) = emeasure (K y) {a} * emeasure (K' None) {b}"
        unfolding KN.Kp_def by (simp del: insert_times_insert add: K'.emeasure_pair_measure_Times)
      also have "emeasure (K' None) {b} = 0"
        by (simp add: K'_def emeasure_distr `b = None` *)
      finally show False
        using t by (simp add: KN.P.E_def t_eq)
    qed
    then obtain b' where b': "b = Some b'" "b' \<in> S"
      using `b \<in> S'` by auto
    then have eq2: "(Some -` {Some b'}) \<inter> S = {b'}" by auto
    from b' eq_0 have "emeasure (K y \<Otimes>\<^sub>M ?N') ({a} \<times> {Some b'}) \<noteq> 0"
      unfolding KN.Kp_def by (simp add: K'_def)
    then have neq_0: "emeasure (K y) {a} \<noteq> 0 \<and> emeasure N {b'} \<noteq> 0"
      by (subst (asm) N'.emeasure_pair_measure_Times)
         (simp_all add: emeasure_distr ab b' eq2)
    then have "a \<in> E y"
      by (simp add: E_def `a \<in> S`)
    with `y \<in> C` have "a \<in> C"
      using essential_classD2[OF `essential_class C` `y \<in> C`] by (auto simp: accessible_single)
    have "b' \<in> C"
    proof (rule ccontr)
      assume "b' \<notin> C"
      with `b' \<in> S` have "measure N {b'} \<le> measure N (S - C)"
        by (intro N.finite_measure_mono) auto
      then have "measure N {b'} = 0"
        by (simp add: measure_nonneg antisym N)
      with neq_0 show False
        by (simp add: N.emeasure_eq_measure)
    qed

    from pos_recurrent[THEN bspec, of "(c0, Some c0)"]
    have recurrent_c0: "KN.P.recurrent (c0, Some c0)"
      by (simp add: KN.P.pos_recurrent_def c0)
    have "C \<times> Some ` C \<in> KN.Sp // KN.P.communicating"
      using aperiodic by (simp add: KN.P.aperiodic_def)
    then have "((c0, Some c0), t) \<in> KN.P.communicating"
      by (rule KN.P.irreducibleD) (simp_all add: t_eq ab c0 b' `b' \<in> C` `a \<in> C`)
    then have "KN.P.accessible (c0, Some c0) t"
      by (simp add: KN.P.communicating_def)
    then have "KN.P.U t (c0, Some c0) = 1"
      by (rule KN.P.recurrent_accessible(1)[OF recurrent_c0])
    then have "AE \<omega> in KN.P.paths t. \<exists>i. \<omega> i = (c0, Some c0)"
      unfolding KN.P.U_def by (subst (asm) KN.P.prob_Collect_eq_1)  simp_all
    then show "AE \<omega> in KN.P.paths t. \<not> (\<forall>i. nat_case t \<omega> i \<noteq> (c0, Some c0))"
      apply eventually_elim
      apply auto
      apply (rule_tac x="Suc i" in exI)
      apply simp
      done
  qed
  finally have "measure (KN.P.paths (y, None)) {\<omega>\<in>space (KN.P.paths (y, None)). \<forall>i. \<omega> i \<notin> D} = 0"
    by (intro antisym measure_nonneg)
  
  have "(\<lambda>n. \<P>(\<omega> in KN.P.paths (y, None). \<not> (\<exists>i<n. smallest \<omega> D i))) ---->
    measure (KN.P.paths (y, None)) (\<Inter>n. {\<omega>\<in>space (KN.P.paths (y, None)). \<not> (\<exists>i<n. smallest \<omega> D i)})"
    by (rule KN.P.finite_Lim_measure_decseq) (auto simp: decseq_def)
  also have "(\<Inter>n. {\<omega>\<in>space (KN.P.paths (y, None)). \<not> (\<exists>i<n. smallest \<omega> D i)}) =
    {\<omega>\<in>space (KN.P.paths (y, None)). \<forall>i. \<omega> i \<notin> D}"
    by (auto intro: smallestI dest: smallestD1)
  also have "measure (KN.P.paths (y, None)) {\<omega>\<in>space (KN.P.paths (y, None)). \<forall>i. \<omega> i \<notin> D} = 0" by fact
  finally have *: "(\<lambda>n. 2 * \<P>(\<omega> in KN.P.paths (y, None). \<not> (\<exists>i<n. smallest \<omega> D i))) ----> 0"
    by (intro tendsto_eq_intros) auto

  show ?thesis
    apply (rule LIMSEQ_imp_Suc)
    apply (rule tendsto_sandwich[OF _ _ tendsto_const *])
    using le_2
    apply (simp_all add: lebesgue_integral_nonneg)
    done
qed

lemma stationary_distribution_imp_p_limit:
  assumes "aperiodic C" "essential_class C" and N: "stationary_distribution N" "prob_space N" "measure N (S - C) = 0"
  assumes [simp]: "x \<in> C" "y \<in> C"
  shows "p x y ----> measure N {y}"
proof -
  interpret N: prob_space N by fact
  def D \<equiv> "\<lambda>y n. \<bar>p x y n - Sigma_Algebra.measure N {y}\<bar>"

  from stationary_distribution_imp_limit[OF assms(1,2,3,4,5,6)]
  have INT: "(\<lambda>n. \<integral>y. D y n \<partial>count_space C) ----> 0"
    unfolding D_def .

  have "C \<subseteq> S"
    using `essential_class C`
    by (simp add: C_subset_S essential_class_def)
  then have [simp]: "countable C"
    using countable_space by (rule countable_subset)

  { fix n
    have "D y n \<le> (\<integral>z. D y n * indicator {y} z \<partial>count_space C)"
      by (subst integral_cmul_indicator) auto
    also have "\<dots> \<le> (\<integral>y. D y n \<partial>count_space C)"
      by (intro integral_mono integral_cmul_indicator)
         (auto split: split_indicator simp: D_def p_def disjoint_family_on_def
               intro!: integrable_abs integral_diff N.integrable_measure integrable_measure)
    finally have "D y n \<le> (\<integral>y. D y n \<partial>count_space C)" . }
  note * = this

  have D_nonneg: "\<And>n. 0 \<le> D y n" by (simp add: D_def)

  have "D y ----> 0"
    by (rule tendsto_sandwich[OF _ _ tendsto_const INT])
       (auto simp: eventually_sequentially * D_nonneg)
  then show ?thesis
    using Lim_null[where l="measure N {y}" and net=sequentially and f="p x y"]
    by (simp add: D_def tendsto_rabs_zero_iff)
qed

end

end

